/*
 * GIMP Tategaki Plug-in -- This is a plug-in for The GIMP 2
 * Copyright (C) 2006-2017 OHTSUKA, Yoshio <ohtsuka.yoshio@gmail.com>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 */
#include "tategaki.h"
#include "encoding.h"

gsize encode (gpointer in, gsize size, gchar** out,
		const gchar *from, const gchar *to)
{
	gsize rlen, wlen;
	GError *error = NULL;
	*out = g_convert(in, size, to, from, &rlen, &wlen, &error);

	if (error != NULL) {
		*out = NULL;
		return 0;
	}
	return wlen;
}


static guint32 encode_to_vender_unicode_win32 (guint32 in)
{
	switch(in) {
		case 0x301C: return 0xFF5E;//`
		case 0x00A5: return 0x005C;
		case 0x203E: return 0x007E;
		case 0x005C: return 0xFF3C;
		case 0x2016: return 0x2225;
		case 0x2212: return 0xFF0D;
		case 0x00A2: return 0xFFE0;
		case 0x00A3: return 0xFFE1;
		case 0x00AC: return 0xFFE2;
	}
	return in;
}
static guint32 encode_from_vender_unicode_win32 (guint32 in)
{
	switch(in) {
		case 0xFF5E: return 0x301C;//`
		case 0x005C: return 0x00A5;
		case 0x007E: return 0x203E;
		case 0xFF3C: return 0x005C;
		case 0x2225: return 0x2016;
		case 0xFF0D: return 0x2212;
		case 0xFFE0: return 0x00A2;
		case 0xFFE1: return 0x00A3;
		case 0xFFE2: return 0x00AC;
	}
	return in;
}


int check_face_charmap(FT_Face face)
{
	int i;
	if (face == NULL) return 1;
	if (face->num_charmaps < 1) return 1;
	if (face->charmap) return 0;

	//1. Search UNICODE charmap, and when not found ...
	//2. Next, select charmaps from platforms and when not found too ...
	//3. select first one.
	
	i = face->num_charmaps;
	while (i--) {
		if (face->charmaps[i]->encoding == FT_ENCODING_UNICODE) {
		FT_Select_Charmap(face, FT_ENCODING_UNICODE);
			return 0;
		}
	}
	i = face->num_charmaps;
	while (i--) {
#ifdef _WIN32
		if (face->charmaps[i]->platform_id == TT_PLATFORM_MICROSOFT) {
			FT_Select_Charmap(face, face->charmaps[i]->encoding);
			return 0;
		}
#endif
	}
	FT_Select_Charmap(face, face->charmaps[0]->encoding_id);
	return 0;
}

FT_UInt unicode_to_charmap_code (FT_Face face, guint32 in)
{
	gsize rlen = 0;
	gsize wlen = 0;
	gchar *out = NULL;
	GError *error = NULL;
	guint32  ret = 0;

	if (face->charmap == NULL) {
		return in;
	}

	switch ( face->charmap->encoding ) {
		case FT_ENCODING_MS_SYMBOL:
			wlen = ucs4_to_mssymbol(&in, sizeof(guint32), &out);
			break;

		case FT_ENCODING_UNICODE:
			return in;

		case FT_ENCODING_SJIS:
		{
			//int n = 0;

			in = encode_from_vender_unicode_win32 (in);
			out = g_convert((char*)&in, sizeof(gunichar), 
					"CP932", "UCS-4LE", &rlen, &wlen, &error);

			if (error) {
				WNGMSG("%s", error->message);
				return 0;
			}

			if ((0x81 <= ((guint8*)out)[0] && ((guint8*)out)[0] <=0x9F) ||
				(0xE0 <= ((guint8*)out)[0] && ((guint8*)out)[0] <=0xEF)) {
				ret = GUINT16_FROM_BE( *((guint16*)out) );
			} else {
				ret = ((guint8*)out)[0];
			}
			if (out) g_free(out);
			return ret;
		}
		case FT_ENCODING_NONE:
			break;

		case FT_ENCODING_GB2312:
			wlen = ucs4_to_gb2312(&in, sizeof(guint32), &out);
			break;

		case FT_ENCODING_BIG5:
			wlen = ucs4_to_big5(&in, sizeof(guint32), &out);
			break;

		case FT_ENCODING_WANSUNG:
			wlen = ucs4_to_wansung(&in, sizeof(guint32), &out);
			break;

		case FT_ENCODING_JOHAB:
			wlen = ucs4_to_johab(&in, sizeof(guint32), &out);
			break;

		case FT_ENCODING_ADOBE_STANDARD:
			wlen = ucs4_to_adobestandard(&in, sizeof(guint32), &out);
			break;

		case FT_ENCODING_ADOBE_EXPERT:
			wlen = ucs4_to_adobeexpert(&in, sizeof(guint32), &out);
			break;

		case FT_ENCODING_ADOBE_CUSTOM:
			wlen = ucs4_to_adobecustom(&in, sizeof(guint32), &out);
			break;

		case FT_ENCODING_ADOBE_LATIN_1:
			wlen = ucs4_to_adobelatin1(&in, sizeof(guint32), &out);
			break;

		case FT_ENCODING_OLD_LATIN_2:
			wlen = ucs4_to_adobelatin2(&in, sizeof(guint32), &out);
			break;

		case FT_ENCODING_APPLE_ROMAN:
			wlen = ucs4_to_appleroman(&in, sizeof(guint32), &out);
			break;

		default:
			ret = 0;
			break;
	}

	if (out) {
		if (wlen == 2) {
			ret = GUINT16_FROM_BE( *((guint16*)out) );
		} else {
			ret = ((guint8*)out)[0];
		}
		g_free(out);
	}
	return ret;
}

// UTF-8 NFC Normalize (except CJK Compatibiliby Kanji etc).
//http://www.unicode.org/Public/UNIDATA/CompositionExclusions.txt
gchar* utf8_nfc_except_cjk_kanji (gchar* src)
{
	gunichar unicode;
	gchar* dst = NULL;
	gchar* src_p = NULL;
	gchar* dst_p = NULL;
	gchar* seq_p = NULL;
	size_t byte_len;
	int counter = 0;

	byte_len = strlen(src);
	dst = g_malloc0(byte_len + 1);
	src_p = src;
	dst_p = dst;
	seq_p = NULL;

	while (*src_p) {
		gchar *next_src_p;
		unicode = g_utf8_get_char(src_p);
		if (
			// Singleton Decompositions (CJK COMPATIBILITY IDEOGRAPH)
			(0xF900 <= unicode && unicode <= 0xFA0D) ||
			unicode == 0xFA10 ||
			unicode == 0xFA12 ||
			(0xFA15 <= unicode && unicode <= 0xFA1E) ||
			unicode == 0xFA20 ||
			unicode == 0xFA22 ||
			(0xFA25 <= unicode && unicode <= 0xFA26) ||
			(0xFA2A <= unicode && unicode <= 0xFA6D) ||
			(0xFA70 <= unicode && unicode <= 0xFAD9) ||
			(0x2F800 <= unicode && unicode <= 0x2FA1D) ||
			// Script Specifics
			unicode == 0x0958 ||
			unicode == 0x0959 ||
			unicode == 0x095A ||
			unicode == 0x095B ||
			unicode == 0x095C ||
			unicode == 0x095D ||
			unicode == 0x095E ||
			unicode == 0x095F ||
			unicode == 0x09DC ||
			unicode == 0x09DD ||
			unicode == 0x09DF ||
			unicode == 0x0A33 ||
			unicode == 0x0A36 ||
			unicode == 0x0A59 ||
			unicode == 0x0A5A ||
			unicode == 0x0A5B ||
			unicode == 0x0A5E ||
			unicode == 0x0B5C ||
			unicode == 0x0B5D ||
			unicode == 0x0F43 ||
			unicode == 0x0F4D ||
			unicode == 0x0F52 ||
			unicode == 0x0F57 ||
			unicode == 0x0F5C ||
			unicode == 0x0F69 ||
			unicode == 0x0F76 ||
			unicode == 0x0F78 ||
			unicode == 0x0F93 ||
			unicode == 0x0F9D ||
			unicode == 0x0FA2 ||
			unicode == 0x0FA7 ||
			unicode == 0x0FAC ||
			unicode == 0x0FB9 ||
			unicode == 0xFB1D ||
			unicode == 0xFB1F ||
			unicode == 0xFB2A ||
			unicode == 0xFB2B ||
			unicode == 0xFB2C ||
			unicode == 0xFB2D ||
			unicode == 0xFB2E ||
			unicode == 0xFB2F ||
			unicode == 0xFB30 ||
			unicode == 0xFB31 ||
			unicode == 0xFB32 ||
			unicode == 0xFB33 ||
			unicode == 0xFB34 ||
			unicode == 0xFB35 ||
			unicode == 0xFB36 ||
			unicode == 0xFB38 ||
			unicode == 0xFB39 ||
			unicode == 0xFB3A ||
			unicode == 0xFB3B ||
			unicode == 0xFB3C ||
			unicode == 0xFB3E ||
			unicode == 0xFB40 ||
			unicode == 0xFB41 ||
			unicode == 0xFB43 ||
			unicode == 0xFB44 ||
			unicode == 0xFB46 ||
			unicode == 0xFB47 ||
			unicode == 0xFB48 ||
			unicode == 0xFB49 ||
			unicode == 0xFB4A ||
			unicode == 0xFB4B ||
			unicode == 0xFB4C ||
			unicode == 0xFB4D ||
			unicode == 0xFB4E ||
			// Post Composition Version precomposed characters
			unicode == 0x2ADC ||
			unicode == 0x1D15E ||
			unicode == 0x1D15F ||
			unicode == 0x1D160 ||
			unicode == 0x1D161 ||
			unicode == 0x1D162 ||
			unicode == 0x1D163 ||
			unicode == 0x1D164 ||
			unicode == 0x1D1BB ||
			unicode == 0x1D1BC ||
			unicode == 0x1D1BD ||
			unicode == 0x1D1BE ||
			unicode == 0x1D1BF ||
			unicode == 0x1D1C0 ||
			// Singleton Decompositions
			(0x0340 <= unicode && unicode <= 0x0341) ||
			unicode == 0x0343 ||
			unicode == 0x0374 ||
			unicode == 0x037E ||
			unicode == 0x0387 ||
			unicode == 0x1F71 ||
			unicode == 0x1F73 ||
			unicode == 0x1F75 ||
			unicode == 0x1F77 ||
			unicode == 0x1F79 ||
			unicode == 0x1F7B ||
			unicode == 0x1F7D ||
			unicode == 0x1FBB ||
			unicode == 0x1FBE ||
			unicode == 0x1FC9 ||
			unicode == 0x1FCB ||
			unicode == 0x1FD3 ||
			unicode == 0x1FDB ||
			unicode == 0x1FE3 ||
			unicode == 0x1FEB ||
			(0x1FEE <= unicode && unicode <= 0x1FEF) ||
			unicode == 0x1FF9 ||
			unicode == 0x1FFB || 
			unicode == 0x1FFD ||
			(0x2000 <= unicode && unicode <= 0x2001) ||
			unicode == 0x2126 ||
			(0x212A <= unicode && unicode <= 0x212B) ||
			unicode == 0x2329 ||
			unicode == 0x232A ||
			// Non-Starter Decompositions
			unicode == 0x0344 ||
			unicode == 0x0F73 ||
			unicode == 0x0F75 ||
			unicode == 0x0F81
			) {
			if (counter > 0) {
				gchar* tmp = g_utf8_normalize (seq_p, src_p - seq_p, G_NORMALIZE_NFC);
				g_utf8_strncpy(dst_p, tmp, counter);
				dst_p += src_p - seq_p;
				seq_p = NULL;
				g_free(tmp);
				counter = 0;
			}
			g_utf8_strncpy(dst_p, src_p, 1);
			next_src_p = g_utf8_find_next_char(src_p, NULL);
			if (next_src_p > 0) {
				dst_p += next_src_p - src_p;
			}
		} else {
			if (seq_p == NULL) {
				seq_p = src_p;
			}
			counter++;
		}
		src_p = g_utf8_find_next_char(src_p, NULL);
	}
	if (counter > 0) {
		gchar* tmp = g_utf8_normalize (seq_p, src_p - seq_p, G_NORMALIZE_NFC);
		g_utf8_strncpy(dst_p, tmp, counter);
		dst_p += src_p - seq_p;
		seq_p = NULL;
		g_free(tmp);
	}

	return dst;
}

/* vim: set fdm=marker ts=4 sw=4:*/
