/*
 * GIMP Tategaki Plug-in -- This is a plug-in for The GIMP 2
 * Copyright (C) 2006-2017 OHTSUKA, Yoshio <ohtsuka.yoshio@gmail.com>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 */
#include "tategaki.h"
#include "font_gsub_table.h"

#define LOOKUP_TYPE_SINGLE                           1
#define LOOKUP_TYPE_MULTIPLE                         2
#define LOOKUP_TYPE_ALTERNATE                        3
#define LOOKUP_TYPE_LIGATURE                         4
#define LOOKUP_TYPE_CONTEXT                          5
#define LOOKUP_TYPE_CHAINING_CONTEXT                 6
#define LOOKUP_TYPE_EXT_SUBSITUATION                 7
#define LOOKUP_TYPE_REVERSED_CHAINING_CONTEXT_SINGLE 8

#define LOOKUP_FLAG_RIGHTTOLEFT        0x0001
#define LOOKUP_FLAG_IGNOREBASEGLYPHS   0x0002
#define LOOKUP_FLAG_IGNORELIGATURES    0x0004
#define LOOKUP_FLAG_IGNOREMARKS        0x0008
#define LOOKUP_FLAG_RESERVED           0x00F0
#define LOOKUP_FLAG_MARKATTACHMENTTYPE 0xFF00


#define INT32(p)  GINT32_FROM_BE(*((gint32*)(p)))
#define UINT32(p) GUINT32_FROM_BE(*((guint32*)(p)))
#define INT16(p)  GINT16_FROM_BE(*((gint16*)(p)))
#define UINT16(p) GUINT16_FROM_BE(*((guint16*)(p)))

static int find_glyph_id_from_coverage_table(char *cov_table, guint16 glyph_id, int *type, int *index);
static int find_glyph_id_from_single_subst(char *subst_table, guint16 glyph_id);
static int find_glyph_id_from_ligature_subst(char *subst_table, int component_count, guint16 *component_glyphs);
static GlyphIdList find_glyph_id_from_lookup (char *lookup, GlyphIdList *in);

static int parse_lookup_list (FT_Byte *lookup_list, char ***lookup_tables, int *num);
static int parse_feature_table(FT_Byte *feature_table, guint32 **lookup_indexes, int *num);
static int parse_feature_list(FT_Byte *feature_list, FeatureRecord **feature_tables, int *num);
static int parse_GSUB_table (GSUB_Table *gsub);


static int find_glyph_id_from_coverage_table(char *cov_table, guint16 glyph_id, int *type, int *index)
{
	char *p = cov_table;
	guint16 format;

	format = UINT16(p); p+=2;
	*type = format;
	switch(format) {
		case 1:
			{
				guint16 *gidlist, key;
				int i;
				int count = UINT16(p); p+=2;
				gidlist = (guint16*)p;
				// use big endian to get more speed!
				key = GINT16_TO_BE(glyph_id);
				for (i=0; i<count; i++) {
					if (key == gidlist[i]) {
						*index = i;
						return GINT16_FROM_BE(key);
					}
				}
			}
			break;
		case 2:
			{
				guint16 gid_st, gid_ed, idx_st;
				int i;
				int count = UINT16(p); p+=2;

				for (i=0; i<count; i++) {
					gid_st = UINT16(p); p+=2;
					gid_ed = UINT16(p); p+=2;
					idx_st = UINT16(p); p+=2;
					if (gid_st <= glyph_id && glyph_id <= gid_ed) {
						*index = idx_st + glyph_id - gid_st;
						return glyph_id;
					} else {
						continue;
					}
				}
			}
			break;
	}
	return 0;
}

static int find_glyph_id_from_single_subst(char *subst_table, guint16 glyph_id)
{
	char *p = subst_table;
	guint16 format;

	format = UINT16(p); p+=2;
	switch (format) {
		case 1:
			{
				int ret, cov_type, cov_idx;
				int coverage_offset, delta_gid;
				coverage_offset = UINT16(p); p+=2;
				delta_gid = UINT16(p);

				ret = find_glyph_id_from_coverage_table(
						subst_table + coverage_offset, glyph_id,
						&cov_type, &cov_idx);
				if (ret == 0) return 0;
				return ret + delta_gid;
			}
			break;
		case 2:
			{
				int ret, cov_type, cov_idx;
				int coverage_offset, glyph_count;
				guint16 *glyph_list = NULL;
				coverage_offset = UINT16(p); p+=2;
				glyph_count = UINT16(p); p+=2;
				glyph_list = (guint16*)p;

				ret = find_glyph_id_from_coverage_table(
						subst_table + coverage_offset, glyph_id,
						&cov_type, &cov_idx);
				if (ret == 0) return 0;
				return GINT16_FROM_BE(glyph_list[cov_idx]);
			}
			break;

		default:
			return 0;
			break;
	}
	return 0;
}

static int find_glyph_id_from_ligature_subst(char *subst_table, int component_count, guint16 *component_glyphs)
{
	char *p = subst_table;
	guint16 format;

	if (component_count < 1) return 0;

	format = UINT16(p); p+=2;
	{
		int ret, cov_type, cov_idx;
		int coverage_offset, ligset_count;
		int i, j;

		coverage_offset = UINT16(p); p+=2;

		ret = find_glyph_id_from_coverage_table (
				subst_table + coverage_offset, component_glyphs[0],
				&cov_type, &cov_idx);

		if (ret == 0) return 0;

		ligset_count = UINT16(p); p+=2;
		p+=cov_idx*2;

		{
			guint16 ligset_offset;
			char *ligset_table;
			char *pp;
			guint16 lig_count;

			ligset_offset = UINT16(p); p+=2;
			ligset_table = subst_table + ligset_offset;
			pp = ligset_table;
			lig_count = UINT16(pp); pp+=2;

			for (i = 0; i < lig_count; i++) {
				guint16 lig_offset;
				char *lig;
				char *ppp;

				lig_offset = UINT16(pp); pp+=2;
				lig = ligset_table + lig_offset;
				ppp = lig;

				{
					guint16 lig_glyph;
					guint16 comp_count;
					lig_glyph = UINT16(ppp); ppp+=2;
					comp_count = UINT16(ppp); ppp+=2;
					if (component_count == comp_count) {
						for (j = 1; j <= comp_count; j++) {
							guint16 comp_glyph = UINT16(ppp); ppp+=2;
							if (comp_glyph != component_glyphs[j]) {
								break;
							}
							return lig_glyph;
						}
					}
				}
			}
		}
	}
	return 0;
}

//sub functions (find_glyph_id_from_XXX) will calld from this function.
//return 0 if not found.
static GlyphIdList find_glyph_id_from_lookup (char *lookup, GlyphIdList *in)
{
	int i ;
	char *p = lookup;
	guint16 ret;
	guint16 glyph_id, type, flag, count, offset;
	GlyphIdList retgid;

	type = UINT16(p); p+=2;
	flag = UINT16(p); p+=2;
	count = UINT16(p); p+=2;

	switch (type) {
		case LOOKUP_TYPE_SINGLE: //type1
			if (in->size < 1 || in->gid == NULL) break;
			glyph_id = in->gid[0];

			for(i=0; i<count; i++) {
				offset = UINT16(p); p+=2;
				ret = find_glyph_id_from_single_subst(
					lookup + offset, glyph_id);
				if (ret == 0) continue;
			}
			if (ret == 0) break;
			retgid.size = 1;
			retgid.gid = g_malloc(sizeof(guint16));
			retgid.gid[0] = ret;
			return retgid;
			break;
		case LOOKUP_TYPE_MULTIPLE: //type2
			// ** Not Impremented **
			break;
		case LOOKUP_TYPE_LIGATURE: //type4
			if (in->size < 1 || in->gid == NULL) break;

			for(i=0; i<count; i++) {
				offset = UINT16(p); p+=2;
				ret = find_glyph_id_from_ligature_subst (
					lookup + offset, in->size, in->gid);
				if (ret == 0) continue;
			}
			if (ret == 0) break;
			retgid.size = 1;
			retgid.gid = g_malloc(sizeof(guint16));
			retgid.gid[0] = ret;
			return retgid;
			break;
		case LOOKUP_TYPE_CONTEXT: //type5
		case LOOKUP_TYPE_CHAINING_CONTEXT: //type6
		case LOOKUP_TYPE_EXT_SUBSITUATION: //type7
		case LOOKUP_TYPE_REVERSED_CHAINING_CONTEXT_SINGLE: //type8
		default:
			// ** Not Impremented **
			break;
	}
	retgid.size = 0;
	retgid.gid = NULL;
	return retgid;
}


static int parse_lookup_list (FT_Byte *lookup_list, char ***lookup_tables, int *num)
{
	int i;
	FT_Byte *p = lookup_list;
	guint16 count;
	guint16 offset;
	count = UINT16(p); p+=2;

	if (count == 0) {
		*num = 0;
		*lookup_tables = NULL;
		return 0;
	}
	*num = count;
	*lookup_tables = g_malloc(sizeof(char*) * count);


	for (i=0; i<count; i++) {
		offset = UINT16(p); p+=2;
		(*lookup_tables)[i] = (char*)(lookup_list + offset);
	}
	return 0;
}

int parse_feature_table(FT_Byte *feature_table, guint32 **lookup_indexes, int *num)
{
	int i;
	FT_Byte *p = feature_table;
	guint16 param, count;
	param = UINT16(p); p+=2; //gp
	count = UINT16(p); p+=2;
	
	if (count == 0) {
		*num = 0;
		*lookup_indexes = NULL;
		return 0;
	}
	*num = count;
	*lookup_indexes = g_malloc(sizeof(guint32) * count);

	for (i=0; i< count; i++) {
		(*lookup_indexes)[i] = UINT16(p); p+=2;
	}
	return 0;
}

int parse_feature_list(FT_Byte *feature_list, FeatureRecord **feature_tables, int *num)
{
	int i;
	FT_Byte *p = feature_list;
	guint16 count;
	guint32 tag;
	guint16 offset;
	count = UINT16(p); p+=2;

	if (count == 0) {
		*feature_tables = NULL;
		*num = 0;
		return 0;
	}
	*num = count;
	*feature_tables = g_malloc(sizeof(FeatureRecord) * count);
	
	for (i=0; i< count; i++) {
		tag    = UINT32(p); p+=4;
		offset = UINT16(p); p+=2;
		(*feature_tables)[i].tag = tag;
		(*feature_tables)[i].feature = offset + feature_list;
	}
	return 0;
}

static int parse_GSUB_table (GSUB_Table *gsub)
{
	int i;
	FT_Byte *pt    = NULL;
	FT_Byte *start = gsub->table;
	guint16 version[2];
	guint16 script_list_offset, feature_list_offset, lookup_list_offset;

	char **lookup_table = NULL;
	int lookup_table_num;
	
	// read header
	pt = start;
	version[0] = UINT16(pt); pt+=2;
	version[1] = UINT16(pt); pt+=2;
	if (version[0] != 1 || version[1] != 0) {
		return 1;
	}
	script_list_offset  = UINT16(pt); pt+=2;
	feature_list_offset = UINT16(pt); pt+=2;
	lookup_list_offset  = UINT16(pt); pt+=2;
	if (script_list_offset  >= gsub->table_length ||
		feature_list_offset >= gsub->table_length ||
		lookup_list_offset  >= gsub->table_length) {
		return 1;
	}
	gsub->script_list_ptr  = start + script_list_offset;
	gsub->feature_list_ptr = start + feature_list_offset;
	gsub->lookup_list_ptr  = start + lookup_list_offset;

	// script list
	// ȗ
	
	// feature record list, feature table
	{
		FeatureRecord *feature_table;
		guint32 *lookup_index;
		int feature_table_num, lookup_index_num;

		parse_feature_list(start + feature_list_offset, &feature_table, &feature_table_num);
		
		gsub->feature_table = g_malloc(sizeof(FeatureTable) * feature_table_num);
		gsub->feature_table_size = feature_table_num;

		for (i=0; i<feature_table_num; i++) {
			parse_feature_table(feature_table[i].feature, &lookup_index, &lookup_index_num);

			gsub->feature_table[i].tag = feature_table[i].tag;
			gsub->feature_table[i].lookup_index_size = lookup_index_num;
			gsub->feature_table[i].lookup_index = lookup_index;
		}
		
		g_free(feature_table);
	}
	// lookup list
	parse_lookup_list(start + lookup_list_offset, &lookup_table, &lookup_table_num);
	gsub->lookup_size = lookup_table_num;
	gsub->lookup = lookup_table;
	return 0;
}

GSUB_Table* gsub_table_create (FT_Face face)
{
	int error;
	GSUB_Table *gsub = NULL;
	FT_Byte *table;
	FT_ULong length;

	length = 0;
	FT_Load_Sfnt_Table(face, TTAG_GSUB, 0, NULL, &length);
	if (length == 0)
		return NULL;

	gsub = g_malloc(sizeof(GSUB_Table));
	table = g_malloc(length);
	FT_Load_Sfnt_Table(face, TTAG_GSUB, 0, table, &length);

/*    {*/
/*        FILE *fp;*/
/*        fp = fopen("./gsub.dump", "wb");*/
/*        fwrite(table, sizeof(char), length, fp);*/
/*        fclose(fp);*/
/*    }*/

	gsub->table = table;
	gsub->table_length = length;

	error = parse_GSUB_table(gsub);
	if (error) {
		if (table) g_free(table);
		if (gsub) g_free(gsub);
		return NULL;
	}
	return gsub;
}

void gsub_table_destroy (GSUB_Table *gsub)
{
	gsize i;
	if (gsub == NULL) return;
	if (gsub->table) g_free(gsub->table);
	if (gsub->lookup) g_free(gsub->lookup);

	if (gsub->feature_table) {
		for (i=0; i<gsub->feature_table_size; i++)
			if (gsub->feature_table[i].lookup_index) {
				g_free(gsub->feature_table[i].lookup_index);
				gsub->feature_table[i].lookup_index = NULL;
			}

		g_free(gsub->feature_table);
	}

	g_free(gsub);
}

//out->gidNULLłȂΎIɉ̂ŒӁB
void gsub_table_get_gid(GSUB_Table *gsub, guint32 tag, GlyphIdList *in, GlyphIdList *out)
{
	gsize i,j;
	GlyphIdList ret;

	if (gsub == NULL) {
		goto notfound;
	}
	if (out->gid) {
		//g_free(out->gid);
		out->gid = NULL;
	}

	//tagT
	for (i=0; i<gsub->feature_table_size; i++) {
		if (gsub->feature_table[i].tag == tag) {
			gsize size = gsub->feature_table[i].lookup_index_size;

			//lookupΉOt
			for (j=0; j<size; j++) {
				int index;
				index = gsub->feature_table[i].lookup_index[j];
				ret = find_glyph_id_from_lookup (gsub->lookup[index], in);
				if (ret.size > 0) {
					out->size = ret.size;
					out->gid = ret.gid;
					return;
				}
			}
		}
	}

notfound:
	out->size = 0;
	out->gid = NULL;
	return;
}

/* vim: set fdm=marker ts=4 sw=4:*/
