/*
 * GIMP Tategaki Plug-in -- This is a plug-in for The GIMP 2
 * Copyright (C) 2006-2017 OHTSUKA, Yoshio <ohtsuka.yoshio@gmail.com>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 */
#include <ft2build.h>
#include FT_FREETYPE_H
#include FT_SFNT_NAMES_H
#include FT_TRUETYPE_IDS_H
#include <glib.h>
#include "tategaki.h"
#include "font_list.h"

// callback compare function
static gint comp_FontInfo_name(const FontInfo *a, const FontInfo *b)
{
	return strcmp(a->key, b->key);
}

// used for debug message
static void output_FontInfo(FontInfo *finfo, gpointer dammy)
{
	DBGMSG("	%s	%s\n", finfo->name, finfo->key);
}


static gchar* get_sfnt_name_utf8_apple_unicode (FT_SfntName *sfnt)
{
	gchar *temp = NULL;
	gsize temp_len;
	switch (sfnt->encoding_id) {
		case TT_APPLE_ID_DEFAULT:
			break;
		case TT_APPLE_ID_UNICODE_1_1:
			break;
		case TT_APPLE_ID_UNICODE_2_0:
			temp_len = utf16be_to_utf8(sfnt->string, sfnt->string_len, &temp);
			break;
		case TT_APPLE_ID_UNICODE_32:
			temp_len = ucs4_to_utf8(sfnt->string, sfnt->string_len, &temp);
			break;
		case TT_APPLE_ID_ISO_10646:
			break;
		case TT_APPLE_ID_VARIANT_SELECTOR:
			break;
	}
	return temp;
}

static gchar* get_sfnt_name_utf8_macintosh (FT_SfntName *sfnt)
{
	gchar *temp = NULL;
	gsize temp_len;
	switch (sfnt->encoding_id) {
		case TT_MAC_ID_ROMAN:
			temp_len = symbol_to_utf8(sfnt->string, sfnt->string_len, &temp);
			break;
		case TT_MAC_ID_JAPANESE:
			temp_len = cp932_to_utf8(sfnt->string, sfnt->string_len, &temp);
			break;
		case TT_MAC_ID_TRADITIONAL_CHINESE:
			temp_len = big5_to_utf8(sfnt->string, sfnt->string_len, &temp);
			break;
		case TT_MAC_ID_SIMPLIFIED_CHINESE:
			temp_len = gb2312_to_utf8(sfnt->string, sfnt->string_len, &temp);
			break;
		default:
			break;
	}
	return temp;
}

static gchar* get_sfnt_name_utf8_microsoft (FT_SfntName *sfnt)
{
	gchar *temp = NULL;
	gsize temp_len;
	switch (sfnt->encoding_id) {
		case TT_MS_ID_UNICODE_CS:
			temp_len = utf16be_to_utf8(sfnt->string, sfnt->string_len, &temp);
			break;
		case TT_MS_ID_SYMBOL_CS:
			temp_len = symbol_to_utf8(sfnt->string, sfnt->string_len, &temp);
			break;
		case TT_MS_ID_SJIS:
			temp_len = cp932_to_utf8(sfnt->string, sfnt->string_len, &temp);
			break;
		case TT_MS_ID_GB2312:
			temp_len = gb2312_to_utf8(sfnt->string, sfnt->string_len, &temp);
			break;
		case TT_MS_ID_BIG_5:
			temp_len = big5_to_utf8(sfnt->string, sfnt->string_len, &temp);
			break;
		case TT_MS_ID_WANSUNG:
			temp_len = wansung_to_utf8(sfnt->string, sfnt->string_len, &temp);
			break;
		case TT_MS_ID_JOHAB:
			temp_len = johab_to_utf8(sfnt->string, sfnt->string_len, &temp);
			break;
		case TT_MS_ID_UCS_4:
			temp_len = ucs4_to_utf8(sfnt->string, sfnt->string_len, &temp);
			break;
		default:
			break;
	}
	return temp;
}

static gchar* get_sfnt_name_utf8 (FT_SfntName *sfnt)
{
	switch (sfnt->platform_id) {
		case TT_PLATFORM_MICROSOFT:
			return get_sfnt_name_utf8_microsoft(sfnt);
		case TT_PLATFORM_MACINTOSH:
			return get_sfnt_name_utf8_macintosh(sfnt);
		case TT_PLATFORM_APPLE_UNICODE:
			return get_sfnt_name_utf8_apple_unicode(sfnt);
	}
	return NULL;
}

// return font names.
// out_name is japanese full name
// out_name_en is English full name
// out_name_old is old style(before version 1.0) japanese font family name + style name.
// out_name_ps is postscript name
static void get_fontname (FT_Face face,
						  gchar **out_name, 
						  gchar **out_name_en,
						  gchar **out_name_old,
						  gchar **out_name_ps)
{
	FT_Error error;
	int i, count;
	FT_SfntName sfnt;
	
	gchar *name = NULL;
	gchar *style = NULL;
	gchar *full_name = NULL;
	gchar *ps_name = NULL;
	gchar *uniq_id = NULL;
	gchar *name_en = NULL;
	gchar *style_en = NULL;
	gchar *full_name_en = NULL;
	gchar *ps_name_en = NULL;

	count = FT_Get_Sfnt_Name_Count(face);
	for (i=0; i<count; i++) {
		gchar *temp = NULL;
		gboolean is_ascii = FALSE;

		error = FT_Get_Sfnt_Name(face, i, &sfnt);
		if (error) {
			goto PARSE_SFNT_NAME_LOOP;
		}

		// Fix some invalid parameter for some fonts.
		// CRccgbqb.ttc CRccghkm.ttc CRccgrtr.ttc CRccgryb.ttc CRccgrym.ttc
		if (sfnt.platform_id == TT_PLATFORM_MACINTOSH &&
			sfnt.language_id == TT_MS_LANGID_JAPANESE_JAPAN && 
			sfnt.encoding_id == TT_MS_ID_SYMBOL_CS) {
			sfnt.platform_id = TT_PLATFORM_MICROSOFT;
			sfnt.encoding_id = TT_MS_ID_SJIS;
		}
		//Fix some fonts encoding_id is TT_MS_ID_SJIS, but the raw encoding is unicode!
		//       We must check if is valid or not. (JRLM00M.TTF, JSLM00H.TTF, JSLM00L.TTF)
		if (sfnt.language_id == TT_MS_LANGID_ENGLISH_UNITED_STATES && 
			sfnt.platform_id == TT_PLATFORM_MICROSOFT &&
			sfnt.encoding_id == TT_MS_ID_SJIS &&
			face->family_name == NULL) {
			sfnt.encoding_id = TT_MS_ID_UNICODE_CS;
		}

		// Fix some invalid parameter for some fonts.
		if (sfnt.platform_id == TT_PLATFORM_MACINTOSH &&
			sfnt.language_id == TT_MS_LANGID_ENGLISH_UNITED_STATES) {
			sfnt.language_id = TT_MAC_LANGID_ENGLISH;
		}
		if (sfnt.platform_id == TT_PLATFORM_MACINTOSH &&
			sfnt.language_id == TT_MS_LANGID_JAPANESE_JAPAN) {
			sfnt.language_id = TT_MAC_LANGID_JAPANESE;
		}

		if (sfnt.language_id == TT_MS_LANGID_JAPANESE_JAPAN && 
			sfnt.platform_id == TT_PLATFORM_MICROSOFT &&
			sfnt.encoding_id == TT_MS_ID_SJIS &&
			face->family_name == NULL) {
			sfnt.encoding_id = TT_MS_ID_UNICODE_CS;
		}

		if (g_strcmp0(face->family_name, "Marlett") == 0 &&
			sfnt.platform_id == TT_PLATFORM_MICROSOFT &&
			sfnt.encoding_id == TT_MS_ID_SYMBOL_CS &&
			sfnt.language_id == TT_MS_LANGID_ENGLISH_UNITED_STATES) {
			sfnt.encoding_id = TT_MS_ID_UNICODE_CS;
		}

		// Fix Dynafont invalid parameters.
		if (sfnt.platform_id == TT_PLATFORM_MACINTOSH &&
			sfnt.encoding_id == TT_MAC_ID_ROMAN &&
			sfnt.language_id == TT_MAC_LANGID_ENGLISH) {
			sfnt.language_id = TT_MAC_LANGID_JAPANESE;
			sfnt.encoding_id = TT_MAC_ID_JAPANESE;
		}

		//Skip this parameter. because some Dynafont (DFGOTU_5.TTF, DFMINU_3.TTF)
		// has utf-16le but the unicode must be big endian.
		if (sfnt.platform_id == TT_PLATFORM_APPLE_UNICODE) {
			goto PARSE_SFNT_NAME_LOOP;
		}

		if (sfnt.platform_id == TT_PLATFORM_MICROSOFT) {
			switch (sfnt.language_id) {
				case 0x0000:
				case TT_MS_LANGID_ENGLISH_GENERAL:
				case TT_MS_LANGID_ENGLISH_UNITED_STATES:
					is_ascii = TRUE;
					break;
				case TT_MS_LANGID_JAPANESE_JAPAN:
				case TT_MS_LANGID_CHINESE_GENERAL:
				case TT_MS_LANGID_CHINESE_PRC:
				case TT_MS_LANGID_CHINESE_HONG_KONG:
				case TT_MS_LANGID_CHINESE_TAIWAN:
				case TT_MS_LANGID_CHINESE_SINGAPORE:
					is_ascii = FALSE;
					break;
				default:
					goto PARSE_SFNT_NAME_LOOP;
			}
		}

		if (sfnt.platform_id == TT_PLATFORM_MACINTOSH) {
			switch (sfnt.language_id) {
				case TT_MAC_LANGID_ENGLISH:
					is_ascii = TRUE;
					break;
				case TT_MAC_LANGID_JAPANESE:
				case TT_MAC_LANGID_CHINESE_SIMPLIFIED:
				case TT_MAC_LANGID_CHINESE_TRADITIONAL:
					is_ascii = FALSE;
					break;
				default:
					goto PARSE_SFNT_NAME_LOOP;
			}
		}

		if (sfnt.string_len == 0) {
			goto PARSE_SFNT_NAME_LOOP;
		}

		temp = get_sfnt_name_utf8 (&sfnt);

		// Check if ascii only, if so it is is_ascii = TRUE.
		if (temp) {
			gchar *p = temp;
			gboolean _is_ascii = TRUE;
			while (*p) {
				if (g_utf8_get_char(p) > 0xFF) {
					_is_ascii = FALSE;
					break;
				}
				p = g_utf8_next_char(p);
			}
			is_ascii = _is_ascii;
		}

		switch (sfnt.name_id) {
			case TT_NAME_ID_FONT_FAMILY:
				if (is_ascii) {
					if (name_en) goto PARSE_SFNT_NAME_LOOP;
					name_en = g_strdup(temp);
				} else {
					if (name) goto PARSE_SFNT_NAME_LOOP;
					name = g_strdup(temp);
				}
				break;
			case TT_NAME_ID_FONT_SUBFAMILY:
				if (is_ascii) {
					if (style_en) goto PARSE_SFNT_NAME_LOOP;
					style_en = g_strdup(temp);
				} else {
					if (style) goto PARSE_SFNT_NAME_LOOP;
					style = g_strdup(temp);
				}
				break;
			case TT_NAME_ID_FULL_NAME:
				if (is_ascii) {
					if (full_name_en) goto PARSE_SFNT_NAME_LOOP;
					full_name_en = g_strdup(temp);
				} else {
					if (full_name) goto PARSE_SFNT_NAME_LOOP;
					full_name = g_strdup(temp);
				}
				break;
			case TT_NAME_ID_PS_NAME:
				if (is_ascii) {
					if (ps_name_en) goto PARSE_SFNT_NAME_LOOP;
					ps_name_en = g_strdup(temp);
				} else {
					if (ps_name) goto PARSE_SFNT_NAME_LOOP;
					ps_name = g_strdup(temp);
				} 
				break;
			case TT_NAME_ID_UNIQUE_ID:
				if (uniq_id) goto PARSE_SFNT_NAME_LOOP;
				uniq_id = g_strdup(temp);
				break;
			default:
				continue;
		}
PARSE_SFNT_NAME_LOOP:
		if (temp) g_free(temp);
	}

	{
		GString *t = NULL;

		// old style font name
		if (name || name_en) {
			t = g_string_new(name ? name : name_en);
			if (style || style_en) {
				g_string_append(t, " ");
				g_string_append(t, style ? style : style_en);
			}
			*out_name_old = g_strdup(t->str);
		} else {
			*out_name_old = g_strdup(face->family_name);
		}
		// font name ja
		if (full_name || full_name_en) {
			*out_name = g_strdup(full_name ? full_name : full_name_en);
		} else if (name || name_en) {
			*out_name = g_strdup(name ? name : name_en);
		} else {
			*out_name = g_strdup(face->family_name);
		}
		// font name en
		if (full_name_en) {
			*out_name_en = g_strdup(full_name_en);
		} else if (name_en) {
			*out_name_en = g_strdup(name_en);
		} else if (ps_name) {
			*out_name_en = g_strdup(ps_name);
		}
		// postscript name
		if (ps_name_en) {
			*out_name_ps = g_strdup(ps_name_en);
		} else {
			*out_name_ps = g_strdup(ps_name);
		}

		if (t) g_string_free(t, TRUE);
		if (name) g_free(name);
		if (style) g_free(style);
		if (full_name) g_free(full_name);
		if (ps_name) g_free(ps_name);
		if (uniq_id) g_free(uniq_id);
		if (name_en) g_free(name_en);
		if (style_en) g_free(style_en);
		if (full_name_en) g_free(full_name_en);
		if (ps_name_en) g_free(ps_name_en);
		return;
	}
}

//append font to list
static int filelist_append_file (FT_Library library, gchar *file, GList **list)
{
	int i;
	FT_Error error = 0;
	FT_Face face = NULL;
	gchar *font_path = NULL;
	GError *gerror = NULL;
	gchar *ext; 

	ext = (gchar*)file + strlen(file) - 4;
	if (g_ascii_strncasecmp(".ttf", ext, 4) != 0 &&
		g_ascii_strncasecmp(".otf", ext, 4) != 0 &&
		g_ascii_strncasecmp(".ttc", ext, 4) != 0) {
		return 1;
	}

#ifdef _WIN32
	//file is utf-8, but freetype requires system locale.
	font_path = g_win32_locale_filename_from_utf8(file);
#else
	font_path = file;
#endif

	i = 0;
	while (!(error = FT_New_Face(library, font_path, i, &face))) {
		FontInfo *finfo = NULL;
		
		finfo = g_malloc0(sizeof(FontInfo));//Don't Free this!
		finfo->path = g_strdup(file);
		finfo->face_index = i;

		DBGMSG("%-46s ", file);
		DBGMSG("SFNT=%d ", FT_IS_SFNT(face) ? 1:0);
		DBGMSG("face_index = %d\n", finfo->face_index);
		//g_usleep(5000000);

		if (! FT_IS_SFNT(face)) {
			FT_Done_Face(face);
			if (finfo) {
				g_free(finfo->path);
				g_free(finfo);
			}
			continue;
		}

		// get ja/cn name as priority
		{
			gchar *name = NULL;
			gchar *name_en = NULL;
			gchar *name_old = NULL;
			gchar *name_ps = NULL;
			get_fontname (face, &name, &name_en, &name_old, &name_ps);

			// sort key
			if (name_en) {
				finfo->key = g_strdup(name_en);
			} else if (name) {
				finfo->key = g_strdup(name);
			} else if (face->family_name && face->style_name) {
				finfo->key = g_strconcat(face->family_name, " ", face->style_name, NULL);
			} else {
				DBGMSG("skipped loading font(no name):%s\n", file);
				i++;
				continue; //skip loading this font. because this font has no name!
			}
			if (name) {
				finfo->name = g_strdup(name);
			} else {
				finfo->name = g_strdup(finfo->key);
			}
			if (name_en) {
				finfo->name_en = g_strdup(name_en);
			} else {
				finfo->name_en = NULL;
			}
			if (name_old) {
				finfo->name_old = g_strdup(name_old);
			} else {
				finfo->name_old = NULL;
			}
			if (name_ps) {
				finfo->name_ps = g_strdup(name_ps);
			} else {
				finfo->name_ps = NULL;
			}
			if (name) g_free(name);
			if (name_en) g_free(name_en);
			if (name_old) g_free(name_old);
			if (name_ps) g_free(name_ps);
		}
		DBGMSG("key      = %s\n", finfo->key);
		DBGMSG("name     = %s\n", finfo->name);
		DBGMSG("name_en  = %s\n", finfo->name_en);
		DBGMSG("name_old = %s\n", finfo->name_old);
		DBGMSG("name_ps  = %s\n", finfo->name_ps);

		*list = g_list_append(*list, finfo);
		i++;

		if (i >= face->num_faces) {
			FT_Done_Face(face);
			break;
		} else {
			FT_Done_Face(face);
		}
	}

#ifdef _WIN32
	if (font_path) g_free(font_path);
#endif
	return 0;
}

static int filelist_append_dir (FT_Library library, const gchar *path, GList **list)
{
	GDir *dir = NULL;
	gchar *font_file = NULL;
	gchar *font_path = NULL;

	dir = g_dir_open(path, 0, NULL);
	if (dir == NULL) return 1;

	// font_file is owned by glib. DON'T FREE IT.
	while ((font_file = (gchar*)g_dir_read_name(dir))) {
		font_path = g_build_filename(path, font_file, NULL);
		
		filelist_append_file (library, font_path, list);
		
		if (font_path) {
			g_free(font_path);
			font_path = NULL;
		}
	}
	if (dir) g_dir_close(dir);
	return 0;
}

//build font list
GList* font_list_create (GList *itemlist)
{
	FT_Library library;
	FT_Error error;
	GList *list = NULL;
	GList *filepath = NULL;

	error = FT_Init_FreeType(&library);
	if (error) return NULL;

	filepath = itemlist;
	while (filepath) {
		if (filepath->data) {
			if (g_file_test(filepath->data, G_FILE_TEST_EXISTS) == TRUE) {
				if (g_file_test(filepath->data, G_FILE_TEST_IS_DIR) == TRUE) {
					filelist_append_dir (library, filepath->data, &list);
				} else {
					filelist_append_file (library, filepath->data, &list);
				}
			}
		}
		filepath = g_list_next(filepath);
	}

	FT_Done_FreeType(library);
	list = g_list_sort(list, (GCompareFunc)comp_FontInfo_name);
	return list;
}

int font_list_destroy(GList *list)
{
	GList *temp = g_list_first(list);
	while (temp) {
		FontInfo *finfo = temp->data;
		if (finfo) {
			if (finfo->name) g_free(finfo->name);
			if (finfo->key) g_free(finfo->key);
			if (finfo->name_old) g_free(finfo->name_old);
			if (finfo->name_en) g_free(finfo->name_en);
			if (finfo->path) g_free(finfo->path);
			g_free(finfo);
		}
		temp = g_list_next(temp);
	}
	g_list_free(list);
	return 0;
}

GList* font_list_find_name(GList *list, gchar *name)
{
	GList *temp = g_list_first(list);
	while (temp) {
		FontInfo *finfo = temp->data;
		if (g_strcmp0(name, finfo->name) == 0 ||
			g_strcmp0(name, finfo->name_en) == 0 ||
			g_strcmp0(name, finfo->name_old) == 0) {
			return temp;
		}
		temp = g_list_next(temp);
	}
	return NULL;
}

/* vim: set fdm=marker ts=4 sw=4:*/
