/*
 * GIMP Tategaki Plug-in -- This is a plug-in for The GIMP 2
 * Copyright (C) 2006-2017 OHTSUKA, Yoshio <ohtsuka.yoshio@gmail.com>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 */
#include "tategaki.h"
#include "font_mort_table.h"

typedef unsigned int fixed32;

#define INT32(p)  GINT32_FROM_BE(*((gint32*)(p)))
#define UINT32(p) GUINT32_FROM_BE(*((guint32*)(p)))
#define INT16(p)  GINT16_FROM_BE(*((gint16*)(p)))
#define UINT16(p) GUINT16_FROM_BE(*((guint16*)(p)))

guint16 mort_table_find_vertical_glyph_id_from_lookup (FT_Byte *lookup_table, guint16 glyph_index)
{
	FT_Byte *pt = NULL;
	FT_Byte *start =lookup_table;
	guint16 format;
	static struct {
		guint16 unitSize;
		guint16 nUnits;
		guint16 searchRange;
		guint16 entrySelector;
		guint16 rangeShift;
	} search_header;
#define GET_SEARCH_HEADER(s) \
	(s).unitSize      = UINT16(pt); pt+=2;\
	(s).nUnits        = UINT16(pt); pt+=2;\
	(s).searchRange   = UINT16(pt); pt+=2;\
	(s).entrySelector = UINT16(pt); pt+=2;\
	(s).rangeShift    = UINT16(pt); pt+=2;

	pt = start;
	format = UINT16(pt); pt+=2;
	switch (format) {
		case 0:
			//not inpremented
			break;
		case 2:
			//not impremented
			GET_SEARCH_HEADER(search_header);
			break;
		case 4:
			//not impremented
			GET_SEARCH_HEADER(search_header);
			break;
		case 6:
			GET_SEARCH_HEADER(search_header);
			{
				int i=0;
				guint16 *data;
				guint16 key = GUINT16_TO_BE(glyph_index);
				data = (guint16*)pt;

				for (i=0; i<search_header.nUnits*2; i+=2) {
					if (key == data[i]) {
						return INT16(data+i+1);
					}
				}
				return 0;
			}
			break;
		case 8:
			//not impremented
			break;
	}
	return 0;
}


static int parse_subtable (FT_Byte *subtale_table, mort_Subtable *subtable)
{
	FT_Byte *pt    = NULL;
	FT_Byte *start = subtale_table;
	pt = start;
	subtable->addr = pt;
	subtable->length            = UINT16(pt); pt+=2;
	subtable->coverage          = UINT16(pt); pt+=2;
	subtable->sub_feature_flags = UINT32(pt); pt+=4;
	subtable->data        = pt; 
	subtable->data_length = subtable->length - 8;
	return 0;
}

static int parse_feature (FT_Byte *feature_table, guint32 default_flags, 
		mort_Feature *feature)
{
	FT_Byte *pt    = NULL;
	guint32 enable_flags;
	guint32 disable_flags;
	FT_Byte *start = feature_table;

	pt = start;
	feature->type    = UINT16(pt); pt+=2;
	feature->setting = UINT16(pt); pt+=2;
	enable_flags     = UINT32(pt); pt+=4;
	disable_flags    = UINT32(pt); pt+=4;

	feature->flags = (default_flags & disable_flags) | enable_flags;
	return 0;
}


static int parse_chain_header (FT_Byte *chain_header, mort_Table *mort)
{
	guint32 i;
	FT_Byte *pt    = NULL;
	FT_Byte *start = chain_header;
	guint32 chain_num;
	
	pt = start;
	chain_num = UINT32(pt); pt+=4;

	if (chain_num < 1) {
		mort->chain_num = 0;
		mort->chains = NULL;
		return 0;
	}
	mort->chain_num = chain_num;
	mort->chains = g_malloc(sizeof(mort_Chain) * chain_num);

	for (i = 0; i < chain_num; i++) {
		int fnum, mnum, k;
		mort->chains[i].addr = pt;
		mort->chains[i].flags               = UINT32(pt); pt+=4;
		mort->chains[i].length              = UINT32(pt); pt+=4;
		mort->chains[i].feature_num  = fnum = UINT16(pt); pt+=2;
		mort->chains[i].subtable_num = mnum = UINT16(pt); pt+=2;
		mort->chains[i].features  = g_malloc(sizeof(mort_Feature) * fnum);
		mort->chains[i].subtables = g_malloc(sizeof(mort_Subtable) * mnum);

		for (k = 0; k < fnum; k++) {
			parse_feature(pt, mort->chains[i].flags, mort->chains[i].features + k);
			pt+= 12; // Feature Table Size
		}

		for (k = 0; k < mnum; k++) {
			parse_subtable(pt, mort->chains[i].subtables + k);
			pt+=mort->chains[i].subtables[k].length;
		}
		pt = mort->chains[i].addr + mort->chains[i].length;
	}
	return 0;
}

static int parse_mort_table (mort_Table *mort)
{
	FT_Byte *pt    = NULL;
	FT_Byte *start = mort->data;
	fixed32 version;
	
	//read header
	pt = start;
	version = UINT32(pt); pt+=4;
	if (version != 0x00010000)
		return 1;
	parse_chain_header(pt, mort);
	return 0;
}


mort_Table* mort_table_create (FT_Face face)
{
	int error;
	mort_Table *mort = NULL;
	FT_Byte *table;
	FT_ULong length;
	FT_ULong tag_mort = FT_MAKE_TAG('m','o','r','t');

	length = 0;
	FT_Load_Sfnt_Table(face, tag_mort, 0, NULL, &length);
	if (length == 0)
		return NULL;

	mort = g_malloc(sizeof(mort_Table));
	table = g_malloc(length);
	FT_Load_Sfnt_Table(face, tag_mort, 0, table, &length);
/*    {*/
/*        FILE *fp;*/
/*        fp = fopen("./mort.dump", "wb");*/
/*        fwrite(table, sizeof(char), length, fp);*/
/*        fclose(fp);*/
/*    }*/
	
	mort->data = table;
	mort->data_length = length;

	error = parse_mort_table (mort);
	if (error) {
		if (table) g_free(table);
		if (mort) g_free(mort);
	}
	return mort;
}

void mort_table_destroy (mort_Table *mort)
{
	if (mort) {
		if (mort->chains) {
			if (mort->chains->features) g_free(mort->chains->features);
			if (mort->chains->subtables) g_free(mort->chains->subtables);
			g_free(mort->chains);
		}
		if (mort->data) g_free(mort->data);
		g_free(mort);
	}
}

/* vim: set fdm=marker ts=4 sw=4:*/
