/*
 * GIMP Tategaki Plug-in -- This is a plug-in for The GIMP 2
 * Copyright (C) 2006-2017 OHTSUKA, Yoshio <ohtsuka.yoshio@gmail.com>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 */
#include <ft2build.h>
#include FT_FREETYPE_H
#include <glib.h>

#include "tategaki.h"
#include "font_vert_cache_table.h"


void vert_cache_table_destroy (VertCacheTable *vcache)
{
	if (vcache == NULL) return;

	if (vcache->vert_table) {
		g_hash_table_destroy(vcache->vert_table);
		vcache->vert_table = NULL;
	}
	if (vcache->gsub_table) {
		gsub_table_destroy(vcache->gsub_table);
		vcache->gsub_table = NULL;
	}
	if (vcache->mort_table) {
		mort_table_destroy(vcache->mort_table);
		vcache->mort_table = NULL;
	}

	g_free(vcache);
}


VertCacheTable* vert_cache_table_create (FT_Face face)
{
	VertCacheTable *vcache = g_malloc0(sizeof(VertCacheTable));
	if (vcache) {
		vcache->face = face;
		vert_cache_table_reset(vcache);
	}
	return vcache;
}


int vert_cache_table_reset (VertCacheTable *vcache)
{
	FT_Face face = vcache->face;

	if (vcache->vert_table) {
		g_hash_table_destroy(vcache->vert_table);
	}
	vcache->vert_table = g_hash_table_new (g_direct_hash, g_direct_equal);

	if (vcache->gsub_table) {
		gsub_table_destroy(vcache->gsub_table);
	}
	vcache->gsub_table = gsub_table_create(face);
	
	if (vcache->mort_table) {
		mort_table_destroy(vcache->mort_table);
	}
	vcache->mort_table = mort_table_create(face);

	return 0;
}


// Get vert glyph id. but if cache doesn't have it, return 0.
FT_UInt vert_cache_table_get_glyph_id (VertCacheTable *vcache, FT_UInt gid)
{
	FT_Face    face        = vcache->face;
	GSUB_Table *gsub_table = vcache->gsub_table;
	mort_Table *mort_table = vcache->mort_table;
	GHashTable *vert_table = vcache->vert_table;
	gpointer pvgid = NULL;
	FT_UInt vgid = 0;
	FT_ULong tag_vert = FT_MAKE_TAG('v','e','r','t');
	FT_ULong tag_vrt2 = FT_MAKE_TAG('v','r','t','2');
	static GlyphIdList gidout = {0, 0};
	static GlyphIdList *gidin = NULL;

	if (gidin == NULL) {
		gidin = g_malloc(sizeof(GlyphIdList));
		gidin->size = 1;
		gidin->gid = g_malloc(sizeof(guint16));
	}
	if (gidin == NULL) {
		return 0;
	}

	//find from cache table
	pvgid = g_hash_table_lookup(vert_table, GUINT_TO_POINTER(gid));
	if ( g_hash_table_lookup_extended(vert_table, 
				GUINT_TO_POINTER(gid), NULL, &pvgid) == TRUE ) {
		vgid = GPOINTER_TO_UINT(pvgid);
		return vgid;
	}
	
	//if the id not found in cache table, find from gsub and mort
	
	//gsub(vrt2)
	/*
	if (gsub_table) {
		gidin->gid[0] = gid;
		gsub_get_gid(gsub_table, tag_vrt2, gidin, &gidout);
		if (gidout.size > 0) {
			vgid = gidout.gid[0];
			g_free(gidout.gid);
			goto EXIT;
		}
	}
	*/

	//gsub(vert)
	if (gsub_table) {
		gidin->gid[0] = gid;
		gsub_table_get_gid(gsub_table, tag_vert, gidin, &gidout);
		if (gidout.size > 0) {
			vgid = gidout.gid[0];
			g_free(gidout.gid);
			goto EXIT;
		}
	}
	
	//mort
	if (mort_table) {
		int i, k;

		for (i = 0; i < mort_table->chain_num; i++) {
			//TODO: Check FeatureType
			//for (k = 0; k < mort_table->chains[i].feature_num-1; k++) {}
			
			for (k = 0; k < mort_table->chains[i].subtable_num; k++) {
				mort_Subtable *msub = mort_table->chains[i].subtables + k;

				if (MORT_SUBTABLE_GET_TYPE( msub ) != 
						MORT_SUBTABLE_TYPE_NONCONTEXTUAL) continue;
				if (MORT_SUBTABLE_IS_VERTICAL( msub ) == FALSE) continue;

				vgid = mort_table_find_vertical_glyph_id_from_lookup (
					msub->data, (guint16)gid);
				if (vgid) goto EXIT;
			}
		}

	}

EXIT:
	if (vgid) {
		g_hash_table_insert(vert_table,
				GUINT_TO_POINTER(gid),
				GUINT_TO_POINTER(vgid));
		return vgid;
	} else {
		g_hash_table_insert(vert_table,
				GUINT_TO_POINTER(gid),
				NULL);
		return 0;
	}
}

/* vim: set fdm=marker ts=4 sw=4:*/
