/*
 * GIMP Tategaki Plug-in -- This is a plug-in for The GIMP 2
 * Copyright (C) 2006-2017 OHTSUKA, Yoshio <ohtsuka.yoshio@gmail.com>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 */
#include <glib.h>
#include <glib/gstdio.h>
#include <gdk/gdkkeysyms.h>
#include <libgimp/gimp.h>
#include <libgimp/gimpui.h>
#include <libgimpwidgets/gimpwidgets.h>

#include "tategaki.h"
#include "tategaki_intl.h"
#include "tategaki_image.h"
#include "tategaki_splash.h"
#include "tategaki_msgbox.h"
#include "tategaki_stub_func.h"

#include "color.h"
#include "font_list.h"
#include "typesetter.h"

enum {
	ERROR_NOTHING=0, 
	ERROR_CANCEL,
	ERROR_UNKNOWN_ERROR, 
	ERROR_INVALID_FONT_LIST_ERROR,
	ERROR_INVALID_FONT_ERROR,
	ERROR_INVALID_TEXT_ERROR,
	ERROR_TOO_LARGE_IMAGE_SIZE_ERROR,
	ERROR_TOO_SMALL_IMAGE_SIZE_ERROR,
	ERROR_FAILED_CREATE_LAYER_ERROR,
	ERROR_RENDERING_ERROR,
	ERROR_UNKNOWN_IMAGE_TYPE_ERROR
} PluginError;

typedef enum _PluginRunMode {
	RUNMODE_ADD=0, //mormal mode.
	RUNMODE_MODIFY //modify mode when Modify button is pushed.
} PluginRunMode;

/*static void init    (void);*/
/*static void quit    (void);*/
static void query	(void);
static void run		(const gchar		*name,
					 gint				nparams,
					 const GimpParam		*param,
					 gint				*nreturn_vals,
					 GimpParam			**return_vals);

GimpPlugInInfo PLUG_IN_INFO = {
	NULL, /* init */
	NULL, /* quit */
	query,
	run
};

#define RCFILE      "tategakirc"
#define TATEGAKI_RESPONSE_CLOSE   0
#define TATEGAKI_RESPONSE_CREATE  1
#define TATEGAKI_RESPONSE_MODIFY  2
#define TATEGAKI_RESPONSE_TOPMOST 3

//static parameter of Dialog
#define MAX_FONT_SIZE 999
#define MIN_FONT_SIZE 1
#define DEF_FONT_SIZE 20

#ifdef _WIN32
#define TEXTVIEW_FONT "MS Gothic 12"
#else
#define TEXTVIEW_FONT NULL
#endif

typedef struct _DialogState {
	int			font_index[FONT_STYLE_COUNT];
	gdouble		font_size[FONT_STYLE_COUNT];
	gboolean	use_fontset[FONT_STYLE_COUNT];
	int x;
	int y;
	int width;
	int height;
	GimpUnit	unit;
	gboolean	topmost;
	gboolean	expanded;
	guint		page_number;
} DialogState;

typedef struct _PluginVals {
	FontStyle			styles[4];
	TypesetterOption	option;
	Typesetter			*typesetter;
	gchar				*fontdirs;
	gchar				*batch_separator;
	gchar				*text;
	int					x;
	int					y;
	gint32				image_ID;
	gint32				layer_ID;
	DialogState			gui;
	PluginRunMode		mode;
	gint32				mode_is_batch_add;
	GimpRGB				fgcolor;
	GimpRGB				bgcolor;
} PluginVals;

typedef struct _PluginValsContainer {
	int id;
	PluginVals *vals;
} PluginValsContainer;

typedef struct _FontWidgetSet {
	char		*name;
	GtkWidget	*use;
	GtkWidget	*label;
	GtkWidget	*font;
	GtkWidget	*size;
	GtkWidget	*unit;
} FontWidgetSet;

static struct {
	//page 0
	GtkWidget	*dialog;
	GtkWidget	*dlg_add;
	GtkWidget	*dlg_close;
	GtkWidget	*dlg_modify;
	GtkWidget	*topmost;
	GtkWidget	*text_view;
	//page 1
	GtkWidget	*font;
	GtkWidget	*size;
	GtkWidget	*unit;
	GtkWidget	*text_dir;
	GtkWidget	*tsume_type;
	GtkWidget	*line_align;
	GtkWidget	*gyou_okuri;
	GtkWidget	*moji_okuri;
	GtkWidget	*kerning;
	GtkWidget	*antialias;
	GtkWidget	*fuchidori;
	//page 2
	FontWidgetSet	font_style[FONT_STYLE_COUNT];
	//page 3
	GtkWidget	*ruby_trans;
	GtkWidget	*ruby_max_trans_rate_label;
	GtkWidget	*ruby_max_trans_rate;
	//page 4
	GtkWidget	*auto_yoko;
	GtkWidget	*auto_yoko_digit;
	GtkWidget	*auto_yoko_digit_label;
	//page 5
	GtkWidget	*dot_skip_yaku;
	GtkWidget	*dot_align;
	//page 6
	GtkWidget	*tool_unicode_combo;
	GtkWidget	*tool_unicode_entry;
	GtkWidget	*tool_unicode_insert;
	GtkWidget	*tool_fontname_combo;
	GtkWidget	*tool_fontname_insert;
	//page 7
	GtkListStore	*preset_store;
	GtkWidget	*preset_tree;
	GtkWidget	*preset_listbox;
	GtkWidget	*preset_frame;
	GtkWidget	*preset_load;
	GtkWidget	*preset_save;
	GtkWidget	*preset_delete;
	GtkWidget	*preset_update;
	//page 8
    GtkWidget	*fdir_patheditor;
} widgets;

// template vals, used when modify mode
GKeyFile *tempvals;

// key name of parameter
const static char PNAME_GROUP[]					= PLUGIN_NAME;
const static char PNAME_VERSION[]				= "version";
const static char PNAME_FONTDIR[]				= "font_directory";
const static char PNAME_BATCH_SEPARATOR[]		= "batch_separator";
const static char PNAME_COLOR[]					= "color";//keep compatibility to old versions
const static char PNAME_FGCOLOR[]				= "fgcolor";
const static char PNAME_BGCOLOR[]				= "bgcolor";
const static char PNAME_TEXT[]					= "text";
const static char PNAME_UNIT[]					= "unit";
const static char PNAME_TEXT_DIR[]				= "text_dir";
const static char PNAME_GYOU_OKURI[]			= "gyou_okuri";
const static char PNAME_MOJI_OKURI[]			= "moji_okuri";
const static char PNAME_KERNING[]				= "kerning";
const static char PNAME_ANTIALIAS[]				= "antialias";
const static char PNAME_FUCHIDORI[]				= "fuchidori";
const static char PNAME_TSUME_TYPE[]			= "tsume_type";
const static char PNAME_LINE_ALIGN[]			= "line_align";
const static char PNAME_RUBY[]					= "ruby"; //keep compatibility to old version. (version1.x does not used this. always TRUE.)
const static char PNAME_RUBY_TRANS[]			= "ruby_trans";
const static char PNAME_RUBY_MAX_TRANS_RATE[]	= "ruby_max_trans_rate";
const static char PNAME_VERT_NUMBER[]			= "vert_number";	   //does not rename for backward comaptibility. now it is auto_yoko
const static char PNAME_VNUM_DIGIT[]			= "vert_number_digit"; //does not rename for backward comaptibility. now it is auto_yoko_digit
const static char PNAME_DOT_SKIP_YAKU[]			= "dot_skip_yaku";
const static char PNAME_DOT_ALIGN[]				= "dot_align";
const static char PNAME_FONT_BASE_USE[]		= "font_base_use";
const static char PNAME_FONT_BASE_PATH[]	= "font_base_path";
const static char PNAME_FONT_BASE_NAME[]	= "font_base_name";
const static char PNAME_FONT_BASE_INDEX[]	= "font_base_index";
const static char PNAME_FONT_BASE_SIZE[]	= "font_base_size";
const static char PNAME_FONT_KANA_USE[]		= "font_kana_use";
const static char PNAME_FONT_KANA_PATH[]	= "font_kana_path";
const static char PNAME_FONT_KANA_NAME[]	= "font_kana_name";
const static char PNAME_FONT_KANA_INDEX[]	= "font_kana_index";
const static char PNAME_FONT_KANA_SIZE[]	= "font_kana_size";
const static char PNAME_FONT_RUBY_USE[]		= "font_ruby_use";
const static char PNAME_FONT_RUBY_PATH[]	= "font_ruby_path";
const static char PNAME_FONT_RUBY_NAME[]	= "font_ruby_name";
const static char PNAME_FONT_RUBY_INDEX[]	= "font_ruby_index";
const static char PNAME_FONT_RUBY_SIZE[]	= "font_ruby_size";
const static char PNAME_FONT_ASCII_USE[]	= "font_ascii_use";
const static char PNAME_FONT_ASCII_PATH[]	= "font_ascii_path";
const static char PNAME_FONT_ASCII_NAME[]	= "font_ascii_name";
const static char PNAME_FONT_ASCII_INDEX[]	= "font_ascii_index";
const static char PNAME_FONT_ASCII_SIZE[]	= "font_ascii_size";

static int run_dialog (PluginVals *vals);
static int run_non_interactive (PluginVals *vals);
static int do_tategaki (PluginVals *vals, gint32 *out_layer_ID);
static char* create_layer_name (char *text);

static void preset_load (gchar *path);
static void preset_save (gchar *path, PluginVals *vals);

static void rc_load (gchar *filename, PluginVals *vals);
static void rc_save (gchar *filename, PluginVals *vals);

static void          parasite_to_widgets      (GimpParasite *parasite);
static GimpParasite* parasite_from_pluginvals (PluginVals *vals);

static GKeyFile* keyfile_from_pluginvals (PluginVals *vals);
static void      keyfile_to_widgets      (GKeyFile *kfile);

static void fontcombo_append(GtkWidget *combo, GList *fontlist);
static void textview_set_text (GtkWidget *textview, char* text);
static void textview_get_text (GtkWidget *textview, char** text);

//Globals
static GList *fontlist = NULL;
static gchar *tategaki_dir = NULL;


/*
 Data Flow of Parameter
                                                                                       
                            +--------------+      +------------+      +-------------+  
 INTERACTIVE -------------->|  GtkWidgets  +<-----+  GKeyFile  +----->|  Parasite   |  
                            +------+-------+      +----------+-+<-----+-------------+  
                                   |                A      A |                         
                                   V                |      | |                         
                            +--------------+        |      | |        +-------------+  
 NON-INTERACTIVE ---------->|  PluginVals  +--------+      | +------->|    Preset   | 
                            +------+-------+               +----------+-------------+  
                                   |                                                   
                                   V                                                   
                            +--------------+                                           
                            | FontRenderer |                                           
                            +--------------+                                           
                                                                                       
 */

// Gimp Plugin Interface
static void
query (void) {
	static GimpParamDef args[] = {
	{ GIMP_PDB_INT32,   "run-mode",     "Interactive=0, non-interactive=1" },
	{ GIMP_PDB_IMAGE,   "image",        "Input image" },
	{ GIMP_PDB_DRAWABLE,"drawable",     "Input drawable" },
	{ GIMP_PDB_INT32,   "dir",          "Text direction LeftToRignt=0, TopToBottom=1" },
	{ GIMP_PDB_INT32,   "x",            "x coordinate" },
	{ GIMP_PDB_INT32,   "y",            "y coordinate" },
	{ GIMP_PDB_STRING,  "text",         "UTF-8 String" },
	{ GIMP_PDB_STRING,  "font_name",    "font family name" },
	{ GIMP_PDB_FLOAT,   "size",         "font size" },
	{ GIMP_PDB_INT32,   "size unit",    "Pixel=0, Point=1" },
	{ GIMP_PDB_INT32,   "pitch",        "line pitch amount 100-200% (default 150)" },
	{ GIMP_PDB_INT32,   "tracking",     "char tracking amount 100-200% (default 100)" },
	{ GIMP_PDB_INT32,   "align",        "TOP=0, BOTTOM=1, CENTER=2, BOTH=3" },
	{ GIMP_PDB_INT32,   "antialias",    "Not Antialias=0, Antialias=1" },
	};
	
	static GimpParamDef return_vals[] = {
	{ GIMP_PDB_LAYER,	"layer",	"The text layer" }
	};
	
	gimp_install_procedure (
		PLUGIN_ID,
		_(PLUGIN_DESCRIPTION),
		_(PLUGIN_DESCRIPTION),
		PLUGIN_AUTHOR,
		PLUGIN_AUTHOR,
		PLUGIN_DATE,
		_("_Tategaki..."),
		"RGB*, GRAY*",
		GIMP_PLUGIN,
		G_N_ELEMENTS (args), G_N_ELEMENTS (return_vals),
		args, return_vals);

	gimp_plugin_menu_register (PLUGIN_ID, "<Image>/Filters/ReddogPlug");
}


static void
run (const gchar	*name,
	gint			nparams,
	const GimpParam	*param,
	gint			*nreturn_vals,
	GimpParam		**return_vals)
{/*{{{*/
	static GimpParam	values[2];
	//GimpDrawable		*drawable;
	gint32				image_ID;
	gint32				layer_ID;
	GimpRunMode			run_mode;
	GimpPDBStatusType	status = GIMP_PDB_SUCCESS;
	
	PluginVals vals;
	
	INIT_I18N();

	*nreturn_vals = 2;
	*return_vals  = values;
	
	run_mode = param[0].data.d_int32;
	image_ID = param[1].data.d_image;
	layer_ID = param[2].data.d_drawable;
	//drawable = gimp_drawable_get (layer_ID);
	
	//return value
	values[0].type          = GIMP_PDB_STATUS;
	values[0].data.d_status = status;
	values[1].type          = GIMP_PDB_LAYER;
	values[1].data.d_int32  = -1;
	
	//Create Typesetter
	vals.typesetter = typesetter_create();

	//init parameter
	vals.styles[FONT_STYLE_BASE].file_path   = NULL;
	vals.styles[FONT_STYLE_BASE].name        = NULL;
	vals.styles[FONT_STYLE_BASE].face_index  = 0;
	vals.styles[FONT_STYLE_BASE].unit        = PIXEL;
	vals.styles[FONT_STYLE_BASE].size_x      = INT32_TO_F26D6(32);
	vals.styles[FONT_STYLE_BASE].size_y      = INT32_TO_F26D6(32);

	vals.option.text_dir            = DIR_TYPE_RTL;
	vals.option.gyou_okuri          = 150;
	vals.option.moji_okuri          = 100;
	vals.option.kerning             = TRUE;
	vals.option.line_align          = TOP;
	vals.option.tsume_type          = BETA;
	vals.option.antialias           = TRUE;
	vals.option.fuchidori           = 0;
	vals.option.auto_yoko           = TRUE;
	vals.option.auto_yoko_digit     = 3;
	vals.option.ruby_trans          = TRUE;
	vals.option.ruby_max_trans_rate = 0.2;
	vals.option.dot_skip_yaku       = TRUE;
	vals.option.dot_align           = FALSE;
	vals.fontdirs                   = NULL;
	vals.batch_separator            = NULL;
	vals.text                       = NULL;
	vals.x                          = 0;
	vals.y                          = 0;
	vals.image_ID                   = image_ID;
	vals.layer_ID                   = layer_ID;
	vals.typesetter->option         = vals.option;

	//save dialog parameter
	vals.gui.font_index[FONT_STYLE_BASE]  = 0;
	vals.gui.font_index[FONT_STYLE_KANA]  = 0;
	vals.gui.font_index[FONT_STYLE_RUBY]  = 0;
	vals.gui.font_index[FONT_STYLE_ASCII] = 0;

	vals.gui.unit = GIMP_UNIT_PIXEL;

	vals.gui.font_size[FONT_STYLE_BASE]  = 20;
	vals.gui.font_size[FONT_STYLE_KANA]  = 20;
	vals.gui.font_size[FONT_STYLE_RUBY]  = 10;
	vals.gui.font_size[FONT_STYLE_ASCII] = 20;

	vals.gui.use_fontset[FONT_STYLE_BASE]  = 1;
	vals.gui.use_fontset[FONT_STYLE_KANA]  = 0;
	vals.gui.use_fontset[FONT_STYLE_RUBY]  = 0;
	vals.gui.use_fontset[FONT_STYLE_ASCII] = 0;

	vals.gui.x           = 0;
	vals.gui.y           = 0;
	vals.gui.width       = -1;
	vals.gui.height      = -1;
	vals.gui.topmost     = FALSE;
	vals.gui.expanded    = TRUE;
	vals.gui.page_number = 0;
	
	// init run mode (default is add mode)
	vals.mode = RUNMODE_ADD;
	vals.mode_is_batch_add = 0;

	switch (run_mode) {
		//interactive mode
		case GIMP_RUN_INTERACTIVE:
		case GIMP_RUN_WITH_LAST_VALS:
		{
			// restore dialog parameters
			PluginVals saved_vals;
			if (gimp_get_data (PLUGIN_ID, &saved_vals)) {
				Typesetter *p_typesetter = vals.typesetter;
				vals = saved_vals;
				vals.image_ID                           = image_ID;
				vals.layer_ID                           = layer_ID;
				//reset these parameter
				vals.styles[FONT_STYLE_BASE].file_path  = NULL;
				vals.fontdirs                           = NULL;
				vals.batch_separator                    = NULL;
				vals.text                               = NULL;
				vals.typesetter                         = p_typesetter;
				vals.typesetter->option                 = vals.option;

				vals.mode = RUNMODE_ADD;
			}

			//load font dirs
			rc_load(RCFILE, &vals);

			//run dialog and enter message loops
			if (ERROR_NOTHING != run_dialog(&vals)) {
				WNGMSG ("Error: Failed to get params from GUI interface");
				status = GIMP_PDB_CALLING_ERROR;
				goto error;
			}

			//save font dirs
			rc_save(RCFILE, &vals);

			if (vals.text) {
				g_free(vals.text);
				vals.text = NULL;
			}

			if (vals.fontdirs) {
				g_free(vals.fontdirs);
				vals.fontdirs = NULL;
			}

			if (vals.batch_separator) {
				g_free(vals.batch_separator);
				vals.batch_separator = NULL;
			}

			gimp_set_data (PLUGIN_ID, &vals, sizeof(vals));
			values[0].data.d_status = GIMP_PDB_SUCCESS;
			values[1].data.d_int32  = -1;
		}
		break;

		// non-interactive mode
		case GIMP_RUN_NONINTERACTIVE:
			if (nparams != 14) {
				WNGMSG (_("error: require 10 args."));
				status = GIMP_PDB_CALLING_ERROR;
				goto error;
			}

			vals.option.text_dir                = param[3].data.d_int32;
			vals.x                              = param[4].data.d_int32;
			vals.y                              = param[5].data.d_int32;
			vals.text                           = param[6].data.d_string;
			vals.styles[FONT_STYLE_BASE].name   = param[7].data.d_string;
			vals.styles[FONT_STYLE_BASE].size_x = DOUBLE_TO_F26D6(param[8].data.d_float);
			vals.styles[FONT_STYLE_BASE].size_y = DOUBLE_TO_F26D6(param[8].data.d_float);
			vals.styles[FONT_STYLE_BASE].unit   = param[9].data.d_int32;
			vals.option.gyou_okuri              = param[10].data.d_int32;
			vals.option.moji_okuri              = param[11].data.d_int32;
			vals.option.line_align              = param[12].data.d_int32;
			vals.option.antialias               = param[13].data.d_int32;

			if (vals.styles[FONT_STYLE_BASE].size_x < 1) {
				WNGMSG (_("error: invalid parameter."));
				status = GIMP_PDB_CALLING_ERROR;
				goto error;
			}
			if (vals.option.gyou_okuri < 100) {
				vals.option.gyou_okuri = 100;
			}
			if (vals.option.moji_okuri < 100) {
				vals.option.moji_okuri = 100;
			}
			if (vals.option.line_align < 0 || vals.option.line_align > 3) {
				vals.option.line_align = 0;
			}

			//load font dirs
			rc_load(RCFILE, &vals);

			values[1].data.d_int32	= run_non_interactive (&vals);
			if (values[1].data.d_int32 == -1) 
				values[0].data.d_status = GIMP_PDB_CALLING_ERROR;
			else
				values[0].data.d_status = GIMP_PDB_SUCCESS;
			break;
	}

	//gimp_drawable_detach(drawable);
	return;

error:
	//gimp_drawable_detach(drawable);
	values[0].data.d_status = status;
	values[1].data.d_int32	= -1;
	return;
}/*}}}*/

static gchar *get_tategaki_dir ()
{/*{{{*/
	gchar *ret        = NULL;
	gchar *dir        = NULL;
	gchar *parent_dir = NULL;
	gchar *gimprc     = gimp_personal_rc_file ("gimprc");
	if (gimprc == NULL) goto done;

	parent_dir = g_path_get_dirname (gimprc);
	dir = g_build_filename (parent_dir, "tategaki", NULL);
	
	if (g_file_test(dir, G_FILE_TEST_EXISTS) == FALSE) {
		if (g_mkdir (dir, 0755) != 0) {
			if (dir) g_free(dir);
			goto done;
		}
	} else if (g_file_test(dir, G_FILE_TEST_IS_DIR) == FALSE) {
		if (dir) g_free(dir);
		goto done;
	}
	ret = dir;
done:
	if (parent_dir) g_free(parent_dir);
	if (gimprc) g_free(gimprc);
	return ret;
}/*}}}*/

static void get_fontinfo_from_widget (int   style_id,
									  gchar **path,
									  gchar **name,
									  int   *face_index)
{/*{{{*/
	FontInfo *finfo = NULL;
	GtkWidget *combo;
	int selected;

	if (fontlist == NULL) goto error;
	combo = widgets.font_style[style_id].font;

	gimp_int_combo_box_get_active (
			GIMP_INT_COMBO_BOX(combo), &selected);
	
	fontlist = g_list_first(fontlist);
	if ( 0 <= selected) {
		finfo = g_list_nth_data(fontlist, selected);
		if (finfo != NULL) {
			*path = finfo->path;
			*name = finfo->name;
			*face_index= finfo->face_index;
			return;
		}
	}
error:
	*path = NULL;
	*face_index = 0;
}/*}}}*/

static int fontcombo_select (int   style_id,
							 gchar *path,
							 gchar *name,
							 int   face_index)
{/*{{{*/
	int n;
	FontInfo *finfo = NULL;
	GtkWidget *combo;
	GList *list;

	combo = widgets.font_style[style_id].font;
	n = 0;
	list = g_list_first(fontlist);
	while (list) {
		finfo = list->data;
		if (finfo->face_index == face_index && 
			(g_strcmp0(finfo->name, name) == 0 || g_strcmp0(finfo->path, path) == 0)) {
			gimp_int_combo_box_set_active (GIMP_INT_COMBO_BOX(combo), n);
			return 0;
		}
		n++;
		list = g_list_next(list);
	}
	return 0;
}/*}}}*/

static FT_F26Dot6 compute_size_with_unit (double   size,
										  GimpUnit unit)
{/*{{{*/
	switch (unit) {
		case GIMP_UNIT_PIXEL:
		case GIMP_UNIT_POINT:
			return DOUBLE_TO_F26D6(size);
		case GIMP_UNIT_INCH:
		case GIMP_UNIT_MM:
		case GIMP_UNIT_PICA:
		default:
		{
			double s = size  * 72.0 / gimp_unit_get_factor (unit);
			return DOUBLE_TO_F26D6(s);
			break;
		}
	}
}/*}}}*/

static void textview_set_text (GtkWidget *textview, gchar* text)
{/*{{{*/
	GtkTextBuffer *textbuf;
	textbuf = gtk_text_view_get_buffer (GTK_TEXT_VIEW(textview));
	gtk_text_buffer_set_text (textbuf, text, -1);
}/*}}}*/
static void textview_get_text (GtkWidget *textview, gchar** text)
{/*{{{*/
	GtkTextBuffer *textbuf;
	GtkTextIter start, end;
	textbuf = gtk_text_view_get_buffer (GTK_TEXT_VIEW(textview));
	gtk_text_buffer_get_start_iter (textbuf, &start);
	gtk_text_buffer_get_end_iter (textbuf, &end);
	*text = gtk_text_buffer_get_text (textbuf, &start, &end, FALSE);
	return;
}/*}}}*/

static gboolean key_press_callback (GtkWidget   *widget,
									GdkEventKey *event,
									PluginVals  *vals)
{/*{{{*/
	if (event->keyval == GDK_KEY_Shift_L ||
		event->keyval == GDK_KEY_Shift_R) {
		if (vals->mode == RUNMODE_ADD) {
			gtk_button_set_label(GTK_BUTTON(widgets.dlg_add), _("Batch Create"));
			vals->mode_is_batch_add++;
		}
	}
	return FALSE;
}/*}}}*/

static gboolean key_release_callback (GtkWidget     *widget,
									  GdkEventKey   *event,
									  PluginVals    *vals)
{/*{{{*/
	if (event->keyval == GDK_KEY_Shift_L ||
		event->keyval == GDK_KEY_Shift_R) {
		switch (vals->mode) {
			case RUNMODE_ADD:
				gtk_button_set_label(GTK_BUTTON(widgets.dlg_add), _("Create"));
				break;
			case RUNMODE_MODIFY:
				gtk_button_set_label(GTK_BUTTON(widgets.dlg_add), _("Apply"));
				break;
		}
		vals->mode_is_batch_add--;
	}
	return FALSE;
}/*}}}*/

static void response_callback (GtkWidget    *dialog,
							   gint          response_id,
							   PluginVals   *vals)
{/*{{{*/
	GtkWidget *add_button;
	GtkWidget *close_button;
	GtkWidget *modify_button;
	GdkColor modify_mode_color;
	GdkColor normal_mode_color;

	add_button = widgets.dlg_add;
	close_button = widgets.dlg_close;
	modify_button = widgets.dlg_modify;

	modify_mode_color.red = 32336;
	modify_mode_color.green = 40046;
	modify_mode_color.blue = 58036;

	normal_mode_color.red = 65535;
	normal_mode_color.green = 65535;
	normal_mode_color.blue = 65535;

	switch (response_id) {
		case TATEGAKI_RESPONSE_CLOSE:
		{
			switch (vals->mode) {
				case  RUNMODE_MODIFY:
				{
					// restore parameter
					if (tempvals) {
						keyfile_to_widgets(tempvals);
						if (tempvals) {
							g_key_file_free(tempvals);
							tempvals = NULL;
						}
					}

					vals->mode = RUNMODE_ADD;
					gtk_button_set_label(GTK_BUTTON(add_button), _("Create"));
					gtk_button_set_label(GTK_BUTTON(close_button), _("Close"));
					gtk_widget_show(modify_button);
					textview_set_text(widgets.text_view, "");
					gtk_widget_modify_base(widgets.text_view, GTK_STATE_NORMAL, &normal_mode_color);
					break;
				}
				case RUNMODE_ADD:
				{
					gtk_window_get_size (GTK_WINDOW(dialog), 
							&vals->gui.width, &vals->gui.height);
					gtk_window_get_position (GTK_WINDOW(dialog),
							&vals->gui.x, &vals->gui.y);
					gtk_widget_destroy (dialog);
					break;
				}
				default:
					break;
			}
			break;
		}

		case TATEGAKI_RESPONSE_CREATE:
		{
			gimp_context_get_foreground(&vals->fgcolor);
			gimp_context_get_background(&vals->bgcolor);
			textview_get_text(widgets.text_view, &vals->text);

			switch (vals->mode) {
				case RUNMODE_MODIFY:
				{
					// do_tategaki as modify mode
					gint32 out_layer_ID;
					do_tategaki (vals, &out_layer_ID);
					// restore parameter
					if (vals->text) {
						g_free(vals->text);
						vals->text = NULL;
					}
					if (tempvals) {
						keyfile_to_widgets(tempvals);
						g_key_file_free(tempvals);
						tempvals = NULL;
					}

					vals->mode = RUNMODE_ADD;
					gtk_button_set_label(GTK_BUTTON(add_button), _("Create"));
					gtk_button_set_label(GTK_BUTTON(close_button), _("Close"));
					gtk_widget_show(modify_button);
					textview_set_text(widgets.text_view, "");
					gtk_widget_modify_base(widgets.text_view, GTK_STATE_NORMAL, &normal_mode_color);
					break;
				}
				case RUNMODE_ADD: 
				{
					if (vals->mode_is_batch_add > 0) {
						//Batch Create
						int i;
						char* text = vals->text;
						char** texts = NULL;
						char * separator = NULL;
						separator = g_strconcat("\n", vals->batch_separator, "\n", NULL);
						texts = g_strsplit(vals->text, separator, 0);

						i = 0;
						while(texts[i]) {
							gint32 out_layer_ID;
							vals->text = g_strstrip(texts[i]);
							do_tategaki (vals, &out_layer_ID);
							i++;
						}

						if (texts) g_strfreev(texts);
						if (separator) g_free(separator);
						vals->text = text;
					} else {
						//Create
						gint32 out_layer_ID;
						do_tategaki (vals, &out_layer_ID);
					}
					textview_set_text(widgets.text_view, "");
					gtk_widget_modify_base(widgets.text_view, GTK_STATE_NORMAL, &normal_mode_color);
					break;
				}
				default:
					break;
			}
			break;
		}

		case TATEGAKI_RESPONSE_MODIFY:
		{
			switch (vals->mode) {
				case RUNMODE_ADD:
				{
					gint32 active_layer;
					GimpParasite *parasite;

					// check if this layer has parasite
					active_layer = gimp_image_get_active_layer(vals->image_ID);
					parasite = gimp_drawable_parasite_find (active_layer, PLUGIN_ID);

					if (parasite != NULL) {
						// save current parameters as temp store
						gimp_context_get_foreground(&vals->fgcolor);
						gimp_context_get_background(&vals->bgcolor);
						if (tempvals) {
							g_key_file_free(tempvals);
							tempvals = NULL;
						}
						tempvals = keyfile_from_pluginvals(vals);
						
						//load parameters from parasite which attached on layer
						parasite_to_widgets(parasite);
						gimp_parasite_free(parasite);
						
						gtk_button_set_label(GTK_BUTTON(add_button), _("Apply"));
						gtk_button_set_label(GTK_BUTTON(close_button), _("Cancel"));
						gtk_widget_hide(modify_button);
						gtk_widget_modify_base(widgets.text_view, GTK_STATE_NORMAL, &modify_mode_color);
						vals->layer_ID = active_layer;
						vals->mode = RUNMODE_MODIFY;
					} else {
						msgbox(GTK_MESSAGE_ERROR, GTK_BUTTONS_CLOSE, NULL,
							_("Active layer is not a Tategaki Layer!"));
					}
					break;
				}
				case RUNMODE_MODIFY:
				{
					// invisible when modify mode
					break;
				}
				default:
					break;
			}
		}

		case TATEGAKI_RESPONSE_TOPMOST:
		{
			static gboolean topmost = FALSE;
			topmost = gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(widgets.topmost));
			gtk_window_set_keep_above (GTK_WINDOW(dialog), topmost);
			vals->gui.topmost = topmost;
		}
	}
}/*}}}*/
static void expander_callback (GtkExpander  *expander,
							   GParamSpec   *param, 
							   PluginVals   *vals)
{/*{{{*/
	vals->gui.expanded = gtk_expander_get_expanded (expander);
	if (vals->gui.expanded) {
		gtk_expander_set_label (expander, _("Hide Options..."));
	} else {
		gtk_expander_set_label (expander, _("Show Options..."));
	}
}/*}}}*/
static void notebook_callback (GtkNotebook      *notebook, 
							   GtkNotebookPage  *page,
							   guint             page_num,
							   PluginVals       *vals)
{/*{{{*/
	vals->gui.page_number = page_num;
	DBGMSG("tategaki gui: selected notebook page %d\n", vals->gui.page_number);
}/*}}}*/
static void antialias_toggled (GtkWidget    *widget,
							   PluginVals   *vals)
{/*{{{*/
	vals->option.antialias = 
		gtk_toggle_button_get_active (GTK_TOGGLE_BUTTON(widget));
	DBGMSG("tategaki gui: change antialias:%d\n", vals->option.antialias);
}/*}}}*/
static void fuchidori_changed (GtkWidget    *widget,
							   PluginVals   *vals)
{/*{{{*/
	vals->option.fuchidori = gtk_spin_button_get_value_as_int(
			GTK_SPIN_BUTTON(widget));
	DBGMSG("tategaki gui: change fuchidori:%d\n", vals->option.fuchidori);
}/*}}}*/
static void moji_okuri_changed (GtkWidget   *widget,
							   PluginVals   *vals)
{/*{{{*/
	vals->option.moji_okuri = gtk_spin_button_get_value_as_int(GTK_SPIN_BUTTON(widget));
	DBGMSG("tategaki gui: change moji_okuri:%d\n", vals->option.moji_okuri);
}/*}}}*/
static void kerning_toggled (GtkWidget  *widget,
							 PluginVals *vals)
{/*{{{*/
	vals->option.kerning = gtk_toggle_button_get_active (GTK_TOGGLE_BUTTON(widget));
	DBGMSG("tategaki gui: change kerning:%d\n", vals->option.antialias);
}/*}}}*/

static void gyou_okuri_changed (GtkWidget   *widget,
								PluginVals  *vals)
{/*{{{*/
	vals->option.gyou_okuri = gtk_spin_button_get_value_as_int(
			GTK_SPIN_BUTTON(widget));
	DBGMSG("tategaki gui: change gyou_okuri:%d\n", vals->option.gyou_okuri);
}/*}}}*/
static void size_changed (GtkWidget             *widget,
						  PluginValsContainer   *container)
{/*{{{*/
	int id = container->id;
	PluginVals	*vals = container->vals;

	vals->gui.font_size[id] = gtk_spin_button_get_value (GTK_SPIN_BUTTON(widget));
	vals->styles[id].size_x =
	vals->styles[id].size_y =
		compute_size_with_unit (vals->gui.font_size[id], vals->gui.unit);
	DBGMSG("tategaki gui: id=%d, change size:%f, optsize:%f\n", 
			id, vals->gui.font_size[id],
			F26D6_TO_DOUBLE(vals->styles[id].size_x));

}/*}}}*/
static void unit_changed (GtkWidget     *widget,
						  PluginVals    *vals)
{/*{{{*/
	int i;
	gimp_int_combo_box_get_active (
			GIMP_INT_COMBO_BOX(widget), (int*)&vals->gui.unit);
	for (i=0; i<FONT_STYLE_COUNT; i++) {
		vals->styles[i].unit = (vals->gui.unit == GIMP_UNIT_PIXEL) ? PIXEL : POINT;
		vals->styles[i].size_x = vals->styles[i].size_y =
		compute_size_with_unit (vals->gui.font_size[i], vals->gui.unit);

		// rewrite labels
		if (i != 0) {
			if (widgets.font_style[i].unit)
				gtk_label_set_text (GTK_LABEL(widgets.font_style[i].unit), 
				gimp_unit_get_plural (vals->gui.unit) );
		}
	}
	DBGMSG("tategaki gui: change unit:%d, optunit:%d optsize:%f\n", 
			vals->gui.unit, vals->styles[FONT_STYLE_BASE].unit, 
			F26D6_TO_DOUBLE(vals->styles[FONT_STYLE_BASE].size_x));
}/*}}}*/
static void fontcombo_changed (GtkWidget            *widget,
							   PluginValsContainer  *container)
{/*{{{*/
	PluginVals *vals = container->vals;
	int id = container->id;
	FontInfo *finfo = NULL;
	int selected = -1;
	int listlen;
	gimp_int_combo_box_get_active (
			GIMP_INT_COMBO_BOX(widget), &selected);
	
	fontlist = g_list_first(fontlist);
	listlen = g_list_length(fontlist);

	if ( 0 <= selected  && selected < listlen ) {
		finfo = g_list_nth_data(fontlist, selected);
		if (finfo != NULL) {
			vals->styles[id].file_path  = finfo->path;
			vals->styles[id].name       = finfo->name;
			vals->styles[id].face_index = finfo->face_index;
			vals->gui.font_index[id]    = selected;
		}
	}
	DBGMSG("tategaki gui: id=%d change font:%d %s\n", id, selected, finfo?finfo->path:"error");
}/*}}}*/
static void textbuf_changed (GtkWidget   *widget,
							PluginVals  *vals)
{/*{{{*/
	GtkTextIter st_text_iter, ed_text_iter;
	if (vals->text) {
		g_free(vals->text);
		vals->text = NULL;
	}
	gtk_text_buffer_get_start_iter (
			GTK_TEXT_BUFFER(widget), &st_text_iter);
	gtk_text_buffer_get_end_iter (
			GTK_TEXT_BUFFER(widget), &ed_text_iter);
	vals->text = gtk_text_buffer_get_text (
			GTK_TEXT_BUFFER(widget),
			&st_text_iter,
			&ed_text_iter,
			FALSE);
	DBGMSG("tategaki gui: change text:%s\n", vals->text);
}/*}}}*/
static void text_dir_changed (GtkWidget    *widget,
							  PluginVals   *vals)
{/*{{{*/
	int selected;
	gimp_int_combo_box_get_active (
			GIMP_INT_COMBO_BOX(widget), &selected);
	vals->option.text_dir = selected;
	DBGMSG("tategaki gui: change text_dir:%d\n", selected);
}/*}}}*/
static void line_align_changed (GtkWidget    *widget,
								PluginVals   *vals)
{/*{{{*/
	int selected;
	gimp_int_combo_box_get_active (
			GIMP_INT_COMBO_BOX(widget), &selected);
	vals->option.line_align = selected;
	DBGMSG("tategaki gui: change line_align:%d\n", selected);
}/*}}}*/
static void tsume_type_changed (GtkWidget    *widget,
								PluginVals   *vals)
{/*{{{*/
	int selected;
	gimp_int_combo_box_get_active (
			GIMP_INT_COMBO_BOX(widget), &selected);
	vals->option.tsume_type = selected;
	DBGMSG("tategaki gui: change tsume_type:%d\n", selected);
}/*}}}*/
static void fontcombo_append (GtkWidget    *combo,
							  GList        *fontlist)
{/*{{{*/
	int i;
	GList *tmplist;

	// Count item number
	i = gtk_tree_model_iter_n_children (
			gtk_combo_box_get_model(GTK_COMBO_BOX (combo)), NULL);

	if (i > 0) {
		GtkTreeIter iter;
		gboolean result;

		result = gtk_tree_model_iter_nth_child ( 
			gtk_combo_box_get_model(GTK_COMBO_BOX (combo)), &iter, NULL, i-1);

		if (result == FALSE) {
			i = 0;
		} else {
			int n;
			gchar *c;
			gtk_tree_model_get (gtk_combo_box_get_model(GTK_COMBO_BOX (combo)), 
								&iter,
								GIMP_INT_STORE_VALUE, &n,
								GIMP_INT_STORE_LABEL, &c,
								-1);
			i = n+1;
		}
	}

	DBGMSG ("append_fonts_to_list : st i=%d\n", i);

	tmplist = fontlist;
	while (tmplist) {
		FontInfo *finfo = tmplist->data;
		gimp_int_combo_box_append (GIMP_INT_COMBO_BOX(combo),
								   GIMP_INT_STORE_VALUE, i++,
								   GIMP_INT_STORE_LABEL, finfo->name,
								   -1);
		tmplist = g_list_next(tmplist);
	}
	DBGMSG ("append_fonts_to_list : ed i=%d\n", i);
	return;
}/*}}}*/
static GtkWidget *fontcombo_create(GList *fontlist)
{/*{{{*/
	GtkWidget *combo;
	combo = gimp_int_combo_box_new (NULL, 0, NULL);
	gtk_widget_set_usize(combo, 180, -1);
	fontcombo_append(combo, fontlist);
	return combo;
}/*}}}*/
static void font_widget_set_toggled (GtkWidget	            *widget,
									 PluginValsContainer    *container)
{/*{{{*/
	int id = container->id;
	PluginVals	*vals = container->vals;
	FontWidgetSet fset = widgets.font_style[id];
	
	vals->gui.use_fontset[id] = 
		gtk_toggle_button_get_active (GTK_TOGGLE_BUTTON(widget));

	gtk_widget_set_sensitive (fset.font, vals->gui.use_fontset[id]);
	gtk_widget_set_sensitive (fset.size, vals->gui.use_fontset[id]);
	gtk_widget_set_sensitive (fset.unit, vals->gui.use_fontset[id]);

	DBGMSG("tategaki gui: toggle font set id=%d, val=%d\n", id, vals->gui.use_fontset[id]);
}/*}}}*/
static void ruby_max_trans_rate_changed (GtkWidget  *widget,
										 PluginVals *vals)
{/*{{{*/
	GtkWidget *ruby_max_trans_rate = widgets.ruby_max_trans_rate;
	vals->option.ruby_max_trans_rate = gtk_spin_button_get_value (
			GTK_SPIN_BUTTON(ruby_max_trans_rate));
	DBGMSG("tategaki gui: change ruby max trans rate:%.4ff\n", vals->option.ruby_max_trans_rate);
}/*}}}*/
static void ruby_trans_toggled (GtkWidget   *widget,
								PluginVals  *vals)
{/*{{{*/
	GtkWidget *ruby_trans = widget;
	GtkWidget *ruby_max_trans_rate = widgets.ruby_max_trans_rate;
	GtkWidget *ruby_max_trans_rate_label = widgets.ruby_max_trans_rate_label;
	gboolean max_rate_sensitive = TRUE;

	vals->option.ruby_trans = 
		gtk_toggle_button_get_active (GTK_TOGGLE_BUTTON(ruby_trans));

	max_rate_sensitive =  (GTK_WIDGET_IS_SENSITIVE(ruby_trans) && vals->option.ruby_trans);
	gtk_widget_set_sensitive (ruby_max_trans_rate, max_rate_sensitive);
	gtk_widget_set_sensitive (ruby_max_trans_rate_label,max_rate_sensitive);
	DBGMSG("tategaki gui: change ruby_trans :%d\n", vals->option.ruby_trans);
}/*}}}*/
static void auto_yoko_toggled (GtkWidget    *widget,
							   PluginVals   *vals)
{/*{{{*/
	GtkWidget *digit = widgets.auto_yoko_digit;
	GtkWidget *digit_label = widgets.auto_yoko_digit_label;

	vals->option.auto_yoko = 
		gtk_toggle_button_get_active (GTK_TOGGLE_BUTTON(widget));

	gtk_widget_set_sensitive (digit, vals->option.auto_yoko);
	gtk_widget_set_sensitive (digit_label, vals->option.auto_yoko);
	DBGMSG("tategaki gui: change auto_yoko:%d\n", vals->option.auto_yoko);
}/*}}}*/
static void auto_yoko_digit_changed (GtkWidget  *widget,
									 PluginVals *vals)
{/*{{{*/
	int n = gtk_spin_button_get_value_as_int(GTK_SPIN_BUTTON(widget));
	if (n < 1) n = 1;
	if (n > 10) n = 10;
	vals->option.auto_yoko_digit = n;
	DBGMSG("tategaki gui: change auto yoko digit:%d\n", vals->option.auto_yoko_digit);
}/*}}}*/
static void dot_skip_yaku_toggled (GtkWidget    *widget,
								   PluginVals   *vals)
{/*{{{*/
	vals->option.dot_skip_yaku = gtk_toggle_button_get_active (GTK_TOGGLE_BUTTON(widget));
	DBGMSG("tategaki gui: change dot_skip_yaku:%d\n", vals->option.dot_skip_yaku);
}/*}}}*/
static void dot_align_toggled (GtkWidget    *widget,
							   PluginVals   *vals)
{/*{{{*/
	vals->option.dot_align = gtk_toggle_button_get_active (GTK_TOGGLE_BUTTON(widget));
	DBGMSG("tategaki gui: change dot_align:%d\n", vals->option.dot_align);
}/*}}}*/
static void tool_unicode_insert_callback (GtkWidget	*widget)
{/*{{{*/
	GtkTextBuffer *textbuf;
	const gchar *text;
	gunichar uc;
	gchar *chars;
	gint32 len;

	text = gtk_entry_get_text(GTK_ENTRY(widgets.tool_unicode_entry));
#ifdef _WIN32
	sscanf_s (text, "%06X", &uc);
#else
	sscanf (text, "%06X", &uc);
#endif

	len = g_unichar_to_utf8 (uc, NULL);
	chars = g_malloc0(len+1);
	g_unichar_to_utf8 (uc, chars);

	textbuf = gtk_text_view_get_buffer (GTK_TEXT_VIEW(widgets.text_view));
	gtk_text_buffer_insert_at_cursor(textbuf, chars, len);

	gtk_entry_select_region(GTK_ENTRY(widgets.tool_unicode_entry), 0, -1);
	gtk_widget_grab_focus (widgets.text_view);

	DBGMSG("tategaki gui: insert unicode:U+%04X(%s)\n", uc, chars);
	g_free(chars);
}/*}}}*/
static void tool_unicode_combo_changed_callback (GtkWidget	*widget)
{/*{{{*/
	int selected;
	gchar *chars = NULL;
	gchar *unicode = NULL;
	gunichar uc;
	gint32 len;
	gimp_int_combo_box_get_active(GIMP_INT_COMBO_BOX(widgets.tool_unicode_combo), &selected);
	uc = (gunichar)selected;

	len = g_unichar_to_utf8 (uc, NULL);
	chars = g_malloc0(len+1);
	g_unichar_to_utf8(uc, chars);
	unicode = g_strdup_printf("%04X", selected);

	gtk_entry_set_text(GTK_ENTRY(widgets.tool_unicode_entry), unicode);
	
	DBGMSG("tategaki gui: select unicode:U+%s(%s)\n", unicode, chars);
	if (chars) g_free(chars);
	if (unicode) g_free(unicode);
}/*}}}*/
static void tool_fontname_insert_callback (GtkWidget	*widget)
{/*{{{*/
	GtkTextBuffer *textbuf;
	int value;
	FontInfo *fontinfo;
	gchar **tokens = NULL;
	gchar *text = NULL;

	gimp_int_combo_box_get_active(GIMP_INT_COMBO_BOX(widgets.tool_fontname_combo), &value);
	fontinfo = g_list_nth_data(fontlist, value);

	tokens = g_strsplit(fontinfo->name, " ", -1);
	text = g_strjoinv("\\ ", tokens);
	
	textbuf = gtk_text_view_get_buffer (GTK_TEXT_VIEW(widgets.text_view));
	gtk_text_buffer_insert_at_cursor(textbuf, text, -1);
	gtk_widget_grab_focus (widgets.text_view);

	DBGMSG("tategaki gui: insert fontname:%s\n", text);
	if (tokens) g_strfreev(tokens);
	if (text) g_free(text);
}/*}}}*/
static void preset_save (gchar      *path,
						 PluginVals *vals)
{/*{{{*/
	FILE *fp;
	GKeyFile *keyfile;
	gsize len;
	gchar *data;

	keyfile = keyfile_from_pluginvals(vals);
	g_key_file_remove_key(keyfile, PNAME_GROUP, PNAME_TEXT, NULL);
	data = g_key_file_to_data(keyfile, &len, NULL);

	//FIXME: fopen waring
	//Because Glib doesn't have g_fclose, we can not use g_fopen.
	//If another C runtime was linked, it makes crash the plug-in.

#ifdef _WIN32
	{
		gunichar2 *mode = NULL;
		gunichar2 *temp = NULL;
		mode = g_utf8_to_utf16("wb", -1, NULL, NULL, NULL);
		temp = g_utf8_to_utf16(path, -1, NULL, NULL, NULL);
		_wfopen_s(&fp, temp, mode);
		if (mode) g_free(mode);
		if (temp) g_free(temp);
	}
#else
	fp = fopen(path, "wb");
#endif
	fprintf(fp, "%s", data);
	fclose(fp);

	if (data) g_free(data);
	if (keyfile) g_key_file_free(keyfile);
}/*}}}*/

static void preset_load (gchar *path)
{/*{{{*/

	GKeyFile *keyfile;
	GError *error;

	keyfile = g_key_file_new();
	if (g_key_file_load_from_file(keyfile, path, G_KEY_FILE_KEEP_COMMENTS, &error) == FALSE) {
		msgbox(GTK_MESSAGE_ERROR, GTK_BUTTONS_CLOSE, NULL, error->message);
		return;
	}
	
	g_key_file_remove_key(keyfile, PNAME_GROUP, "text", NULL);
	keyfile_to_widgets(keyfile);
	g_key_file_free(keyfile);

}/*}}}*/

static void preset_delete_callback (GtkWidget    *dummy,
								   gpointer     data)
{/*{{{*/
	GtkResponseType ret;
	gchar *fname, *path;
	gchar msg[1024];
	gboolean has_sel;
	GtkTreeIter iter;
	GtkTreeModel *model;
	GtkTreeSelection *sel;

	sel = gtk_tree_view_get_selection (
			GTK_TREE_VIEW(widgets.preset_tree));
	has_sel = gtk_tree_selection_get_selected(sel, &model, &iter);
	if (has_sel == FALSE) return;

	gtk_tree_model_get(model, &iter, 0, &fname, -1);

	g_snprintf(msg, sizeof(msg), _("Do you want to delete '%s' ?"), fname);
	ret = msgbox (
			GTK_MESSAGE_QUESTION,
			GTK_BUTTONS_OK_CANCEL, 
			widgets.dialog, msg);
	if (ret != GTK_RESPONSE_OK) return;

	path = g_build_filename (tategaki_dir, fname, NULL);
	g_free(fname);
	if (g_file_test(path, G_FILE_TEST_EXISTS)) {
		if (g_remove(path) == 0)
			gtk_list_store_remove(GTK_LIST_STORE(model), &iter);
	}
	g_free(path);
}/*}}}*/
static void preset_update_callback (GtkWidget   *dummy,
									gpointer     data)
{/*{{{*/
	GDir *dir;
	const gchar *fname;

	gtk_list_store_clear (widgets.preset_store);

	if (tategaki_dir == NULL) {
		WNGMSG(_("Cannot open tategaki directory."));
		return;
	}
	dir = g_dir_open(tategaki_dir, 0, NULL);
	if (dir == NULL) return;

	while ((fname = g_dir_read_name (dir))) {
		gchar *fpath = g_build_filename (tategaki_dir, fname, NULL);

		if (g_file_test(fpath, G_FILE_TEST_IS_DIR) == FALSE) {
			GtkTreeIter iter;
			gtk_list_store_append (widgets.preset_store, &iter);
			gtk_list_store_set (widgets.preset_store, &iter, 0, fname, -1);
		}
		g_free(fpath);
	}
	g_dir_close(dir); 
}/*}}}*/
static void preset_load_callback (GtkButton  *dummy,
								  gpointer    data)
{/*{{{*/
	gchar *fname, *path;
	gboolean has_sel;
	GtkTreeIter iter;
	GtkTreeModel *model;
	GtkTreeSelection *sel;

	tategaki_dir = get_tategaki_dir();
	if (tategaki_dir == NULL) return;

	sel = gtk_tree_view_get_selection (GTK_TREE_VIEW(widgets.preset_tree));
	has_sel = gtk_tree_selection_get_selected(sel, &model, &iter);
	if (has_sel == FALSE) return;
	
	gtk_tree_model_get(model, &iter, 0, &fname, -1);
	path = g_build_filename (tategaki_dir, fname, NULL);

	preset_load (path);
	goto done;

done:
	preset_update_callback(NULL, NULL);
	if (fname) g_free(fname);
	if (path) g_free(path);
	return;
}/*}}}*/
static void preset_save_ok_callback (GtkWidget   *dummy,
									 const gchar *name,
									 gpointer     data)
{/*{{{*/
	PluginVals *vals;
	gchar *path = NULL;
	gchar *name_locale = NULL;
	GtkResponseType ret;
	gchar msg[2048];

	vals = (PluginVals*)data;

	tategaki_dir = get_tategaki_dir();
	if (tategaki_dir == NULL) return;

	name_locale = g_filename_from_utf8(name, -1, NULL, NULL, NULL);

	if (name_locale) {
		path = g_build_filename (tategaki_dir, name_locale, NULL);
		g_free(name_locale);
		name_locale = NULL;
	}

	if (g_file_test(path, G_FILE_TEST_EXISTS) == TRUE) {
		g_snprintf(msg, sizeof(msg),
				_("The name '%s' was already exists. Overwrite?"), name);
		ret = msgbox (
				GTK_MESSAGE_QUESTION,
				GTK_BUTTONS_OK_CANCEL, 
				widgets.dialog, msg);
		if (ret != GTK_RESPONSE_OK) goto done;
	}

	preset_save(path, vals);

done:
	preset_update_callback(NULL, NULL);
	if (path) g_free(path);
	return;
}/*}}}*/
static void preset_save_callback (GtkWidget  *dummy,
								 gpointer   data)
{/*{{{*/
	GtkWidget *dialog;
	PluginVals *vals;

	vals = (PluginVals*)data;

	dialog = gimp_query_string_box (_("Tategaki New Parameter Preset"), 
		widgets.dialog, 
		 gimp_standard_help_func, 0, 
		_("Enter a name for the new Tategaki Parameter Preset"), 
		_("Unnamed"), 
		NULL, NULL,
		preset_save_ok_callback, vals);
	gtk_window_set_modal(GTK_WINDOW(dialog), TRUE);
	gtk_window_set_keep_above (GTK_WINDOW (dialog), TRUE);
	gtk_widget_show (dialog);
}/*}}}*/
static void preset_tree_dbclicked (GtkTreeView   *tree_view,
								   GtkTreeIter   *iter,
								   GtkTreePath   *path,
								   gpointer      user_data)
{/*{{{*/
	int i;
	GtkTreeSelection *sel;
	GtkTreeModel *model;
	GtkTreeIter siter;
	GtkTreePath *spath;
	gboolean has_sel;
	
	model = gtk_tree_view_get_model(tree_view);
	sel = gtk_tree_view_get_selection (tree_view);
	has_sel = gtk_tree_selection_get_selected(sel, &model, &siter);
	if (!has_sel) return;
	spath = gtk_tree_model_get_path (model, &siter);

	preset_load_callback (NULL, NULL);
	
	//_ł
	for (i=0; i< 5; i++) {
		if ((i%2) == 0) {
			gtk_tree_selection_select_path (sel, spath);
		} else {
			gtk_tree_selection_unselect_path (sel, spath);
		}
		while (gtk_events_pending ()) {
			gtk_main_iteration ();
		}
		g_usleep(50000);
	}
	gtk_tree_selection_select_path (sel, spath);
    gtk_tree_path_free(spath);
}/*}}}*/
static void fdir_change_path_callback (GimpPathEditor   *editor,
									   PluginVals       *vals)
{/*{{{*/
	gchar *newpath = gimp_path_editor_get_path(editor);

	if (newpath == NULL) {
		if (vals->fontdirs)
			g_free(vals->fontdirs);
		vals->fontdirs = NULL;
	return;
	}

	if (vals->fontdirs == NULL) {
		vals->fontdirs = newpath;
		return;
	}

	if (strcmp(vals->fontdirs, newpath) == 0) {
		g_free(newpath);
	} else {
		g_free(vals->fontdirs);
		vals->fontdirs = newpath;
	}
}/*}}}*/

static int run_dialog (PluginVals	*vals) 
{/*{{{*/
	int i;
	GtkWidget *dlg;
	GtkWidget *topmost;
	GtkWidget *main_vbox;
	GtkExpander *expander;
	GtkNotebook *notebook;

	GtkWidget *basic_vbox;
	GtkWidget *basic_font_hbox;
	GtkWidget *basic_option_table;
	GtkWidget *mixfont_table;
	GtkWidget *ruby_table;
	GtkWidget *vert_num_table;
	GtkWidget *dot_table;
	GtkWidget *tool_table;
	GtkWidget *preset_table;
	GtkWidget *fdir_table;

	GtkWidget		*textframe;
	GtkWidget		*scrolledtext;
	GtkTextBuffer	*textbuf;

	GtkWidget		*tmp;
	int				select_page;
	static PluginValsContainer container[4];

	gimp_ui_init (PLUGIN_NAME, TRUE);
	
	show_splash();

	fontlist = (GList*)font_list_create(gimp_path_parse(vals->fontdirs, 16, FALSE, NULL));
	if (fontlist == NULL) {
		WNGMSG ("Error: No fontlist");
	}
	vals->typesetter->font_list = fontlist;

	for (i=0; i<FONT_STYLE_COUNT; i++) {
		container[i].id   = i;
		container[i].vals = vals;
	}
	select_page = vals->gui.page_number;

	// create dialog
	dlg = gtk_dialog_new ();
	topmost = gtk_check_button_new_with_label (_("Topmost"));
	gtk_window_set_title (GTK_WINDOW(dlg),
						  "Tategaki" " Version " PLUGIN_VERSION);

	gtk_window_set_icon(GTK_WINDOW(dlg), 
	gdk_pixbuf_new_from_inline (-1, img_window_icon, FALSE, NULL));
	gtk_window_set_role (GTK_WINDOW(dlg), PLUGIN_NAME);
	gtk_window_set_resizable (GTK_WINDOW(dlg), TRUE);
	gtk_dialog_set_has_separator (GTK_DIALOG(dlg), FALSE);
#ifdef _WIN32
	widgets.dlg_modify = gtk_dialog_add_button (GTK_DIALOG(dlg), _("Modify"), TATEGAKI_RESPONSE_MODIFY);
	widgets.dlg_add = gtk_dialog_add_button (GTK_DIALOG(dlg), _("Create"), TATEGAKI_RESPONSE_CREATE);
	widgets.dlg_close = gtk_dialog_add_button (GTK_DIALOG(dlg), _("Close"), TATEGAKI_RESPONSE_CLOSE);
#else
	widgets.dlg_close = gtk_dialog_add_button (GTK_DIALOG(dlg), _("Close"), TATEGAKI_RESPONSE_CLOSE);
	widgets.dlg_add = gtk_dialog_add_button (GTK_DIALOG(dlg), _("Create"), TATEGAKI_RESPONSE_CREATE);
	widgets.dlg_modify = gtk_dialog_add_button (GTK_DIALOG(dlg), _("Modify"), TATEGAKI_RESPONSE_MODIFY);
#endif
	gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON(topmost), vals->gui.topmost);

	gtk_dialog_add_action_widget (GTK_DIALOG(dlg), 
								  topmost,
								  TATEGAKI_RESPONSE_TOPMOST);
	gtk_button_box_set_child_secondary (GTK_BUTTON_BOX (GTK_DIALOG(dlg)->action_area), 
										topmost,
										TRUE);
	
	g_signal_connect (dlg, "key-press-event", G_CALLBACK (key_press_callback), vals);
	g_signal_connect (dlg, "key-release-event", G_CALLBACK (key_release_callback), vals);
	g_signal_connect (dlg, "response", G_CALLBACK(response_callback), vals);
	g_signal_connect (dlg, "destroy", G_CALLBACK(gtk_main_quit), NULL);
	
	widgets.dialog = dlg;
	widgets.topmost = topmost;

	// main vbox
	main_vbox = gtk_vbox_new (FALSE, 2);
	gtk_container_set_border_width (GTK_CONTAINER(main_vbox), 
										2);
	gtk_container_add(GTK_CONTAINER( GTK_DIALOG(dlg)->vbox ), 
					  main_vbox);

	// options expander
	expander = GTK_EXPANDER(gtk_expander_new (NULL));
	gtk_box_pack_start(GTK_BOX(main_vbox),
					   GTK_WIDGET(expander),
					   FALSE,
					   TRUE,
					   0);
	g_signal_connect (GTK_WIDGET(expander), "notify::expanded", 
					  G_CALLBACK(expander_callback), vals);
	gtk_expander_set_expanded (expander, vals->gui.expanded);
	expander_callback(expander, NULL, vals);
	
	// notebook
	notebook = GTK_NOTEBOOK(gtk_notebook_new());
	g_signal_connect (notebook, "switch_page",
					  G_CALLBACK(notebook_callback), vals);
	gtk_container_add(GTK_CONTAINER(expander), GTK_WIDGET(notebook));

	// pange 1 basic infomation
	basic_vbox = gtk_vbox_new (FALSE, 2);
	gtk_container_set_border_width (GTK_CONTAINER(basic_vbox), 2);
	gtk_notebook_append_page (notebook, basic_vbox, NULL);
	gtk_notebook_set_tab_label_text (notebook, basic_vbox, _("Basic"));
	// page 1 upper row layout
	basic_font_hbox = gtk_hbox_new (FALSE, 2);
	gtk_box_pack_start( GTK_BOX(basic_vbox), basic_font_hbox, FALSE, TRUE, 0);
	// page 1 under row layout
	basic_option_table = gtk_table_new (3, 8, FALSE);
	gtk_box_pack_start( GTK_BOX(basic_vbox), basic_option_table, FALSE, TRUE, 0);

	// page 2 mixed font
	mixfont_table = gtk_table_new (4, 4, FALSE);
	gtk_container_set_border_width (GTK_CONTAINER(mixfont_table), 2);
	gtk_notebook_append_page (notebook, mixfont_table, NULL);
	gtk_notebook_set_tab_label_text (notebook, mixfont_table, _("Mixed Font"));

	// page 3 ruby
	ruby_table = gtk_table_new (3, 4, FALSE);
	gtk_container_set_border_width (GTK_CONTAINER(ruby_table), 2);
	gtk_notebook_append_page (notebook, ruby_table, NULL);
	gtk_notebook_set_tab_label_text (notebook, ruby_table, _("Ruby"));

	// page 4 vert num
	vert_num_table = gtk_table_new (4, 4, FALSE);
	gtk_container_set_border_width (GTK_CONTAINER(vert_num_table), 2);
	gtk_notebook_append_page (notebook, vert_num_table, NULL);
	gtk_notebook_set_tab_label_text (notebook, vert_num_table, _("Yoko"));

	//page 5 dot
	dot_table = gtk_table_new (2, 1, FALSE);
	gtk_container_set_border_width (GTK_CONTAINER(dot_table), 2);
	gtk_notebook_append_page (notebook, dot_table, NULL);
	gtk_notebook_set_tab_label_text (notebook, dot_table, _("Dot"));

	// page 6 tool 
	tool_table = gtk_table_new(2, 2, FALSE);
	gtk_container_set_border_width (GTK_CONTAINER(tool_table), 2);
	gtk_notebook_append_page (notebook, tool_table, NULL);
	gtk_notebook_set_tab_label_text (notebook, tool_table, _("Tool"));

	// page 7 preset 
	preset_table = gtk_table_new (4, 4, FALSE); 
	gtk_container_set_border_width (GTK_CONTAINER(preset_table), 2);
	gtk_notebook_append_page (notebook, preset_table, NULL);
	gtk_notebook_set_tab_label_text (notebook, preset_table, _("Preset"));

	// page 8 font directory
	fdir_table = gtk_table_new (1, 1, FALSE);
	gtk_container_set_border_width (GTK_CONTAINER(fdir_table), 2);
	gtk_notebook_append_page (notebook, fdir_table, NULL);
	gtk_notebook_set_tab_label_text (notebook, fdir_table, _("FontDir"));

	//
	// text input
	//
/*{{{*/
	// text frame
	textframe = gtk_frame_new (NULL);
	gtk_frame_set_shadow_type(GTK_FRAME(textframe), GTK_SHADOW_IN);
	gtk_box_pack_start(GTK_BOX(main_vbox), textframe, TRUE, TRUE, 0);

	// scrolled text
	scrolledtext = gtk_scrolled_window_new(NULL, NULL);
	gtk_scrolled_window_set_policy (GTK_SCROLLED_WINDOW (scrolledtext),
									GTK_POLICY_AUTOMATIC,
									GTK_POLICY_ALWAYS);
	gtk_container_add(GTK_CONTAINER(textframe), scrolledtext);
	gtk_widget_set_usize(scrolledtext, -1, 100);

	// text view
	widgets.text_view = gtk_text_view_new ();
	gtk_text_view_set_left_margin  (GTK_TEXT_VIEW(widgets.text_view), 2);
	gtk_text_view_set_right_margin (GTK_TEXT_VIEW(widgets.text_view), 2);
	gtk_text_view_set_pixels_above_lines (GTK_TEXT_VIEW(widgets.text_view), 1);
	gtk_text_view_set_pixels_below_lines (GTK_TEXT_VIEW(widgets.text_view), 1);
	textbuf = gtk_text_view_get_buffer (GTK_TEXT_VIEW(widgets.text_view));
	if (vals->text) {
		textview_set_text(widgets.text_view, vals->text);
	}
	gtk_widget_modify_font (GTK_WIDGET(widgets.text_view),
							pango_font_description_from_string(TEXTVIEW_FONT));
	g_signal_connect (textbuf, "changed", G_CALLBACK (textbuf_changed), vals);
	gtk_container_add(GTK_CONTAINER(scrolledtext), widgets.text_view);
	/*}}}*/

	//
	// page 1 basic information
	//
/*{{{*/
	widgets.font = fontcombo_create(fontlist);
	gtk_widget_set_usize(widgets.font, 200, -1);
	gimp_int_combo_box_set_active (GIMP_INT_COMBO_BOX(widgets.font),
								   vals->gui.font_index[FONT_STYLE_BASE]);
	gtk_box_pack_start(GTK_BOX(basic_font_hbox), widgets.font, FALSE, FALSE, 0);
	g_signal_connect (widgets.font, "changed", G_CALLBACK (fontcombo_changed), container);
	fontcombo_changed (widgets.font, container);

	//spacer
	tmp = gimp_frame_new (NULL);
	gtk_box_pack_start(GTK_BOX(basic_font_hbox), tmp, FALSE, FALSE, 0);
	gtk_widget_set_usize(tmp, 10, -1);

	//size
	widgets.size = gtk_spin_button_new_with_range (MIN_FONT_SIZE, MAX_FONT_SIZE, 1);
	gtk_spin_button_configure (GTK_SPIN_BUTTON(widgets.size), NULL, 1, 1);
	gtk_spin_button_set_value (GTK_SPIN_BUTTON(widgets.size),
							   vals->gui.font_size[FONT_STYLE_BASE]);
	g_signal_connect (widgets.size, "value_changed", G_CALLBACK (size_changed), container);
	size_changed (widgets.size, container);
	gtk_box_pack_start (GTK_BOX(basic_font_hbox), widgets.size, FALSE, FALSE, 0);

	widgets.font_style[FONT_STYLE_BASE].font = widgets.font;
	widgets.font_style[FONT_STYLE_BASE].size = widgets.size;

	widgets.unit = gimp_int_combo_box_new (
							gimp_unit_get_plural (GIMP_UNIT_PIXEL), GIMP_UNIT_PIXEL,
							gimp_unit_get_plural (GIMP_UNIT_INCH),  GIMP_UNIT_INCH,
							gimp_unit_get_plural (GIMP_UNIT_MM),    GIMP_UNIT_MM,
							gimp_unit_get_plural (GIMP_UNIT_POINT), GIMP_UNIT_POINT,
							gimp_unit_get_plural (GIMP_UNIT_PICA),  GIMP_UNIT_PICA,
							NULL);
    gimp_int_combo_box_set_active (GIMP_INT_COMBO_BOX(widgets.unit), vals->gui.unit);
	gtk_widget_set_usize(widgets.unit, 90, -1);
	g_signal_connect (widgets.unit, "changed", G_CALLBACK (unit_changed), vals);
	gtk_box_pack_start ( GTK_BOX(basic_font_hbox), widgets.unit, FALSE, FALSE, 0);


	// here under is page 1 under row layout table

	// text dir
	widgets.text_dir = gimp_int_combo_box_new (
							_("Vertical"),   DIR_TYPE_RTL,
							_("Horizontal"), DIR_TYPE_TTB,
							NULL);
	gimp_int_combo_box_set_active (GIMP_INT_COMBO_BOX(widgets.text_dir), vals->option.text_dir);
	gtk_widget_set_usize(widgets.text_dir, 90, -1);
	g_signal_connect (widgets.text_dir, "changed", G_CALLBACK (text_dir_changed), vals);
	text_dir_changed (widgets.text_dir, vals);
	
	// gyou okuri
	widgets.gyou_okuri = gtk_spin_button_new_with_range (100, 500, 1);
	gtk_spin_button_configure (GTK_SPIN_BUTTON(widgets.gyou_okuri), NULL, 1, 0);
	gtk_spin_button_set_value (GTK_SPIN_BUTTON(widgets.gyou_okuri), vals->option.gyou_okuri);
	g_signal_connect (widgets.gyou_okuri, "value_changed", G_CALLBACK (gyou_okuri_changed), vals);

	// moji okuji
	widgets.moji_okuri = gtk_spin_button_new_with_range (80, 500, 1);
	gtk_spin_button_configure (GTK_SPIN_BUTTON(widgets.moji_okuri), NULL, 1, 0);
	gtk_spin_button_set_value (GTK_SPIN_BUTTON(widgets.moji_okuri), vals->option.moji_okuri);
	g_signal_connect (widgets.moji_okuri, "value_changed", G_CALLBACK (moji_okuri_changed), vals);

	//kerning
	widgets.kerning = gtk_check_button_new_with_label (_("Kerning"));
	gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON(widgets.kerning), vals->option.kerning);
	g_signal_connect (widgets.kerning, "toggled", G_CALLBACK (kerning_toggled), vals);

	// yakumono tume type
	widgets.tsume_type = gimp_int_combo_box_new (
							_("BetaGumi"),      BETA,
							_("YakuTsume"),     TSUME,
							_("JISX4051"),      JISX4051,
							NULL);
	gimp_int_combo_box_set_active (GIMP_INT_COMBO_BOX(widgets.tsume_type), vals->option.tsume_type);
	gtk_widget_set_usize(widgets.tsume_type, 90, -1);
	g_signal_connect (widgets.tsume_type, "changed", G_CALLBACK (tsume_type_changed), vals);
	tsume_type_changed (widgets.tsume_type, vals);

	// line align
	widgets.line_align = gimp_int_combo_box_new (
							_("Top Align"),     TOP,
							_("Center Align"),  CENTER,
							_("Bottom Align"),  BOTTOM,
							_("Both Align"),    BOTH,
							NULL);
	gimp_int_combo_box_set_active (GIMP_INT_COMBO_BOX(widgets.line_align), vals->option.line_align);
	gtk_widget_set_usize(widgets.line_align, 90, -1);
	g_signal_connect (widgets.line_align, "changed", G_CALLBACK (line_align_changed), vals);
	line_align_changed (widgets.line_align, vals);

	// antialias
	widgets.antialias = gtk_check_button_new_with_label (_("Antialias"));
	gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON(widgets.antialias), vals->option.antialias);
	g_signal_connect (widgets.antialias, "toggled", G_CALLBACK (antialias_toggled), vals);
	
	// fuchidori
	widgets.fuchidori = gtk_spin_button_new_with_range (0, 50, 1);
	gtk_spin_button_configure (GTK_SPIN_BUTTON(widgets.fuchidori), NULL, 1, 0);
	gtk_spin_button_set_value (GTK_SPIN_BUTTON(widgets.fuchidori), vals->option.fuchidori);
	g_signal_connect (widgets.fuchidori, "value_changed", G_CALLBACK (fuchidori_changed), vals);
	
	// build tables
	tmp = gtk_alignment_new(0.0, 0.5, 0.0, 0.0);
	gtk_container_add(GTK_CONTAINER(tmp), gtk_label_new (_("Text Dir:")));
	gtk_table_attach(GTK_TABLE(basic_option_table), tmp,
					 0, 1, 0, 1, GTK_FILL, GTK_FILL, 2, 2);
	gtk_table_attach(GTK_TABLE(basic_option_table), widgets.text_dir, 
					 1, 2, 0, 1, GTK_FILL, GTK_FILL, 2, 2);

	tmp = gtk_alignment_new(0.0, 0.5, 0.0, 0.0);
	gtk_container_add(GTK_CONTAINER(tmp), gtk_label_new (_("Line Align:")));
	gtk_table_attach(GTK_TABLE(basic_option_table), tmp,
					 0, 1, 1, 2, GTK_FILL, GTK_FILL, 2, 2);
	gtk_table_attach(GTK_TABLE(basic_option_table), widgets.line_align,
					 1, 2, 1, 2, GTK_FILL, GTK_FILL, 2, 2);

	tmp = gtk_alignment_new(0.0, 0.5, 0.0, 0.0);
	gtk_container_add(GTK_CONTAINER(tmp), gtk_label_new (_("Yakumono:")));
	gtk_table_attach(GTK_TABLE(basic_option_table), tmp,
					 0, 1, 2, 3, GTK_FILL, GTK_FILL, 2, 2);
	gtk_table_attach(GTK_TABLE(basic_option_table), widgets.tsume_type,
					 1, 2, 2, 3, GTK_FILL, GTK_FILL, 2, 2);

	gtk_table_attach(GTK_TABLE(basic_option_table), gtk_label_new (_("Line Pitch:")),
					 2, 3, 1, 2, GTK_FILL, GTK_FILL, 2, 2);
	gtk_table_attach(GTK_TABLE(basic_option_table), widgets.gyou_okuri,
					 3, 4, 1, 2, GTK_FILL, GTK_FILL, 2, 2);
	gtk_table_attach(GTK_TABLE(basic_option_table), gtk_label_new ("%"),
					 4, 5, 1, 2, GTK_FILL, GTK_FILL, 0, 2);

	gtk_table_attach(GTK_TABLE(basic_option_table), gtk_label_new (_("Tracking:")),
					 2, 3, 2, 3, GTK_FILL, GTK_FILL, 2, 2);
	gtk_table_attach(GTK_TABLE(basic_option_table), widgets.moji_okuri,
					 3, 4, 2, 3, GTK_FILL, GTK_FILL, 2, 2);
	gtk_table_attach(GTK_TABLE(basic_option_table), gtk_label_new ("%"),
					 4, 5, 2, 3, GTK_FILL, GTK_FILL, 0, 2);


	gtk_table_attach(GTK_TABLE(basic_option_table), widgets.kerning,
					 5, 7, 0, 1, GTK_FILL, GTK_FILL, 2, 2);

	gtk_table_attach(GTK_TABLE(basic_option_table), widgets.antialias,
					 5, 7, 1, 2, GTK_FILL, GTK_FILL, 2, 2);

	gtk_table_attach(GTK_TABLE(basic_option_table), gtk_label_new (_("Edging:")),
					 5, 6, 2, 3, GTK_FILL, GTK_FILL, 2, 2);
	gtk_table_attach(GTK_TABLE(basic_option_table), widgets.fuchidori,
					 6, 7, 2, 3, GTK_FILL, GTK_FILL, 2, 2);
	gtk_table_attach(GTK_TABLE(basic_option_table), gtk_label_new ("%"),
					 7, 8, 2, 3, GTK_FILL, GTK_FILL, 0, 2);


/*}}}*/

	//
	// page 2 mixed font
	//
	{/*{{{*/
		int i;
		FontWidgetSet *fset = widgets.font_style;
		
		fset[FONT_STYLE_KANA].name = _("Kana Font");
		fset[FONT_STYLE_RUBY].name  = _("Ruby Font");
		fset[FONT_STYLE_ASCII].name = _("ASCII Font");
		
		//checkbutton, rabel, font, fontsize
		//don't use i=0
		for (i=1; i< FONT_STYLE_COUNT; i++) {
			//used checkbutton
			fset[i].use = gtk_check_button_new_with_label (fset[i].name);
			gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON(fset[i].use),
										  vals->gui.use_fontset[i]);
			g_signal_connect (fset[i].use, "toggled", G_CALLBACK (font_widget_set_toggled), container + i);

			//font combobox
			fset[i].font = fontcombo_create(fontlist);

			gimp_int_combo_box_set_active (GIMP_INT_COMBO_BOX(fset[i].font), 
										   vals->gui.font_index[i]);
			g_signal_connect (fset[i].font, "changed", G_CALLBACK (fontcombo_changed), container + i);
			fontcombo_changed(fset[i].font , &container[i]);

			//spacer
			tmp = gimp_frame_new (NULL);
			gtk_widget_set_usize(tmp, 10, -1);

			//size
			fset[i].size = gtk_spin_button_new_with_range (MIN_FONT_SIZE, MAX_FONT_SIZE, 1);
			gtk_spin_button_configure (GTK_SPIN_BUTTON(fset[i].size), NULL, 1, 1);
			gtk_spin_button_set_value (GTK_SPIN_BUTTON(fset[i].size), vals->gui.font_size[i]);
			g_signal_connect (fset[i].size, "value_changed", G_CALLBACK (size_changed), container+i);
			size_changed (fset[i].size, container+i);

			//unit
			fset[i].unit = gtk_label_new (NULL);

			gtk_table_attach(GTK_TABLE(mixfont_table), fset[i].use, 
							 0, 1, i, i+1, GTK_FILL, GTK_FILL, 2, 2);
			gtk_table_attach(GTK_TABLE(mixfont_table), fset[i].font,
							 1, 2, i, i+1, GTK_FILL, GTK_FILL, 2, 2);
			gtk_table_attach(GTK_TABLE(mixfont_table), tmp,
							 2, 3, i, i+1, GTK_FILL, GTK_FILL, 2, 2);
			gtk_table_attach(GTK_TABLE(mixfont_table), fset[i].size,
							 4, 5, i, i+1, GTK_FILL, GTK_FILL, 2, 2);
			gtk_table_attach(GTK_TABLE(mixfont_table), fset[i].unit,
							 5, 6, i, i+1, GTK_FILL, GTK_FILL, 2, 2);

			font_widget_set_toggled (fset[i].use, container + i);
		}
	}/*}}}*/

	//
	// page 3 ruby
	//
/*{{{*/
	;
	//ruby transform
	widgets.ruby_trans = gtk_check_button_new_with_label (_("Enable Ruby Transform"));
	gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON(widgets.ruby_trans),
								  vals->option.ruby_trans);
	g_signal_connect (widgets.ruby_trans, "toggled", G_CALLBACK (ruby_trans_toggled), vals);
	
	//ruby transform max rate
	widgets.ruby_max_trans_rate_label = gtk_label_new(_("Ruby Max Transform Rate:"));
	widgets.ruby_max_trans_rate = gtk_spin_button_new_with_range (0.1, 1.0, 0.01);
	gtk_spin_button_configure (GTK_SPIN_BUTTON(widgets.ruby_max_trans_rate), NULL, 1, 3);
	gtk_spin_button_set_value (GTK_SPIN_BUTTON(widgets.ruby_max_trans_rate),
							   vals->option.ruby_max_trans_rate);
	g_signal_connect (widgets.ruby_max_trans_rate, "value_changed", G_CALLBACK (ruby_max_trans_rate_changed), vals);


	ruby_trans_toggled (widgets.ruby_trans, vals);

	gtk_table_attach(GTK_TABLE(ruby_table), widgets.ruby_trans,
					 0, 1, 0, 1, GTK_FILL, GTK_FILL, 2, 2);

	//gtk_table_attach(GTK_TABLE(ruby_table), gtk_label_new("  "),
	//		1, 2, 0, 1, GTK_FILL, GTK_FILL, 2, 2);

	gtk_table_attach(GTK_TABLE(ruby_table), widgets.ruby_max_trans_rate_label,
					 0, 1, 1, 2, GTK_FILL, GTK_FILL, 2, 2);
	gtk_table_attach(GTK_TABLE(ruby_table), widgets.ruby_max_trans_rate,
					 1, 2, 1, 2, GTK_FILL, GTK_FILL, 2, 2);
/*}}}*/

	//
	// page 4 auto yoko(vertival number)
	//
/*{{{*/
	;
	//auto yoko(vertical number)
	widgets.auto_yoko = gtk_check_button_new_with_label(_("Auto Yoko"));
	gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON(widgets.auto_yoko), vals->option.auto_yoko);
	g_signal_connect (widgets.auto_yoko, "toggled", G_CALLBACK (auto_yoko_toggled), vals);

	// sequence number digits
	widgets.auto_yoko_digit = gtk_spin_button_new_with_range (1, 10, 1);
	gtk_spin_button_configure (GTK_SPIN_BUTTON(widgets.auto_yoko_digit), NULL, 1, 0);
	gtk_spin_button_set_value (GTK_SPIN_BUTTON(widgets.auto_yoko_digit), vals->option.auto_yoko_digit);
	g_signal_connect (widgets.auto_yoko_digit, "value_changed", G_CALLBACK (auto_yoko_digit_changed), vals);

	widgets.auto_yoko_digit_label = gtk_label_new (_("Auto Yoko Digits:")),

	auto_yoko_toggled(widgets.auto_yoko, vals);

	gtk_table_attach(GTK_TABLE(vert_num_table), widgets.auto_yoko,
					 0, 2, 0, 1, GTK_FILL, GTK_FILL, 2, 2);

	gtk_table_attach(GTK_TABLE(vert_num_table), widgets.auto_yoko_digit_label,
					 0, 1, 1, 2, GTK_FILL, GTK_FILL, 2, 2);

	gtk_table_attach(GTK_TABLE(vert_num_table), widgets.auto_yoko_digit,
					 1, 2, 1, 2, GTK_FILL, GTK_FILL, 2, 2);
/*}}}*/

	//
	// page 5 dot
	//
/*{{{*/
	{
	widgets.dot_skip_yaku = gtk_check_button_new_with_label(_("Ignore Yakumono"));
	gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON(widgets.dot_skip_yaku),
													vals->option.dot_skip_yaku);
	g_signal_connect (widgets.dot_skip_yaku, "toggled", G_CALLBACK (dot_skip_yaku_toggled), vals);

	widgets.dot_align = gtk_check_button_new_with_label(_("Align Dot"));
	gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON(widgets.dot_align), vals->option.dot_align);
	g_signal_connect (widgets.dot_align, "toggled", G_CALLBACK (dot_align_toggled), vals);

	gtk_table_attach(GTK_TABLE(dot_table), widgets.dot_skip_yaku,
					 0, 1, 0, 1, GTK_FILL, GTK_FILL, 2, 2);

	gtk_table_attach(GTK_TABLE(dot_table), widgets.dot_align,
					 0, 1, 1, 2, GTK_FILL, GTK_FILL, 2, 2);
	}
/*}}}*/

	//
	// page 6 tool
	//
/*{{{*/
	{
		gint w, h;
		GtkWidget *align1 = gtk_hbox_new(FALSE, 2);

		// Unicode Insert
		widgets.tool_unicode_entry = gtk_entry_new_with_max_length(5);
		gtk_widget_get_size_request(widgets.tool_unicode_entry, &w, &h);
		gtk_widget_set_size_request(widgets.tool_unicode_entry, 50, h);
		g_signal_connect(widgets.tool_unicode_entry, "activate", G_CALLBACK(tool_unicode_insert_callback), NULL);

		widgets.tool_unicode_insert = gtk_button_new_with_label(_("Insert"));
		g_signal_connect(widgets.tool_unicode_insert, "clicked", G_CALLBACK(tool_unicode_insert_callback), NULL);

		gtk_box_pack_end(GTK_BOX(align1), widgets.tool_unicode_entry, FALSE, FALSE, 0);
		gtk_box_pack_end(GTK_BOX(align1), gtk_label_new("U+"), FALSE, FALSE, 0);

		// Unicode Selector Combo
		widgets.tool_unicode_combo = gimp_int_combo_box_new ("U+3030(\xE3\x80\xB0) WAVY DASH",      0x3030,
															 "U+2013(\xE2\x80\x93) EN DASH",        0x2013,
															 "U+2014(\xE2\x80\x94) EM DASH",        0x2014,
															 "U+2015(\xE2\x80\x95) HORIZONTAL BAR", 0x2015,
															 "U+E0100 VARIATION SELECTOR-17",       0xE0100,
															 "U+E0101 VARIATION SELECTOR-18",       0xE0101,
															 "U+E0102 VARIATION SELECTOR-19",       0xE0102,
															 "U+E0103 VARIATION SELECTOR-20",       0xE0103,
															 "U+E0104 VARIATION SELECTOR-21",       0xE0104,
															 "U+E0105 VARIATION SELECTOR-22",       0xE0105,
															 "U+E0106 VARIATION SELECTOR-23",       0xE0106,
															 "U+E0107 VARIATION SELECTOR-24",       0xE0107,
															 "U+E0108 VARIATION SELECTOR-25",       0xE0108,
															 "U+E0109 VARIATION SELECTOR-26",       0xE0109,
															 "U+E010A VARIATION SELECTOR-27",       0xE010A,
															 "U+E010B VARIATION SELECTOR-28",       0xE010B,
															 "U+E010C VARIATION SELECTOR-29",       0xE010C,
															 "U+E010D VARIATION SELECTOR-30",       0xE010D,
															 "U+E010E VARIATION SELECTOR-31",       0xE010E,
															 "U+E010F VARIATION SELECTOR-32",       0xE010F,
															 NULL);
		g_signal_connect(widgets.tool_unicode_combo, "changed", G_CALLBACK(tool_unicode_combo_changed_callback), vals);

		gtk_box_pack_end(GTK_BOX(align1), widgets.tool_unicode_combo, TRUE, TRUE, 0);

		//Packing to table
		gtk_table_attach(GTK_TABLE(tool_table), align1,
						 0, 1, 0, 1, GTK_FILL, GTK_FILL, 2, 2);
		gtk_table_attach(GTK_TABLE(tool_table), widgets.tool_unicode_insert,
						 1, 2, 0, 1, GTK_FILL, GTK_FILL, 2, 2);

		// FontName Insert
		widgets.tool_fontname_combo = fontcombo_create(fontlist);
		widgets.tool_fontname_insert = gtk_button_new_with_label(_("Insert"));
		g_signal_connect(widgets.tool_fontname_insert, "clicked", G_CALLBACK(tool_fontname_insert_callback), NULL);

		gtk_table_attach(GTK_TABLE(tool_table), widgets.tool_fontname_combo,
						 0, 1, 1, 2, GTK_FILL, GTK_FILL, 2, 2);

		gtk_table_attach(GTK_TABLE(tool_table), widgets.tool_fontname_insert,
						 1, 2, 1, 2, GTK_FILL, GTK_FILL, 2, 2);
	}
/*}}}*/
	
	//
	// page 7 user defined preset save/load
	//
/*{{{*/
	tategaki_dir = get_tategaki_dir();

	widgets.preset_frame = gtk_frame_new(NULL);
	gtk_frame_set_shadow_type(GTK_FRAME(widgets.preset_frame), GTK_SHADOW_IN);

	widgets.preset_listbox = gtk_scrolled_window_new (NULL, NULL);
	gtk_scrolled_window_set_policy (GTK_SCROLLED_WINDOW (widgets.preset_listbox), 
									GTK_POLICY_AUTOMATIC, GTK_POLICY_AUTOMATIC);
	gtk_widget_set_size_request (widgets.preset_listbox, 200, 80);
	gtk_container_add (GTK_CONTAINER(widgets.preset_frame), 
					   widgets.preset_listbox);

	widgets.preset_store = gtk_list_store_new (1, G_TYPE_STRING);
	widgets.preset_tree = gtk_tree_view_new_with_model (GTK_TREE_MODEL (widgets.preset_store));
	gtk_tree_view_set_headers_visible (GTK_TREE_VIEW(widgets.preset_tree), FALSE);
	gtk_tree_view_append_column (GTK_TREE_VIEW (widgets.preset_tree),
								 gtk_tree_view_column_new_with_attributes (NULL,
								 gtk_cell_renderer_text_new (), "text", 0, NULL));

	g_signal_connect (widgets.preset_tree, "row-activated", G_CALLBACK(preset_tree_dbclicked), NULL);

	gtk_container_add (GTK_CONTAINER(widgets.preset_listbox), widgets.preset_tree);

	widgets.preset_load   = gtk_button_new_with_label(_("Load"));
	widgets.preset_save   = gtk_button_new_with_label(_("Save"));
	widgets.preset_delete = gtk_button_new_with_label(_("Delete"));
	widgets.preset_update = gtk_button_new_with_label(_("Update List"));
	g_signal_connect (widgets.preset_load, "clicked", G_CALLBACK(preset_load_callback), NULL);
	g_signal_connect (widgets.preset_save, "clicked", G_CALLBACK(preset_save_callback), vals);
	g_signal_connect (widgets.preset_delete, "clicked", G_CALLBACK(preset_delete_callback), NULL);
	g_signal_connect (widgets.preset_update, "clicked", G_CALLBACK(preset_update_callback), NULL);

	preset_update_callback (widgets.preset_update, NULL);

	gtk_table_attach(GTK_TABLE(preset_table), widgets.preset_frame,
					 0, 1, 0, 4, GTK_FILL, GTK_EXPAND|GTK_FILL, 2, 2);

	gtk_table_attach(GTK_TABLE(preset_table), widgets.preset_load,
					 1, 2, 0, 1, GTK_FILL, 0, 2, 2);

	gtk_table_attach(GTK_TABLE(preset_table), widgets.preset_save,
					 1, 2, 1, 2, GTK_FILL, 0, 2, 2);

	gtk_table_attach(GTK_TABLE(preset_table), widgets.preset_delete,
					 1, 2, 2, 3, GTK_FILL, 0, 2, 2);

	gtk_table_attach(GTK_TABLE(preset_table), widgets.preset_update,
					 2, 3, 2, 3, GTK_FILL, 0, 20, 2);
/*}}}*/

    //
    // page 8 font directory
    //
/*{{{*/
    widgets.fdir_patheditor = 
        gimp_path_editor_new(_("Select Font Directory."), vals->fontdirs);

    g_signal_connect (widgets.fdir_patheditor, "path-changed", G_CALLBACK (fdir_change_path_callback), vals);

	gtk_table_attach(GTK_TABLE(fdir_table), widgets.fdir_patheditor,
					 0, 1, 0, 1, GTK_EXPAND|GTK_FILL, GTK_EXPAND|GTK_FILL, 2, 2);
/*}}}*/

	//
	// show all dialog
	//
	if (vals->gui.width > 100 && vals->gui.height > 100) {
		gtk_window_resize (GTK_WINDOW(dlg), vals->gui.width, vals->gui.height);
	}
	if (vals->gui.x < 0) {
		vals->gui.x = 0;
	}
	if (vals->gui.y < 0) {
		vals->gui.y = 0;
	}
	if (vals->gui.x == 0 && vals->gui.y == 0) {
		gtk_window_set_position(GTK_WINDOW(dlg), GTK_WIN_POS_CENTER_ALWAYS);
	} else {
		gtk_window_move (GTK_WINDOW(dlg), vals->gui.x, vals->gui.y);
	}

	unit_changed(widgets.unit, vals);

	if (fontlist == NULL) {
		gtk_widget_set_sensitive(widgets.dlg_add, FALSE);
	}

	gtk_widget_grab_focus (widgets.text_view);
	gtk_widget_show_all (dlg);
	gtk_notebook_set_current_page (notebook, select_page);
	gtk_window_set_keep_above (GTK_WINDOW(dlg), vals->gui.topmost);

	hide_splash();
	
	gtk_main();
	return ERROR_NOTHING;
}/*}}}*/

static int run_non_interactive (PluginVals	*vals)
{/*{{{*/
	gint32  out_layer_ID;
	
	gimp_context_get_foreground(&vals->fgcolor);
	gimp_context_get_background(&vals->bgcolor);
	fontlist = (GList*)font_list_create(gimp_path_parse(vals->fontdirs, 16, FALSE, NULL));
	if (fontlist == NULL) {
		WNGMSG ("Error: No fontlist");
	}
	vals->typesetter->font_list = fontlist;

	{
		GList *font = fontlist;
		while (font) {
			FontInfo *finfo = font->data;
			gchar *input_name = vals->styles[FONT_STYLE_BASE].name;
			if (g_strcmp0(input_name, finfo->name) == 0 ||
				g_strcmp0(input_name, finfo->name_en) == 0 ||
				g_strcmp0(input_name, finfo->name_old) == 0 ||
				g_strcmp0(input_name, finfo->name_ps) == 0) {
				vals->styles[FONT_STYLE_BASE].file_path = finfo->path;
				vals->styles[FONT_STYLE_BASE].face_index = finfo->face_index;
				break;
			}
			font = g_list_next(font);
		}
		if (vals->styles[FONT_STYLE_BASE].file_path == NULL) {
			return -1;
		}
		vals->styles[FONT_STYLE_KANA]  = vals->styles[FONT_STYLE_BASE];
		vals->styles[FONT_STYLE_RUBY]  = vals->styles[FONT_STYLE_BASE];
		vals->styles[FONT_STYLE_ASCII] = vals->styles[FONT_STYLE_BASE];
	}
	do_tategaki (vals, &out_layer_ID);
	return out_layer_ID;
}/*}}}*/

static int do_tategaki (PluginVals  *vals,
						gint32      *out_layer_ID)
{/*{{{*/
	/* Freetype */
	FontStyle   *base_style = NULL;
	FontStyle   *kana_style = NULL;
	FontStyle   *ruby_style = NULL;
	FontStyle   *ascii_style = NULL;
	TypesetterOption *option = &vals->option;
	char        *text   = vals->text;
	guint8      *fbuf   = NULL;
	int         selection_ID = 0;
	gboolean    have_selection;
	gint        width, height;
	gsize       size;
	/* Layer */
	gint32          layer_ID = 0;
	GimpImageType   type;
	GimpDrawable    *drawable = NULL;
	GimpRGB         bgcolor;
	guint8          rgba_bg[3];
	guint8          *buf = NULL;
	GimpPixelRgn    rgn;
	char            *layer_name = NULL;
	gint32          fuchidori;
	int             planes;

	int error = ERROR_NOTHING;

	if (text == NULL || strlen(text) == 0) {
		error = ERROR_INVALID_TEXT_ERROR;
		goto done;
	}

	if (fontlist == NULL ||
		vals->styles[FONT_STYLE_BASE].name == NULL ||
		vals->styles[FONT_STYLE_KANA].name == NULL ||
		vals->styles[FONT_STYLE_RUBY].name == NULL ||
		vals->styles[FONT_STYLE_ASCII].name == NULL) {
		error = ERROR_INVALID_FONT_LIST_ERROR;
		goto done;
	}

	base_style = typesetter_fontstyle_copy(&vals->styles[FONT_STYLE_BASE]);
	kana_style = typesetter_fontstyle_copy(&vals->styles[FONT_STYLE_KANA]);
	ruby_style = typesetter_fontstyle_copy(&vals->styles[FONT_STYLE_RUBY]);
	ascii_style = typesetter_fontstyle_copy(&vals->styles[FONT_STYLE_ASCII]);

	{
		gdouble resx, resy;
		gimp_image_get_resolution(vals->image_ID, &resx, &resy);
		base_style->dpi_x =
		kana_style->dpi_x =
		ruby_style->dpi_x =
		ascii_style->dpi_x = (guint32)resx;
		base_style->dpi_y =
		kana_style->dpi_y =
		ruby_style->dpi_y =
		ascii_style->dpi_y = (guint32)resy;
	}

	vals->typesetter->option = *option;
	switch (gimp_image_base_type (vals->image_ID)) {
		case GIMP_RGB:
			vals->typesetter->option.render_mode = RENDER_MODE_RGBA;
			gimp_rgb_get_uchar (&vals->fgcolor, 
								&vals->typesetter->option.color.r,
								&vals->typesetter->option.color.g,
								&vals->typesetter->option.color.b);
			vals->typesetter->option.color.a = 0xFF;
			planes = 4;
			break;
		case GIMP_GRAY:
			vals->typesetter->option.render_mode = RENDER_MODE_GA;
			gimp_rgb_get_uchar (&vals->fgcolor, 
								&vals->typesetter->option.color.r,
								&vals->typesetter->option.color.g,
								&vals->typesetter->option.color.b);
			vals->typesetter->option.color.r = color_to_intensity_uchar(vals->typesetter->option.color);
			vals->typesetter->option.color.g = vals->typesetter->option.color.r;
			vals->typesetter->option.color.b = vals->typesetter->option.color.r;
			vals->typesetter->option.color.a = 0xFF;
			planes = 2;
			break;
		default: g_assert_not_reached(); break;
	}

	if ( typesetter_reset(vals->typesetter) != 0 ) {
		WNGMSG(_("Error:Invalid font options."));
		error = ERROR_INVALID_FONT_ERROR;
		goto done;
	}
	if ( typesetter_set_style(vals->typesetter, FONT_STYLE_BASE, base_style) != 0 ) {
		WNGMSG(_("Error:Invalid font base style."));
		error = ERROR_INVALID_FONT_ERROR;
		goto done;
	}
	if ( vals->gui.use_fontset[FONT_STYLE_KANA] ) {
		if (typesetter_set_style(vals->typesetter, FONT_STYLE_KANA, kana_style) != 0 ) {
		WNGMSG(_("Error:Invalid font kana style."));
		error = ERROR_INVALID_FONT_ERROR;
		goto done;
		}
		DBGMSG("tategaki : set kana font");
	}
	if ( vals->gui.use_fontset[FONT_STYLE_RUBY] ) {
		if (typesetter_set_style(vals->typesetter, FONT_STYLE_RUBY, ruby_style) != 0 ) {
		WNGMSG(_("Error:Invalid font ruby style."));
		error = ERROR_INVALID_FONT_ERROR;
		goto done;
		}
		DBGMSG("tategaki : set ruby font");
	}
	if ( vals->gui.use_fontset[FONT_STYLE_ASCII] ) {
		if (typesetter_set_style(vals->typesetter, FONT_STYLE_ASCII, ascii_style) != 0 ) {
		WNGMSG(_("Error:Invalid font ascii style."));
		error = ERROR_INVALID_FONT_ERROR;
		goto done;
		}
		DBGMSG("tategaki : set ascii font");
	}
	if ( typesetter_render(vals->typesetter, text, &fbuf, &width, &height) != 0) {
		WNGMSG(_("Error:Font rendering failed."));
		error = ERROR_RENDERING_ERROR;
		goto done;
	}
	if (fbuf == NULL || width < 1 || height < 1) {
		WNGMSG(_("Error:Font rendering failed. or too small."));
		error = ERROR_RENDERING_ERROR;
		goto done;
	}
	size = width * height;

	if (GIMP_MAX_IMAGE_SIZE < width || 
		GIMP_MAX_IMAGE_SIZE < height) {
		WNGMSG(_("Error:Image is too large."));
		error = ERROR_TOO_LARGE_IMAGE_SIZE_ERROR;
		goto done;
	}
	if (GIMP_MIN_IMAGE_SIZE > width || 
		GIMP_MIN_IMAGE_SIZE > height) {
		WNGMSG(_("Error:Image is too small."));
		error = ERROR_TOO_SMALL_IMAGE_SIZE_ERROR;
		goto done;
	}

	bgcolor = vals->bgcolor;
	gimp_rgb_get_uchar (&bgcolor, rgba_bg, rgba_bg+1, rgba_bg+2);
	fuchidori = (int)(typesetter_get_pixel_size(vals->typesetter, FONT_STYLE_BASE) * vals->option.fuchidori / 100.0);


	switch (gimp_image_base_type (vals->image_ID)) {
		case GIMP_RGB:
		{
			gsize buf_size = (width + fuchidori * 2) * (height + fuchidori * 2) * planes;
			guint8 *dst;

			buf = g_malloc0(sizeof(guint8) * buf_size);
			dst = buf;

			if (fuchidori > 0) {
				int y, x, yy, xx;
				//background(edging)
				for (y=0; y<height; y++) {
					for (x=0; x<width; x++) {
						int pos = (x+y*width) * planes;
						guint8 a = fbuf[pos+3];
						if (a == 0) continue;
						for (yy = y; yy <= y + fuchidori*2; yy++) {
							for (xx = x; xx <= x + fuchidori*2; xx++) {
								dst = &buf[(xx + yy * (width + fuchidori*2))*planes];
								*dst++ = rgba_bg[0];
								*dst++ = rgba_bg[1];
								*dst++ = rgba_bg[2];
								*dst   = MIN(0xFF, *dst + a);
							}
						}
					}
				}
				
				//foreground
				for (y=0; y<height; y++) {
					for (x=0; x<width; x++) {
						int pos = (x+y*width) * planes;
						guint32 p = ((x+fuchidori) + (y+fuchidori)*(width+fuchidori*2)) * planes;
						gdouble r = fbuf[pos+0];
						gdouble g = fbuf[pos+1];
						gdouble b = fbuf[pos+2];
						gdouble a = fbuf[pos+3];
						gdouble a_norm = fbuf[pos+3] / 255.0;
						gdouble a_norm_v = 1.0 - a_norm;
						if (a == 0) continue;
						dst = &buf[p];
						*dst++ = (guint8)CLAMP0255(rgba_bg[0] * a_norm_v + r * a_norm);
						*dst++ = (guint8)CLAMP0255(rgba_bg[1] * a_norm_v + g * a_norm);
						*dst++ = (guint8)CLAMP0255(rgba_bg[2] * a_norm_v + b * a_norm);
						*dst   = (guint8)CLAMP0255(*dst + a);
					}
				}
			} else {
				memcpy(dst, fbuf, size*planes);
			}
			type = GIMP_RGBA_IMAGE;
			break;
		}
		case GIMP_GRAY:
		{
			guint8 i_bg;
			gsize buf_size = (width + fuchidori * 2) * (height + fuchidori * 2) * 2;
			guint8 *dst;
			Color bg_color;
			bg_color.r = rgba_bg[0];
			bg_color.g = rgba_bg[1];
			bg_color.b = rgba_bg[2];
			bg_color.a = 0xFF;

			i_bg = color_to_intensity_uchar(bg_color);
			buf = g_malloc0(sizeof(guint8) * buf_size);
			dst = buf;

			if (fuchidori > 0) {
				int y, x, yy, xx;

				//background(edging)
				for (y=0; y<height; y++) {
					for (x=0; x<width; x++) {
						int pos = (x+y*width) * planes;
						guint8 a = fbuf[pos+1];
						if (a == 0) continue;
						for (yy = y; yy <= y + fuchidori*2; yy++) {
							for (xx = x; xx <= x + fuchidori*2; xx++) {
								dst = &buf[(xx + yy * (width + fuchidori*2))*planes];
								*dst++ = i_bg;
								*dst   = CLAMP0255(*dst + a);
							}
						}
					}
				}

				//foreground
				for (y=0; y<height; y++) {
					for (x=0; x<width; x++) {
						int pos = (x+y*width) * planes;
						guint32 p = ((x+fuchidori) + (y+fuchidori)*(width+fuchidori*2)) * planes;
						gdouble i = fbuf[pos+0];
						gdouble a = fbuf[pos+1];
						gdouble a_norm = fbuf[pos+1] / 255.0;
						if (a == 0) continue;
						dst = &buf[p];
						*dst++ = (guint8)CLAMP0255(i_bg * (1.0 - a_norm) + i * a_norm);
						*dst   = (guint8)CLAMP0255(*dst + a);
					}
				}
			} else {
				memcpy(dst, fbuf, size*planes);
			}
			type = GIMP_GRAYA_IMAGE;
			break;
		}
		default:
			error = ERROR_UNKNOWN_IMAGE_TYPE_ERROR;
			goto done;
			break;
	}
	g_free(fbuf); fbuf = NULL;

	//update width and height, if fuchidori has value.
	width = width   + (fuchidori * 2);
	height = height + (fuchidori * 2);

	//remove fuchidori antialias
	if (fuchidori > 0 && vals->option.antialias == FALSE) {
		int x, y;
		int chan_num = 4;
		guint8 *dst;
		switch (gimp_image_base_type (vals->image_ID)) {
			case GIMP_RGB: chan_num = 4; break;
			case GIMP_GRAY: chan_num = 2; break;
			case GIMP_INDEXED: chan_num = 1; break; //don't used
		}
		for (y=0; y<height; y++) {
			for (x=0; x<width; x++) {
				dst = &buf[(x+y*width) * chan_num];
				dst += chan_num -1;
				*dst = *dst > 64 ? 0xFF : 0x00;
			}
		}
	}

	have_selection = !gimp_selection_is_empty(vals->image_ID);

	gimp_image_undo_group_start(vals->image_ID);

	//add layer
	if (vals->mode == RUNMODE_ADD) {
		// create new layer and copy the image
		layer_name = create_layer_name (text);
		layer_ID = gimp_layer_new (
				vals->image_ID,
				layer_name,
				width,
				height,
				type,
				100,
				GIMP_NORMAL_MODE);
		if (layer_name) {
			g_free(layer_name);
			layer_name = NULL;
		}
		if (layer_ID < 0) {
			WNGMSG("Error:Failed to create new layer");
			error = ERROR_FAILED_CREATE_LAYER_ERROR;
			goto done;
		}

		stub_func_gimp_image_insert_layer(vals->image_ID, layer_ID, -1, -1);
		gimp_image_set_active_layer(vals->image_ID, layer_ID);

		//offset layer to selection rectangle
		if (!have_selection) {
			if (vals->x != 0 || vals->y != 0) {
				gimp_layer_set_offsets(layer_ID, vals->x, vals->y);
			} else {
				gimp_layer_set_offsets(layer_ID, gimp_image_width(vals->image_ID) - width ,0);
			}
		} else {
			gboolean non_empty;
			gint sx1, sy1, sx2, sy2;
			gimp_selection_bounds(vals->image_ID, &non_empty, &sx1, &sy1, &sx2, &sy2);
			gimp_layer_set_offsets(layer_ID, sx2 - width, sy1);
		}
	}

	//modify layer
	if (vals->mode == RUNMODE_MODIFY) {
		int oldwidth, oldheight;
		layer_ID = vals->layer_ID;

		layer_name = create_layer_name (text);
		gimp_drawable_set_name(layer_ID, layer_name);

		if (layer_name) {
			g_free(layer_name);
			layer_name = NULL;
		}

		oldwidth = gimp_drawable_width(layer_ID);
		oldheight = gimp_drawable_height(layer_ID);

		gimp_layer_resize(layer_ID, width, height, width - oldwidth, 0);
	}

	if (have_selection) {
		selection_ID = gimp_selection_save (vals->image_ID);
		gimp_selection_none(vals->image_ID);
	}

	drawable = gimp_drawable_get (layer_ID);

	gimp_tile_cache_ntiles ( 4 * (drawable->width / gimp_tile_width() + 1));

	gimp_pixel_rgn_init (&rgn, drawable, 0, 0, width, height, TRUE, TRUE);

	gimp_pixel_rgn_set_rect (&rgn, buf, 0, 0, width, height);

	gimp_drawable_flush (drawable);

	gimp_drawable_merge_shadow( drawable->drawable_id, TRUE);

	gimp_drawable_update (drawable->drawable_id, 0, 0, drawable->width, drawable->height);

	gimp_drawable_detach (drawable);

	// attatch parasite
	{
		GimpParasite *parasite;
		parasite = parasite_from_pluginvals(vals);
		if (parasite != NULL) {
			gimp_drawable_parasite_detach(layer_ID, PLUGIN_ID);
			gimp_drawable_parasite_attach(layer_ID, parasite);
		}
		gimp_parasite_free(parasite);
	}

	if (have_selection) {
		gimp_selection_load(selection_ID);
		gimp_image_remove_channel(vals->image_ID, selection_ID);
	}

	gimp_image_undo_group_end(vals->image_ID);

	gimp_displays_flush();

done:
	if (layer_name) g_free(layer_name);
	if (fbuf) g_free(fbuf);
	if (buf) g_free(buf);
	*out_layer_ID = (error == ERROR_NOTHING) ? layer_ID : -1;
	return error;
}/*}}}*/

static char* create_layer_name (char *text)
{/*{{{*/
	gboolean overflow = 0;
	int all_num, cpy_num;
	char *dst, *ret;
	all_num = g_utf8_strlen (text, -1);
	overflow = all_num > 20;
	cpy_num = overflow ? 20 : all_num;
	
	dst = ret = g_malloc0(64);
	g_utf8_strncpy (dst, text, cpy_num);
	dst += strlen(dst);
	if (overflow) {
		g_utf8_strncpy (dst, "...", 3);
	}
	
	dst = ret;
	while (*dst) {
		if (*dst == 0x0a) {
			*dst = 0x2F;
		}
		dst = g_utf8_next_char(dst);
	}
	return ret;
}/*}}}*/

static void rc_load(gchar *filename, PluginVals *vals)
{/*{{{*/
	GKeyFile *keyfile;
	gchar *rcfile = gimp_personal_rc_file (filename);

	keyfile = g_key_file_new();
	g_key_file_load_from_file(keyfile, rcfile, G_KEY_FILE_NONE, NULL);
	if (g_key_file_has_key(keyfile, PNAME_GROUP, PNAME_FONTDIR, NULL)) {
		gchar *pathlist = g_key_file_get_string(keyfile, PNAME_GROUP, PNAME_FONTDIR, NULL);
		if (vals->fontdirs) {
			g_free(vals->fontdirs);
		}
		vals->fontdirs = pathlist;
	}

	if (g_key_file_has_key(keyfile, PNAME_GROUP, PNAME_BATCH_SEPARATOR, NULL)) {
		if (vals->batch_separator) {
			g_free(vals->batch_separator);
		}
		vals->batch_separator = g_key_file_get_string(keyfile, PNAME_GROUP, PNAME_BATCH_SEPARATOR, NULL);
	} else {
		vals->batch_separator = g_strdup("----");
	}

	if (rcfile) g_free(rcfile);
	if (keyfile) g_key_file_free(keyfile);
}/*}}}*/

static void rc_save(gchar *filename, PluginVals *vals)
{/*{{{*/
	FILE *fp;
	GKeyFile *keyfile;
    gchar *rcfile = NULL;
	gchar *data = NULL;
    
	if (vals->fontdirs == NULL) {
		return;
	}
	rcfile = gimp_personal_rc_file (filename);
	
	keyfile = g_key_file_new();
	if (vals->fontdirs != NULL) {
		g_key_file_set_string(keyfile, PNAME_GROUP, PNAME_FONTDIR, vals->fontdirs);
	}
	if (vals->batch_separator != NULL) {
		g_key_file_set_string(keyfile, PNAME_GROUP, PNAME_BATCH_SEPARATOR, vals->batch_separator);
	}
	data = g_key_file_to_data(keyfile, NULL, NULL);

	//FIXME: fopen waring
	//Because Glib doesn't have g_fclose, we can not use g_fopen.
	//If another C runtime was linked, it makes crash the plug-in.
	
#ifdef _WIN32
	{
		gunichar2 *mode = NULL;
		gunichar2 *temp = NULL;
		mode = g_utf8_to_utf16("wb", -1, NULL, NULL, NULL);
		temp = g_utf8_to_utf16(rcfile, -1, NULL, NULL, NULL);
		_wfopen_s(&fp, temp, mode);
		if (mode) g_free(mode);
		if (temp) g_free(temp);
	}
#else
	fp = fopen(rcfile, "wb");
#endif
	fprintf(fp, "%s", data);
	fclose(fp);

	if (data) g_free(data);
	if (rcfile) g_free(rcfile);
	if (keyfile) g_key_file_free(keyfile);
}/*}}}*/

static void parasite_to_widgets (GimpParasite *parasite)
{/*{{{*/
	gchar* data;
	gint32 size;
	GKeyFile *kfile;
	gboolean result;
	if (parasite == NULL) return;

	data = (gchar*)gimp_parasite_data(parasite);
	size = gimp_parasite_data_size(parasite);

	kfile = g_key_file_new();
	result = g_key_file_load_from_data(kfile, data, size, G_KEY_FILE_NONE, NULL);
	if (kfile != NULL) {
		keyfile_to_widgets(kfile);
		g_key_file_free(kfile);
	}
}/*}}}*/

static GimpParasite* parasite_from_pluginvals (PluginVals *vals)
{/*{{{*/
	GKeyFile *kfile;
	gchar *data;
	GimpParasite *parasite;

	kfile = keyfile_from_pluginvals(vals);

	{
		gchar **list;
		list = g_key_file_get_keys(kfile, PNAME_GROUP, NULL, NULL);
	}
	data = g_key_file_to_data(kfile, NULL, NULL);
	parasite = gimp_parasite_new(PLUGIN_ID, 
								GIMP_PARASITE_PERSISTENT | GIMP_PARASITE_UNDOABLE,
								(guint32)strlen(data) * sizeof(gchar), data);

	g_free(data);
	g_key_file_free(kfile);

	return parasite;
}/*}}}*/ 

static GKeyFile* keyfile_from_pluginvals (PluginVals *vals)
{/*{{{*/
    GKeyFile *kfile;
	guchar rgb[3];
	gint   list[3];

	kfile = g_key_file_new();

	g_key_file_set_string(kfile, PNAME_GROUP, PNAME_VERSION, PLUGIN_VERSION);

	if (vals->text != NULL) {
		g_key_file_set_string(kfile, PNAME_GROUP, PNAME_TEXT, vals->text);
	}

	gimp_context_get_foreground(&vals->fgcolor);
	gimp_context_get_background(&vals->bgcolor);

	gimp_rgb_get_uchar(&vals->fgcolor, rgb, rgb+1, rgb+2);
	list[0] = (gint)rgb[0];
	list[1] = (gint)rgb[1];
	list[2] = (gint)rgb[2];
	g_key_file_set_integer_list(kfile, PNAME_GROUP, PNAME_FGCOLOR, list, 3);
	g_key_file_set_integer_list(kfile, PNAME_GROUP, PNAME_COLOR, list, 3); //keep compatibility to old versions

	gimp_rgb_get_uchar(&vals->bgcolor, rgb, rgb+1, rgb+2);
	list[0] = (gint)rgb[0];
	list[1] = (gint)rgb[1];
	list[2] = (gint)rgb[2];
	g_key_file_set_integer_list(kfile, PNAME_GROUP, PNAME_BGCOLOR, list, 3);

	g_key_file_set_integer(kfile, PNAME_GROUP, PNAME_UNIT, vals->gui.unit);

	g_key_file_set_integer(kfile, PNAME_GROUP, PNAME_TEXT_DIR, vals->option.text_dir);

	g_key_file_set_integer(kfile, PNAME_GROUP, PNAME_TSUME_TYPE, vals->option.tsume_type);

	g_key_file_set_integer(kfile, PNAME_GROUP, PNAME_LINE_ALIGN, vals->option.line_align);

	g_key_file_set_integer(kfile, PNAME_GROUP, PNAME_GYOU_OKURI, vals->option.gyou_okuri);

	g_key_file_set_integer(kfile, PNAME_GROUP, PNAME_MOJI_OKURI, vals->option.moji_okuri);

	g_key_file_set_boolean(kfile, PNAME_GROUP, PNAME_KERNING, vals->option.kerning);

	g_key_file_set_boolean(kfile, PNAME_GROUP, PNAME_ANTIALIAS, vals->option.antialias);

	g_key_file_set_integer(kfile, PNAME_GROUP, PNAME_FUCHIDORI, vals->option.fuchidori);

	g_key_file_set_boolean(kfile, PNAME_GROUP, PNAME_RUBY, TRUE);//keep it for backward-compatibility

	g_key_file_set_boolean(kfile, PNAME_GROUP, PNAME_RUBY_TRANS, vals->option.ruby_trans);

	g_key_file_set_double(kfile, PNAME_GROUP, PNAME_RUBY_MAX_TRANS_RATE, vals->option.ruby_max_trans_rate);

	g_key_file_set_boolean(kfile, PNAME_GROUP, PNAME_VERT_NUMBER, vals->option.auto_yoko);

	g_key_file_set_integer(kfile, PNAME_GROUP, PNAME_VNUM_DIGIT, vals->option.auto_yoko_digit);

	g_key_file_set_boolean(kfile, PNAME_GROUP, PNAME_DOT_SKIP_YAKU, vals->option.dot_skip_yaku);
	
	g_key_file_set_boolean(kfile, PNAME_GROUP, PNAME_DOT_ALIGN, vals->option.dot_align);

	//Font Style
	{
		FontStyle *style;

		//base
		style = &vals->styles[FONT_STYLE_BASE];
		g_key_file_set_string(kfile, PNAME_GROUP, PNAME_FONT_BASE_PATH, style->file_path);
		g_key_file_set_string(kfile, PNAME_GROUP, PNAME_FONT_BASE_NAME, style->name);
		g_key_file_set_integer(kfile, PNAME_GROUP, PNAME_FONT_BASE_INDEX, style->face_index);
		g_key_file_set_double(kfile, PNAME_GROUP, PNAME_FONT_BASE_SIZE, vals->gui.font_size[FONT_STYLE_BASE]);
		g_key_file_set_double(kfile, PNAME_GROUP, PNAME_FONT_BASE_USE, vals->gui.use_fontset[FONT_STYLE_BASE]);

		//kana
		style = &vals->styles[FONT_STYLE_KANA];
		g_key_file_set_string(kfile, PNAME_GROUP, PNAME_FONT_KANA_PATH, style->file_path ? style->file_path : "");
		g_key_file_set_string(kfile, PNAME_GROUP, PNAME_FONT_KANA_NAME, style->name ? style->name :  "");
		g_key_file_set_integer(kfile, PNAME_GROUP, PNAME_FONT_KANA_INDEX, style->face_index);
		g_key_file_set_double(kfile, PNAME_GROUP, PNAME_FONT_KANA_SIZE, vals->gui.font_size[FONT_STYLE_KANA]);
		g_key_file_set_double(kfile, PNAME_GROUP, PNAME_FONT_KANA_USE, vals->gui.use_fontset[FONT_STYLE_KANA]);

		//ruby
		style = &vals->styles[FONT_STYLE_RUBY];
		g_key_file_set_string(kfile, PNAME_GROUP, PNAME_FONT_RUBY_PATH, style->file_path ? style->file_path : "");
		g_key_file_set_string(kfile, PNAME_GROUP, PNAME_FONT_RUBY_NAME, style->name ? style->name : "");
		g_key_file_set_integer(kfile, PNAME_GROUP, PNAME_FONT_RUBY_INDEX, style->face_index);
		g_key_file_set_double(kfile, PNAME_GROUP, PNAME_FONT_RUBY_SIZE, vals->gui.font_size[FONT_STYLE_RUBY]);
		g_key_file_set_double(kfile, PNAME_GROUP, PNAME_FONT_RUBY_USE, vals->gui.use_fontset[FONT_STYLE_RUBY]);

		//ascii
		style = &vals->styles[FONT_STYLE_ASCII];
		g_key_file_set_string(kfile, PNAME_GROUP, PNAME_FONT_ASCII_PATH, style->file_path ? style->file_path : "");
		g_key_file_set_string(kfile, PNAME_GROUP, PNAME_FONT_ASCII_NAME, style->name ? style->name : "");
		g_key_file_set_integer(kfile, PNAME_GROUP, PNAME_FONT_ASCII_INDEX, style->face_index);
		g_key_file_set_double(kfile, PNAME_GROUP, PNAME_FONT_ASCII_SIZE, vals->gui.font_size[FONT_STYLE_ASCII]);
		g_key_file_set_double(kfile, PNAME_GROUP, PNAME_FONT_ASCII_USE, vals->gui.use_fontset[FONT_STYLE_ASCII]);
    }

    return kfile;
}/*}}}*/

static void keyfile_to_widgets(GKeyFile *kfile)
{/*{{{*/
	gint selected;
	gchar *str;
	gint *list;
	GimpRGB color;

	//g_key_file_get_string(kfile, PNAME_GROUP, PNAME_VERSION, PLUGIN_VERSION, null);

	if (g_key_file_has_key(kfile, PNAME_GROUP, PNAME_TEXT, NULL) == TRUE) {
		str = g_key_file_get_string(kfile, PNAME_GROUP, PNAME_TEXT, NULL);
		textview_set_text(widgets.text_view, str);
		if (str) {
			g_free(str);
			str = NULL;
		}
	}

	// keep compatibility to old versions. (renamed to PNAME_FGCOLOR)
	if (g_key_file_has_key(kfile, PNAME_GROUP, PNAME_COLOR, NULL) == TRUE) {
		list = g_key_file_get_integer_list(kfile, PNAME_GROUP, PNAME_COLOR, NULL, NULL);
		gimp_rgb_set_uchar(&color, CLAMP0255(list[0]), CLAMP0255(list[1]), CLAMP0255(list[2]));
		color.a = 1;
		gimp_context_set_foreground(&color);
		if (list) {
			g_free(list);
			list = NULL;
		}
	}

	if (g_key_file_has_key(kfile, PNAME_GROUP, PNAME_FGCOLOR, NULL) == TRUE) {
		list = g_key_file_get_integer_list(kfile, PNAME_GROUP, PNAME_FGCOLOR, NULL, NULL);
		gimp_rgb_set_uchar(&color, CLAMP0255(list[0]), CLAMP0255(list[1]), CLAMP0255(list[2]));
		color.a = 1;
		gimp_context_set_foreground(&color);
		if (list) {
			g_free(list);
			list = NULL;
		}
	}

	if (g_key_file_has_key(kfile, PNAME_GROUP, PNAME_BGCOLOR, NULL) == TRUE) {
		list = g_key_file_get_integer_list(kfile, PNAME_GROUP, PNAME_BGCOLOR, NULL, NULL);
		gimp_rgb_set_uchar(&color, CLAMP0255(list[0]), CLAMP0255(list[1]), CLAMP0255(list[2]));
		color.a = 1;
		gimp_context_set_background(&color);
		if (list) {
			g_free(list);
			list = NULL;
		}
	}

	if (g_key_file_has_key(kfile, PNAME_GROUP, PNAME_UNIT, NULL) == TRUE) {
		gimp_int_combo_box_set_active (GIMP_INT_COMBO_BOX(widgets.unit), 
			(GimpUnit) g_key_file_get_integer(kfile, PNAME_GROUP, PNAME_UNIT, NULL) );
	}

	if (g_key_file_has_key(kfile, PNAME_GROUP, PNAME_TEXT_DIR, NULL) == TRUE) {
		selected = g_key_file_get_integer(kfile, PNAME_GROUP, PNAME_TEXT_DIR, NULL);
		gimp_int_combo_box_set_active (GIMP_INT_COMBO_BOX(widgets.text_dir), selected);
	} else {
		gimp_int_combo_box_set_active (GIMP_INT_COMBO_BOX(widgets.text_dir), DIR_TYPE_RTL);
	}

	if (g_key_file_has_key(kfile, PNAME_GROUP, PNAME_TSUME_TYPE, NULL) == TRUE) {
		selected = g_key_file_get_integer(kfile, PNAME_GROUP, PNAME_TSUME_TYPE, NULL);
		gimp_int_combo_box_set_active (GIMP_INT_COMBO_BOX(widgets.tsume_type), selected);
	}

	if (g_key_file_has_key(kfile, PNAME_GROUP, PNAME_LINE_ALIGN, NULL) == TRUE) {
		selected = g_key_file_get_integer(kfile, PNAME_GROUP, PNAME_LINE_ALIGN, NULL);
		gimp_int_combo_box_set_active (GIMP_INT_COMBO_BOX(widgets.line_align), selected);
	}

	if (g_key_file_has_key(kfile, PNAME_GROUP, PNAME_GYOU_OKURI, NULL) == TRUE) {
		gtk_spin_button_set_value(GTK_SPIN_BUTTON(widgets.gyou_okuri),
			(gdouble)g_key_file_get_integer(kfile, PNAME_GROUP, PNAME_GYOU_OKURI, NULL));
	}

	if (g_key_file_has_key(kfile, PNAME_GROUP, PNAME_MOJI_OKURI, NULL) == TRUE) {
		gtk_spin_button_set_value(GTK_SPIN_BUTTON(widgets.moji_okuri),
			g_key_file_get_integer(kfile, PNAME_GROUP, PNAME_MOJI_OKURI, NULL));
	}

	if (g_key_file_has_key(kfile, PNAME_GROUP, PNAME_KERNING, NULL) == TRUE) {
		gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON(widgets.kerning),
			g_key_file_get_boolean(kfile, PNAME_GROUP, PNAME_KERNING, NULL));
	}

	if (g_key_file_has_key(kfile, PNAME_GROUP, PNAME_ANTIALIAS, NULL) == TRUE) {
		gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON(widgets.antialias),
			g_key_file_get_boolean(kfile, PNAME_GROUP, PNAME_ANTIALIAS, NULL));
	}

	if (g_key_file_has_key(kfile, PNAME_GROUP, PNAME_FUCHIDORI, NULL) == TRUE) {
		gtk_spin_button_set_value(GTK_SPIN_BUTTON(widgets.fuchidori),
			(gdouble)g_key_file_get_integer(kfile, PNAME_GROUP, PNAME_FUCHIDORI, NULL));
	}

	if (g_key_file_has_key(kfile, PNAME_GROUP, PNAME_RUBY_TRANS, NULL) == TRUE) {
		gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON(widgets.ruby_trans),
			g_key_file_get_boolean(kfile, PNAME_GROUP, PNAME_RUBY_TRANS, NULL));
	}

	if (g_key_file_has_key(kfile, PNAME_GROUP, PNAME_RUBY_MAX_TRANS_RATE, NULL) == TRUE) {
		gtk_spin_button_set_value(GTK_SPIN_BUTTON(widgets.ruby_max_trans_rate),
			(gdouble)g_key_file_get_double(kfile, PNAME_GROUP, PNAME_RUBY_MAX_TRANS_RATE, NULL));
	}

	if (g_key_file_has_key(kfile, PNAME_GROUP, PNAME_VERT_NUMBER, NULL) == TRUE) {
		gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON(widgets.auto_yoko),
			g_key_file_get_boolean(kfile, PNAME_GROUP, PNAME_VERT_NUMBER, NULL));
	}

	if (g_key_file_has_key(kfile, PNAME_GROUP, PNAME_VNUM_DIGIT, NULL) == TRUE) {
		gtk_spin_button_set_value(GTK_SPIN_BUTTON(widgets.auto_yoko_digit),
			(gdouble)g_key_file_get_integer(kfile, PNAME_GROUP, PNAME_VNUM_DIGIT, NULL));
	}

	if (g_key_file_has_key(kfile, PNAME_GROUP, PNAME_DOT_SKIP_YAKU, NULL) == TRUE) {
		gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(widgets.dot_skip_yaku),
			g_key_file_get_boolean(kfile, PNAME_GROUP, PNAME_DOT_SKIP_YAKU, NULL));
	}

	if (g_key_file_has_key(kfile, PNAME_GROUP, PNAME_DOT_ALIGN, NULL) == TRUE) {
		gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(widgets.dot_align),
			g_key_file_get_boolean(kfile, PNAME_GROUP, PNAME_DOT_ALIGN, NULL));
	}
	
	//Font Style
	{
		GtkWidget *usewidget;
		GtkWidget *sizewidget;
		gchar *path = 0;
		gchar *name = 0;
		int index = 0;
		gboolean usedstyle;

		//base
		if (g_key_file_has_key(kfile,PNAME_GROUP, PNAME_FONT_BASE_PATH, NULL) == TRUE) {
			path  = g_key_file_get_string(kfile, PNAME_GROUP, PNAME_FONT_BASE_PATH, NULL);
		}
		if (g_key_file_has_key(kfile,PNAME_GROUP, PNAME_FONT_BASE_NAME, NULL) == TRUE) {
			name  = g_key_file_get_string(kfile, PNAME_GROUP, PNAME_FONT_BASE_NAME, NULL);
		}
		if (g_key_file_has_key(kfile,PNAME_GROUP, PNAME_FONT_BASE_INDEX, NULL) == TRUE) {
			index = g_key_file_get_integer(kfile, PNAME_GROUP, PNAME_FONT_BASE_INDEX, NULL);
		}
		if (name != NULL) {
			GList *finfolist = font_list_find_name(fontlist, name);
			if (finfolist != NULL) {
				fontcombo_select (FONT_STYLE_BASE, path, ((FontInfo*)finfolist->data)->name, index);
			}
		}
		if (g_key_file_has_key(kfile,PNAME_GROUP, PNAME_FONT_BASE_SIZE, NULL) == TRUE) {
			gtk_spin_button_set_value(GTK_SPIN_BUTTON(widgets.size),
				g_key_file_get_double(kfile, PNAME_GROUP, PNAME_FONT_BASE_SIZE, NULL));
		}
		if (path) g_free(path);
		if (name) g_free(name);
		path = NULL;
		name = NULL;
		index = 0;

		//kana
		usedstyle = FALSE;
		usewidget  = widgets.font_style[FONT_STYLE_KANA].use;
		sizewidget = widgets.font_style[FONT_STYLE_KANA].size;

		if (g_key_file_has_key(kfile,PNAME_GROUP,PNAME_FONT_KANA_USE , NULL) == TRUE) {
			usedstyle = g_key_file_get_boolean(kfile, PNAME_GROUP, PNAME_FONT_KANA_USE, NULL);
		}
		gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON(usewidget), usedstyle);

		if (usedstyle == TRUE) {
			if (g_key_file_has_key(kfile,PNAME_GROUP, PNAME_FONT_KANA_PATH, NULL) == TRUE) {
				path  = g_key_file_get_string(kfile, PNAME_GROUP, PNAME_FONT_KANA_PATH, NULL);
			}
			if (g_key_file_has_key(kfile,PNAME_GROUP, PNAME_FONT_KANA_NAME, NULL) == TRUE) {
				name  = g_key_file_get_string(kfile, PNAME_GROUP, PNAME_FONT_KANA_NAME, NULL);
			}
			if (g_key_file_has_key(kfile,PNAME_GROUP, PNAME_FONT_KANA_INDEX, NULL) == TRUE) {
				index = g_key_file_get_integer(kfile, PNAME_GROUP, PNAME_FONT_KANA_INDEX, NULL);
			}
			if (name) {
				GList *finfolist = font_list_find_name(fontlist, name);
				if (finfolist != NULL) {
					fontcombo_select (FONT_STYLE_KANA, path, name, index);
				}
			}
			if (g_key_file_has_key(kfile,PNAME_GROUP,PNAME_FONT_KANA_SIZE , NULL) == TRUE) {
				gtk_spin_button_set_value(GTK_SPIN_BUTTON(sizewidget),
					g_key_file_get_double(kfile, PNAME_GROUP, PNAME_FONT_KANA_SIZE, NULL));
			}
		}
		if (path) g_free(path);
		if (name) g_free(name);
		path = NULL;
		name = NULL;
		index = 0;

		//ruby
		usedstyle = FALSE;
		usewidget = widgets.font_style[FONT_STYLE_RUBY].use;
		sizewidget = widgets.font_style[FONT_STYLE_RUBY].size;

		if (g_key_file_has_key(kfile,PNAME_GROUP,PNAME_FONT_RUBY_USE , NULL) == TRUE) {
			usedstyle = g_key_file_get_boolean(kfile, PNAME_GROUP, PNAME_FONT_RUBY_USE, NULL);
		}
		gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON(usewidget), usedstyle);

		if (usedstyle == TRUE) {
			if (g_key_file_has_key(kfile,PNAME_GROUP, PNAME_FONT_RUBY_PATH, NULL) == TRUE) {
				path  = g_key_file_get_string(kfile, PNAME_GROUP, PNAME_FONT_RUBY_PATH, NULL);
			}
			if (g_key_file_has_key(kfile,PNAME_GROUP, PNAME_FONT_RUBY_NAME, NULL) == TRUE) {
				name  = g_key_file_get_string(kfile, PNAME_GROUP, PNAME_FONT_RUBY_NAME, NULL);
			}
			if (g_key_file_has_key(kfile,PNAME_GROUP, PNAME_FONT_RUBY_INDEX, NULL) == TRUE) {
				index = g_key_file_get_integer(kfile, PNAME_GROUP, PNAME_FONT_RUBY_INDEX, NULL);
			}
			if (name) {
				GList *finfolist = font_list_find_name(fontlist, name);
				if (finfolist != NULL) {
					fontcombo_select (FONT_STYLE_RUBY, path, name, index);
				}
			}
			if (g_key_file_has_key(kfile,PNAME_GROUP,PNAME_FONT_RUBY_SIZE , NULL) == TRUE) {
				gtk_spin_button_set_value(GTK_SPIN_BUTTON(sizewidget),
					g_key_file_get_double(kfile, PNAME_GROUP, PNAME_FONT_RUBY_SIZE, NULL));
			}
		}
		if (path) g_free(path);
		if (name) g_free(name);
		path = NULL;
		name = NULL;
		index = 0;

		//ascii
		usedstyle = FALSE;
		usewidget  = widgets.font_style[FONT_STYLE_ASCII].use;
		sizewidget = widgets.font_style[FONT_STYLE_ASCII].size;

		if (g_key_file_has_key(kfile,PNAME_GROUP,PNAME_FONT_ASCII_USE , NULL) == TRUE) {
			usedstyle = g_key_file_get_boolean(kfile, PNAME_GROUP, PNAME_FONT_ASCII_USE, NULL);
		}
		gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON(usewidget), usedstyle);

		if (usedstyle == TRUE) {
			if (g_key_file_has_key(kfile,PNAME_GROUP, PNAME_FONT_ASCII_PATH, NULL) == TRUE) {
				path  = g_key_file_get_string(kfile, PNAME_GROUP, PNAME_FONT_ASCII_PATH, NULL);
			}
			if (g_key_file_has_key(kfile,PNAME_GROUP, PNAME_FONT_ASCII_NAME, NULL) == TRUE) {
				name  = g_key_file_get_string(kfile, PNAME_GROUP, PNAME_FONT_ASCII_NAME, NULL);
			}
			if (g_key_file_has_key(kfile,PNAME_GROUP, PNAME_FONT_ASCII_INDEX, NULL) == TRUE) {
				index = g_key_file_get_integer(kfile, PNAME_GROUP, PNAME_FONT_ASCII_INDEX, NULL);
			}
			if (name) {
				GList *finfolist = font_list_find_name(fontlist, name);
				if (finfolist != NULL) {
					fontcombo_select (FONT_STYLE_ASCII, path, name, index);
				}
			}
			if (g_key_file_has_key(kfile,PNAME_GROUP,PNAME_FONT_ASCII_SIZE , NULL) == TRUE) {
				gtk_spin_button_set_value(GTK_SPIN_BUTTON(sizewidget),
					g_key_file_get_double(kfile, PNAME_GROUP, PNAME_FONT_ASCII_SIZE, NULL));
			}
		}
		if (path) g_free(path);
		if (name) g_free(name);
		path = NULL;
		name = NULL;
		index = 0;
	}
}/*}}}*/

MAIN()

/* vim: set fdm=marker ts=4 sw=4:*/
