/*
 * GIMP Tategaki Plug-in -- This is a plug-in for The GIMP 2
 * Copyright (C) 2006-2017 OHTSUKA, Yoshio <ohtsuka.yoshio@gmail.com>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 */
#include "typesetter.h"
#include "typeset_obj_dot.h"
#include "typeset_obj.h"
#include FT_OUTLINE_H


static Typesetter *typ = NULL;
FT_Bitmap* bitmap_table[DOT_TYPE_COUNT]; //0=NONE, 1=MARU,2=GOMA

static FT_Bitmap* build_bitmap (DotType type);

void dot_init(Typesetter *typesetter)
{/*{{{*/
	typ = typesetter;
	bitmap_table[DOT_TYPE_NONE] = NULL;
	bitmap_table[DOT_TYPE_MARU] = build_bitmap(DOT_TYPE_MARU);
	bitmap_table[DOT_TYPE_GOMA] = build_bitmap(DOT_TYPE_GOMA);
}/*}}}*/

void dot_destroy()
{/*{{{*/
	int i;
	typ = NULL;
	for (i = 0; i < DOT_TYPE_COUNT; i++) {
		if (bitmap_table[i]) {
			if (bitmap_table[i]->buffer) {
				g_free(bitmap_table[i]->buffer);
			}
			g_free(bitmap_table[i]);
		}
		bitmap_table[i] = NULL;
	}
}/*}}}*/

void dot_get_size (DotType dottype, double *advance, double *offset, Rect *real_size, Rect *typo_size)
{/*{{{*/
	FT_Face face = typ->face_table[FONT_STYLE_BASE].face;
	Rect zero_size = {0, 0, 0, 0};

	g_assert(typ != NULL);

	*advance = 0;
	*offset  = 0;
	*real_size = zero_size;
	*typo_size = zero_size;
	
	real_size->top = 0;
	real_size->bottom = face->size->metrics.x_ppem/ 2;
	real_size->left = 0;
	real_size->right = real_size->bottom;
	*typo_size = *real_size;
	*advance = real_size->bottom;
	*offset = 0;
}/*}}}*/

void dot_render (DotType dottype, Color color, BufferContainer *buf)
{/*{{{*/
	FT_Bitmap *bitmap;
	FT_Face face = typ->face_table[FONT_STYLE_BASE].face;
	int dx, dy;
	double ppem = face->size->metrics.x_ppem / 2.0;

	g_assert(typ != NULL);

	bitmap = bitmap_table[dottype];
	dx = (int)(buf->x + (ppem - bitmap->width) / 2.0);
	dy = (int)(buf->y + (ppem - bitmap->rows)  / 2.0);

	copy_glyph_image (color,
					  0,
					  typ->option.antialias, 
					  buf->data, 
					  dx,
					  dy,
					  buf->width,
					  buf->height,
					  buf->planes,
					  bitmap->buffer,
					  bitmap->width,
					  bitmap->rows,
					  bitmap->pitch);

}/*}}}*/


static FT_Bitmap* build_bitmap (DotType type)
{/*{{{*/
	FT_Library library = typ->library;
	FT_Face face = typ->face_table[FONT_STYLE_BASE].face;
	FT_Outline outline;
	double ppem = face->size->metrics.x_ppem / 2.0;

	char *tag;
	FT_Vector *vec;
	FT_BBox cbox;
	FT_Bitmap *bitmap = NULL;
	
	if (type == DOT_TYPE_NONE) goto EXIT;
	bitmap = (FT_Bitmap*)g_malloc0(sizeof(FT_Bitmap));
	
#define SET_P(X,Y,TAG) vec->x=DOUBLE_TO_F26D6((X)*scale);vec->y=DOUBLE_TO_F26D6((Y)*scale);vec++;*tag++=(TAG);
	switch (type) {
		case DOT_TYPE_MARU: {
				double em = 2048;
				double scale = ppem / em * 1.8;
				FT_Outline_New(library, 12, 1, &outline);
				outline.n_points   = 12;
				outline.n_contours = 1;
				outline.contours[0] = outline.n_points-1;
				vec = outline.points;
				tag = outline.tags;
				SET_P(-228,   0, FT_CURVE_TAG_ON);
				SET_P(-228, 125, FT_CURVE_TAG_CUBIC);
				SET_P(-125, 228, FT_CURVE_TAG_CUBIC);
				SET_P(   0, 228, FT_CURVE_TAG_ON);
				SET_P( 125, 228, FT_CURVE_TAG_CUBIC);
				SET_P( 228, 125, FT_CURVE_TAG_CUBIC);
				SET_P( 228 ,  0, FT_CURVE_TAG_ON);
				SET_P( 228,-125, FT_CURVE_TAG_CUBIC);
				SET_P( 125,-228, FT_CURVE_TAG_CUBIC);
				SET_P(   0,-228, FT_CURVE_TAG_ON);
				SET_P(-125,-228, FT_CURVE_TAG_CUBIC);
				SET_P(-228,-125, FT_CURVE_TAG_CUBIC);
			}
			break;
		case DOT_TYPE_GOMA: {
				double em = 1000;
				double scale = ppem / em * 2.0;
				FT_Outline_New(library, 19, 1, &outline);
				outline.n_points   = 19;
				outline.n_contours = 1;
				outline.contours[0] = outline.n_points-1;
				vec = outline.points;
				tag = outline.tags;
				SET_P( 10,241, FT_CURVE_TAG_ON);
				SET_P( 94,220, FT_CURVE_TAG_CUBIC);
				SET_P(171,188, FT_CURVE_TAG_CUBIC);
				SET_P(207,150, FT_CURVE_TAG_ON);
				SET_P(219,126, FT_CURVE_TAG_CUBIC);
				SET_P(245,101, FT_CURVE_TAG_CUBIC);
				SET_P(245, 71, FT_CURVE_TAG_ON);
				SET_P(245, 31, FT_CURVE_TAG_CUBIC);
				SET_P(220,  0, FT_CURVE_TAG_CUBIC);
				SET_P(183,  0, FT_CURVE_TAG_ON);
				SET_P(155,  0, FT_CURVE_TAG_CUBIC);
				SET_P(115,  3, FT_CURVE_TAG_CUBIC);
				SET_P(106, 71, FT_CURVE_TAG_ON);
				SET_P( 96,126, FT_CURVE_TAG_CUBIC);
				SET_P( 72,180, FT_CURVE_TAG_CUBIC);
				SET_P(  6,232, FT_CURVE_TAG_ON);
				SET_P(  0,236, FT_CURVE_TAG_CUBIC);
				SET_P(  3,243, FT_CURVE_TAG_CUBIC);
				SET_P( 10,241, FT_CURVE_TAG_ON);
			}
			break;

		case DOT_TYPE_NONE:
		default:
			break;
	}
#undef SET_P

	FT_Outline_Get_CBox(&outline, &cbox );
	cbox.xMin &= ~63;
	cbox.yMin &= ~63;
	cbox.xMax  = (cbox.xMax + 63) & ~63;
	cbox.yMax  = (cbox.yMax + 63) & ~63;

	bitmap->width = (cbox.xMax - cbox.xMin) >> 6;
	bitmap->rows  = (cbox.yMax - cbox.yMin) >> 6;
	bitmap->pitch = bitmap->width;
	if (typ->option.antialias) {
		bitmap->num_grays = 0xFF;
		bitmap->pixel_mode = FT_PIXEL_MODE_GRAY;
	} else {
		bitmap->num_grays = 0;
		bitmap->pixel_mode = FT_PIXEL_MODE_MONO;
	}
	bitmap->buffer = g_malloc0(bitmap->pitch * bitmap->rows);

	FT_Outline_Translate(&outline, -cbox.xMin, -cbox.yMin );
	FT_Outline_Get_Bitmap(library, &outline, bitmap );
	FT_Outline_Done(library, &outline);
EXIT:
	return bitmap;
}/*}}}*/

/* vim: set fdm=marker ts=4 sw=4:*/
