/*
 * GIMP Tategaki Plug-in -- This is a plug-in for The GIMP 2
 * Copyright (C) 2006-2017 OHTSUKA, Yoshio <ohtsuka.yoshio@gmail.com>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 */
#include "encoding.h"
#include "typesetter.h"
#include "typeset_obj_glyphbox.h"
#include "typeset_obj_ruby.h"
#include "typeset_obj.h"

static void scale_size (GlyphBox *gbox, double *advance, double *offset, Rect *real_size, Rect *typo_size);
static void glyphbox_bitblt (GlyphBox *gbox, FT_BitmapGlyph glyph, FT_Glyph_Metrics *metrics, BufferContainer *buf, gboolean use_cache);
static void embolden (FT_GlyphSlot glyphslot, double bold_x, double bold_y);

// IuWFNg
GlyphBox* glyphbox_create (TypesetObj *parent)
{/*{{{*/
	GlyphBox *gbox = g_slice_alloc(sizeof(GlyphBox));
	OBJ_INIT(gbox,
			 parent,
			 TYPESET_OBJ_TYPE_GLYPH,
			 parent ? parent->typesetter : NULL,
			 DIR_TYPE_TTB,
			 glyphbox_destroy,
			 glyphbox_reset_style,
			 glyphbox_set_style,
			 glyphbox_set_dir,
			 glyphbox_set_dot,
			 glyphbox_set_bold,
			 glyphbox_set_color,
			 glyphbox_get_glyph_id,
			 glyphbox_combine_glyph,
			 glyphbox_cache_glyph_image,
			 glyphbox_compute_gbox_size, 
			 glyphbox_compute_self_size,
			 glyphbox_get_size,
			 glyphbox_render);
	gbox->combining_sequence = NULL;
	return gbox;
}/*}}}*/

void glyphbox_destroy (GlyphBox *gbox) 
{/*{{{*/
	GList *p;

	p = g_list_first(gbox->combining_sequence);
	while (p) {
		((TypesetObj*)p->data)->func_destroy(p->data);
		p = g_list_next(p);
	}
	g_list_free (p);

	g_slice_free1 (sizeof(GlyphBox), gbox);
}/*}}}*/

void glyphbox_reset_style (GlyphBox *gbox)
{/*{{{*/
	static int prev_style = FONT_STYLE_BASE;
	if (gbox == NULL) {
		prev_style = FONT_STYLE_BASE;
		return;
	}

	gbox->char_class = get_char_class (gbox->unicode);
	switch (gbox->char_class) {
		case CHAR_CLASS_ALPHABET:
		case CHAR_CLASS_NUMBER:
		case CHAR_CLASS_EXCLAM:
			if (gbox->obj.typesetter->face_table[FONT_STYLE_ASCII].face != NULL) {
				gbox->obj.style = FONT_STYLE_ASCII;
			} else {
				gbox->obj.style = FONT_STYLE_BASE;
			}
			gbox->rotate = TRUE;
			break;
		case CHAR_CLASS_YAKU_KUTEN:
		case CHAR_CLASS_KANA:
			if (gbox->obj.typesetter->face_table[FONT_STYLE_KANA].face != NULL) {
				gbox->obj.style = FONT_STYLE_KANA;
			} else {
				gbox->obj.style = FONT_STYLE_BASE;
			}
			gbox->rotate = FALSE;
			break;
		case CHAR_CLASS_COMBINING:
			gbox->obj.style = prev_style;
			gbox->rotate = FALSE;
			break;
		default:
			gbox->obj.style = FONT_STYLE_BASE;
			gbox->rotate = FALSE;
			break;
	}
	prev_style = gbox->obj.style;
}/*}}}*/

void glyphbox_set_style (GlyphBox *gbox, int style)
{/*{{{*/
	gbox->obj.style = style;
}/*}}}*/

void glyphbox_set_dir (GlyphBox *gbox, DirType dir)
{/*{{{*/
	gbox->obj.dir = dir;
}/*}}}*/

void glyphbox_set_dot (GlyphBox *gbox, DotType dot)
{/*{{{*/
	if (gbox->obj.typesetter->option.dot_skip_yaku) {
		switch (gbox->char_class) {
			case CHAR_CLASS_YAKU_START:
			case CHAR_CLASS_YAKU_CENTER:
			case CHAR_CLASS_YAKU_END:
			case CHAR_CLASS_YAKU_KUTEN:
				gbox->obj.dot = DOT_TYPE_NONE;
				return;
		}
		if (gbox->unicode == 0x0020 ||
			gbox->unicode == 0x3000) {
				gbox->obj.dot = DOT_TYPE_NONE;
				return;
		}
	}
	gbox->obj.dot = dot;
}/*}}}*/

void glyphbox_set_bold (GlyphBox *gbox, double bold_x, double bold_y)
{/*{{{*/
	gbox->obj.bold_x += bold_x;
	gbox->obj.bold_y += bold_y;
}/*}}}*/

void glyphbox_set_color (GlyphBox *gbox, Color color)
{/*{{{*/
	gbox->obj.color = color;
}/*}}}*/

void glyphbox_get_glyph_id (GlyphBox *gbox)
{/*{{{*/
	FT_UInt hgid, vgid;
	FT_Face	face;
	VertCacheTable	*vcache_table;

	if (gbox->obj.type != TYPESET_OBJ_TYPE_GLYPH) return;

	face			= gbox->obj.typesetter->face_table[gbox->obj.style].face;
	vcache_table	= gbox->obj.typesetter->face_table[gbox->obj.style].vcache_table;

	gbox->code = unicode_to_charmap_code(face, gbox->unicode);

	hgid = FT_Get_Char_Index(face, gbox->code);
	if (gbox->variant_selector) {
		FT_UInt gid = FT_Face_GetCharVariantIndex(face, gbox->code, gbox->variant_selector);
		if (gid) {
			hgid = gid;
		}
	}

	if (gbox->obj.dir == DIR_TYPE_TTB) {
		vgid = vert_cache_table_get_glyph_id(vcache_table, hgid);
		
		gbox->is_vert_glyph = vgid ? TRUE : FALSE;
		gbox->glyph_id = vgid ? vgid : hgid;

		//rotate the glyph if U+3030(wavy dash) have no vertical glyph.
		if (gbox->unicode == 0x3030) {
			if (gbox->is_vert_glyph == FALSE) {
				gbox->rotate = TRUE;
			}
		}
	} else {
		vgid = 0;
		gbox->is_vert_glyph = FALSE;
		gbox->rotate = TRUE;
		gbox->glyph_id = hgid;
	}

	DBGMSG("getglyph:");
	DBGMSG("dir:%d, ", gbox->obj.dir);
	DBGMSG("glyphid=0x%04x, ", gbox->glyph_id);
	DBGMSG("vs=0x%04x, ", gbox->variant_selector);
	DBGMSG("hgid=0x%04x, vgid=0x%04x ", hgid, vgid);
	DBGMSG("is_vert= %d\n", gbox->is_vert_glyph);
	return;
}/*}}}*/

void glyphbox_combine_glyph (GlyphBox *gbox)
{/*{{{*/
	//Do nothing!
	//combine-glyph function is called in 
	//parent object frame.
}/*}}}*/

void glyphbox_cache_glyph_image (GlyphBox *gbox)
{/*{{{*/
	FT_Error error;
	FT_Face	face;
	VertCacheTable	*vcache_table;
	GlyphCacheTable	*gcache_table;
	GlyphCache *gcache = NULL;

	FT_Int32 render_flag = gbox->obj.fontflag.render;
	FT_Int32 load_flag   = gbox->obj.fontflag.load;
	FT_UInt  gid = gbox->glyph_id;

	face			= gbox->obj.typesetter->face_table[gbox->obj.style].face;
	vcache_table	= gbox->obj.typesetter->face_table[gbox->obj.style].vcache_table;
	gcache_table	= gbox->obj.typesetter->face_table[gbox->obj.style].gcache_table;

	if (gbox->is_vert_glyph) {
		load_flag |= FT_LOAD_VERTICAL_LAYOUT;
	}
	gbox->obj.fontflag.render = render_flag;
	gbox->obj.fontflag.load   = load_flag;

	gcache = glyph_cache_table_get(gcache_table, gid);
	if (gcache == NULL) {
		//cache outline image
		error = FT_Load_Glyph(face, gid, load_flag);
		if (error) goto next;
		
next:
		glyph_cache_table_set(gcache_table, gid, face->glyph, gbox->is_vert_glyph);
	}
}/*}}}*/

void glyphbox_compute_gbox_size (GlyphBox *gbox, TypesetObj *prev, TypesetObj *next)
{/*{{{*/
	FT_Face			face;
	VertCacheTable	*vcache_table;
	GlyphCacheTable	*gcache_table;
	GlyphCache *gcache;
	double height;
	FT_Size_Metrics  *size_metrics;
	int yppem, xppem;
	gboolean use_kerning = gbox->obj.typesetter->option.kerning;
	double kerning = 0;

	face			= gbox->obj.typesetter->face_table[gbox->obj.style].face;
	vcache_table	= gbox->obj.typesetter->face_table[gbox->obj.style].vcache_table;
	gcache_table	= gbox->obj.typesetter->face_table[gbox->obj.style].gcache_table;
	gcache = glyph_cache_table_get(gcache_table, gbox->glyph_id);
	size_metrics = &(face->size->metrics);
	yppem = size_metrics->y_ppem;
	xppem = size_metrics->x_ppem;

	if (use_kerning && FT_HAS_KERNING(face) && gbox->glyph_id && prev &&
		prev->type == TYPESET_OBJ_TYPE_GLYPH &&
		prev->style == gbox->obj.style &&
		((GlyphBox*)prev)->glyph_id != 0) {
		FT_Vector delta;
		FT_Get_Kerning(face, ((GlyphBox*)prev)->glyph_id, gbox->glyph_id, 
							  FT_KERNING_DEFAULT, &delta);
		kerning = F26D6_TO_DOUBLE(delta.x);
	}

	if (gbox->obj.dir == DIR_TYPE_TTB &&  gbox->rotate) {
		if (gbox->is_vert_glyph) {
			//Not come here. vert glyph must not rotate. even if gbox->rotate == TRUE
			gbox->rotate = FALSE;
			gbox->width = xppem;
			gbox->height = F26D6_TO_DOUBLE(gcache->metrics.vertAdvance);
			gbox->offset = kerning;
			gbox->advance = gbox->height;
		} else {
			gbox->width = F26D6_TO_DOUBLE(size_metrics->height);
			gbox->height = F26D6_TO_DOUBLE(gcache->metrics.horiAdvance);
			gbox->offset = kerning;
			gbox->advance = gbox->height;
		}
	}
	if (gbox->obj.dir == DIR_TYPE_TTB && !gbox->rotate) {
		if (gbox->is_vert_glyph) {
			gbox->width = xppem;
			gbox->height = F26D6_TO_DOUBLE(gcache->metrics.vertAdvance);
			gbox->offset = kerning;
			gbox->advance = gbox->height;
		} else {
			gbox->width = F26D6_TO_DOUBLE(gcache->metrics.horiAdvance);
			gbox->height = yppem;
			gbox->offset = kerning;
			gbox->advance = gbox->height;
		}
	}
	if (gbox->obj.dir == DIR_TYPE_LTR &&  gbox->rotate) {
		gbox->rotate = FALSE;
	}
	if (gbox->obj.dir == DIR_TYPE_LTR && !gbox->rotate) {
		gbox->width = F26D6_TO_DOUBLE(gcache->metrics.horiAdvance);
		gbox->height = yppem;
		gbox->offset = kerning;
		gbox->advance = gbox->width;
	}

	height = gbox->height;
	switch (gbox->obj.typesetter->option.tsume_type) {
		case BETA:
			break;
		case TSUME:
			switch (gbox->char_class) {
				case CHAR_CLASS_YAKU_KUTEN:
				case CHAR_CLASS_YAKU_END:
					gbox->offset = 0;
					gbox->advance = height/2;
					break;
				case CHAR_CLASS_YAKU_START:
					gbox->offset = - height/2;
					break;
				case CHAR_CLASS_YAKU_CENTER:
					gbox->offset = - height/4;
					gbox->advance  = height - height/4;
					break;
			}
			break;
		case JISX4051:
		{
			double offset, advance;
			int prev_type = prev ? ((GlyphBox*)prev)->char_class : CHAR_CLASS_NONE;
			int now_type  = gbox->char_class;
			int next_type = next ? ((GlyphBox*)next)->char_class : CHAR_CLASS_NONE;

			get_yaku_tsume_jisx4051 (prev_type, now_type, next_type, &offset, &advance);
			if (gbox->obj.dir == DIR_TYPE_TTB) {
				gbox->offset  = height * offset;
				gbox->advance = height * advance;
			} else {
				gbox->offset  = gbox->width * offset;
				gbox->advance = gbox->width * advance;
			}
		}
		break;
	}
}/*}}}*/

void glyphbox_compute_self_size (GlyphBox *gbox)
{/*{{{*/
	//Do nothing!
	// Glyphbox size is already computed when
	// glyphbox_compute_gbox_size was called.
}/*}}}*/

void glyphbox_get_size (GlyphBox *gbox, double *advance, double *offset, Rect *real_size, Rect *typo_size)
{/*{{{*/
	FT_Face	face;
	GlyphCacheTable	*gcache_table;
	GlyphCache *gcache = NULL;
	FT_Glyph glyph;
	FT_Glyph_Metrics glyph_metrics;
	double ascender, descender, linegap;
	const Rect zero_size = {0, 0, 0, 0};

	face			= gbox->obj.typesetter->face_table[gbox->obj.style].face;
	gcache_table	= gbox->obj.typesetter->face_table[gbox->obj.style].gcache_table;
	gcache			= glyph_cache_table_get(gcache_table, gbox->glyph_id);



	if (gbox->obj.bold_x != 0 || gbox->obj.bold_y != 0) {
		FT_Load_Glyph(face, gbox->glyph_id, gbox->obj.fontflag.load);
		embolden(face->glyph, gbox->obj.bold_x, gbox->obj.bold_y);
		FT_Render_Glyph(face->glyph, FT_RENDER_MODE_MONO);
		FT_Get_Glyph(face->glyph, &glyph);
		glyph_metrics = face->glyph->metrics;
	} else {
		glyph = (FT_Glyph)gcache->glyph;
		FT_Glyph_To_Bitmap(&glyph, FT_RENDER_MODE_MONO, 0, 0);
		glyph_metrics = gcache->metrics;
	}

	get_typo_metrics(face, &ascender, &descender, &linegap);

	if (gbox->obj.dir == DIR_TYPE_TTB) {
		//cg݂̏ꍇ̃Ot͊_㒆ƂB
		//}CiXAEvXAオ}CiXAvXƂB
		if (gbox->rotate == FALSE) {
			//FIXME: vertBearingY of some otf fonts may be wrong.
			/*
			real_size->top    =  F26D6_TO_DOUBLE(glyph_metrics.vertBearingY);
			real_size->bottom =  F26D6_TO_DOUBLE(glyph_metrics.vertBearingY + glyph_metrics.height);
			*/
			real_size->top    =  ascender - ((FT_BitmapGlyph)glyph)->top;
			real_size->bottom =  real_size->top + ((FT_BitmapGlyph)glyph)->bitmap.rows;
			real_size->left   =  F26D6_TO_DOUBLE(glyph_metrics.vertBearingX);
			real_size->right  =  F26D6_TO_DOUBLE(glyph_metrics.vertBearingX + glyph_metrics.width);
			typo_size->top    =  0;
			typo_size->bottom =  F26D6_TO_DOUBLE(glyph_metrics.vertAdvance);
			typo_size->left   = -F26D6_TO_DOUBLE(glyph_metrics.horiAdvance) / 2.0;
			typo_size->right  =  F26D6_TO_DOUBLE(glyph_metrics.horiAdvance) / 2.0;
			*advance          =  gbox->advance;
			*offset           =  gbox->offset;
		} else {
			real_size->top    =  F26D6_TO_DOUBLE(glyph_metrics.horiBearingX);
			real_size->bottom =  F26D6_TO_DOUBLE(glyph_metrics.horiBearingX + glyph_metrics.width);
			real_size->left   =  -(double)face->size->metrics.y_ppem / 2.0 - descender + F26D6_TO_DOUBLE(glyph_metrics.horiBearingY - glyph_metrics.height);
			real_size->right  =  -(double)face->size->metrics.y_ppem / 2.0 - descender + F26D6_TO_DOUBLE(glyph_metrics.horiBearingY);
			typo_size->top    =  0;
			typo_size->bottom =  F26D6_TO_DOUBLE(glyph_metrics.horiAdvance);
			typo_size->left   = -face->size->metrics.y_ppem / 2.0;
			typo_size->right  =  face->size->metrics.y_ppem / 2.0;
			*advance          = gbox->advance;
			*offset           = gbox->offset;
		}

		if (gbox->char_class == CHAR_CLASS_COMBINING &&
			(gbox->unicode == 0x3099 || gbox->unicode == 0x309a ||
			 gbox->unicode == 0x309b || gbox->unicode == 0x309c)) {
				if (gcache->glyph_id == 0) {
					*real_size = zero_size;
					*typo_size = zero_size;
					*advance = 0;
					*offset  = 0;
					return;
				}
				real_size->top    = 0;
				real_size->bottom = F26D6_TO_DOUBLE(glyph_metrics.height);
				real_size->left   = F26D6_TO_DOUBLE(glyph_metrics.horiAdvance - glyph_metrics.width - glyph_metrics.horiAdvance / 2);
				real_size->right  = F26D6_TO_DOUBLE(glyph_metrics.horiAdvance/2);
				*typo_size = zero_size;
				*advance = 0;
				*offset  = 0;
		}
	} else {
		//g݂̏ꍇ̃Ot͊_̃x[XCƂB
		//}CiXAEvXAオ}CiXAvXƂB
		//gNXascender, descenderƂ̓vX}CiXtɂȂ̂ŒӁB
		real_size->top    = -F26D6_TO_DOUBLE(glyph_metrics.horiBearingY);
		real_size->bottom = -F26D6_TO_DOUBLE(glyph_metrics.horiBearingY - glyph_metrics.height);
		real_size->left   =  F26D6_TO_DOUBLE(glyph_metrics.horiBearingX);
		real_size->right  =  F26D6_TO_DOUBLE(glyph_metrics.horiBearingX + glyph_metrics.width);
		typo_size->top    =  -ascender;
		typo_size->bottom =  -descender;
		typo_size->left   =  0;
		typo_size->right  =  F26D6_TO_DOUBLE(glyph_metrics.horiAdvance);
		*advance          =  gbox->advance;
		*offset           =  gbox->offset;

		if (gbox->char_class == CHAR_CLASS_COMBINING &&
			(gbox->unicode == 0x3099 || gbox->unicode == 0x309a ||
			 gbox->unicode == 0x309b || gbox->unicode == 0x309c)) {
				if (gcache->glyph_id == 0) {
					*real_size = zero_size;
					*typo_size = zero_size;
					*advance = 0;
					*offset  = 0;
					return;
				}
				real_size->top    = -ascender;
				real_size->bottom = -descender - F26D6_TO_DOUBLE(glyph_metrics.height);
				real_size->left   = F26D6_TO_DOUBLE(glyph_metrics.horiAdvance - glyph_metrics.width);
				real_size->right  = F26D6_TO_DOUBLE(glyph_metrics.horiAdvance);
				*typo_size = zero_size;
				*advance = 0;
				*offset  = 0;
		}
	}
	
	scale_size(gbox, advance, offset, real_size, typo_size);
}/*}}}*/

void glyphbox_render (GlyphBox *gbox, BufferContainer *buf)
{/*{{{*/
	double scale_x, scale_y;
	double bold_x, bold_y;
	DBGMSG("render glyph: U+%04x, class=%2d, x=%6.1f, y=%6.1f\n", gbox->unicode, gbox->char_class, buf->x, buf->y);

	if (gbox->glyph_id == 0) return;

	get_scale((TypesetObj*)gbox, &scale_x, &scale_y);
	bold_x = gbox->obj.bold_x;
	bold_y = gbox->obj.bold_y;

	if (scale_x != 1 || scale_y != 1 || bold_x != 0 || bold_y != 0) {
		FT_Glyph glyph;
		double tmp_width, tmp_height;
		FT_Face face = gbox->obj.typesetter->face_table[gbox->obj.style].face;
		FontStyle default_style = *gbox->obj.typesetter->face_table[gbox->obj.style].style;
		FontStyle style = default_style;
		style.size_x = (FT_F26Dot6)(style.size_x * scale_x);
		style.size_y = (FT_F26Dot6)(style.size_y * scale_y);

		face_set_size (face, &style);
		FT_Load_Glyph(face, gbox->glyph_id, gbox->obj.fontflag.load);
		embolden(face->glyph, bold_x, bold_y);
		FT_Render_Glyph(face->glyph, gbox->obj.fontflag.render);
		FT_Get_Glyph(face->glyph, &glyph);

		tmp_width  = gbox->width;
		tmp_height = gbox->height;
		gbox->width  *= scale_x;
		gbox->height *= scale_y;

		glyphbox_bitblt(gbox, (FT_BitmapGlyph)glyph, &(face->glyph->metrics), buf, FALSE);
		gbox->width  = tmp_width;
		gbox->height = tmp_height;

		face_set_size (face, &default_style);
	} else {
		GlyphCache *gcache;
		gcache = glyph_cache_table_get(gbox->obj.typesetter->face_table[gbox->obj.style].gcache_table, gbox->glyph_id);

		if (gcache->glyph->format != FT_GLYPH_FORMAT_BITMAP) {
			FT_Glyph_To_Bitmap(&(gcache->glyph), gbox->obj.fontflag.render, 0, 0);
		}
		glyphbox_bitblt(gbox, (FT_BitmapGlyph)gcache->glyph, &(gcache->metrics), buf, TRUE);
	}

}/*}}}*/

//transfer bitmap to buffer
void glyphbox_bitblt (GlyphBox *gbox, FT_BitmapGlyph glyph, FT_Glyph_Metrics *metrics, BufferContainer *buf, gboolean use_cache)
{/*{{{*/
	gboolean			antialias;
	FT_Face 			face            = gbox->obj.typesetter->face_table[gbox->obj.style].face;
	TT_OS2				*os2            = gbox->obj.typesetter->face_table[gbox->obj.style].os2;
	FT_Size_Metrics		*size_metrics;
	int yppem, xppem;
	double descender, ascender;
	int dx, dy;
	double height, width;
	double base_line;

	size_metrics	= &(face->size->metrics);
	yppem			= size_metrics->y_ppem;
	xppem			= size_metrics->x_ppem;
	ascender	= (double)yppem * os2->sTypoAscender  / face->units_per_EM;
	descender	= (double)yppem * os2->sTypoDescender / face->units_per_EM;
	height	= gbox->height;
	width	= gbox->width; 

	if (use_cache) {
		base_line = gbox->obj.typesetter->face_table[gbox->obj.style].base_line;
	} else {
		base_line = face_get_base_line(face);
	}

	//memset(glyph->bitmap.buffer, 0xFF, sizeof(char) * glyph->bitmap.rows * glyph->bitmap.width);
#ifdef DEBUG
	{ //Draw a base line for debug
		int x, y;
		int base_x, base_y;
		int guide_w, guide_h;
		
		if (gbox->obj.dir == DIR_TYPE_TTB) {
			if (gbox->rotate == FALSE) { 
				base_x = (int)buf->x;
				base_y = (int)(buf->y < 0 ? 0 : buf->y);
				guide_w = xppem;
				guide_h = (int)height;
			} else {
				base_x = (int)(buf->x - (ascender - descender) / 2 + base_line);
				base_y = (int)(buf->y < 0 ? 0 : buf->y);
				guide_w = yppem;
				guide_h = (int)(height * gbox->obj.scale_x);
			}
			//base line
			for (y = 0; y < guide_h; y++) {
				int pos = base_x + (buf->width * (base_y+y));
				if (0 <= pos && pos < buf->width * buf->height) {
					pos *= buf->planes;
					if (buf->planes == 4) {
						buf->data[pos+0] = 0x60;
						buf->data[pos+1] = 0x60;
						buf->data[pos+2] = 0x60;
						buf->data[pos+3] = 0xFF;
					}
					if (buf->planes == 2) {
						buf->data[pos+0] = 0x60;
						buf->data[pos+1] = 0xFF;
					}
				}
			}

			// top side
			if (buf->y >= 0) {
				base_x = (int)(buf->x - guide_w / 2);
				base_y = (int)buf->y;
				for (x =0; x < guide_w; x++) {
					int pos = x + base_x + (base_y * buf->width);
					if (0 <= pos && pos < buf->width * buf->height) {
						pos *= buf->planes;
						if (buf->planes == 4) {
							buf->data[pos+0] = 0x60;
							buf->data[pos+1] = 0x60;
							buf->data[pos+2] = 0x60;
							buf->data[pos+3] = 0xFF;
						}
						if (buf->planes == 2) {
							buf->data[pos+0] = 0x60;
							buf->data[pos+1] = 0xFF;
						}
					}
				}
			}
		}
		if (gbox->obj.dir == DIR_TYPE_LTR) {
			base_x = (int)buf->x;
			base_y = (int)(buf->y < 0 ? 0 : buf->y);
			//base line
			for (x =0; x < width; x++) {
				int pos = x + base_x + (base_y * buf->width);
				if (0 <= pos && pos < buf->width * buf->height) {
					pos *= buf->planes;
					if (buf->planes == 4) {
						buf->data[pos+0] = 0x60;
						buf->data[pos+1] = 0x60;
						buf->data[pos+2] = 0x60;
						buf->data[pos+3] = 0xFF;
					}
					if (buf->planes == 2) {
						buf->data[pos+0] = 0x60;
						buf->data[pos+1] = 0xFF;
					}
				}
			}

			// top side
			base_x = (int)buf->x;
			base_y = (int)(buf->y - ascender);
			for (y = 0; y < height; y++) {
				int pos = base_x + (buf->width * (base_y+y));
				if (0 <= pos && pos < buf->width*buf->height) {
					pos *= buf->planes;
					if (buf->planes == 4) {
						buf->data[pos+0] = 0x60;
						buf->data[pos+1] = 0x60;
						buf->data[pos+2] = 0x60;
						buf->data[pos+3] = 0xFF;
					}
					if (buf->planes == 2) {
						buf->data[pos+0] = 0x60;
						buf->data[pos+1] = 0xFF;
					}
				}
			}
		}

	}
#endif

	switch (glyph->bitmap.pixel_mode) {
		case FT_PIXEL_MODE_MONO: antialias = FALSE; break;
		case FT_PIXEL_MODE_GRAY: antialias = TRUE;  break;
		default: goto done;
	}

	if (gbox->rotate == FALSE) {
		if (gbox->obj.dir == DIR_TYPE_LTR) {
			dx = (int)buf->x + F26D6_TO_INT32(metrics->horiBearingX);
			dy = (int)buf->y - F26D6_TO_INT32(metrics->horiBearingY);
		} else {
			dx = (int)(buf->x - width / 2 + glyph->left);
			//FIXME: vertBearingY of some otf fonts may be wrong.
			//dy = (int)(buf->y + F26D6_TO_INT32(metrics->vertBearingY));
			dy = (int)(buf->y + (ascender - glyph->top));

			while (dy + glyph->bitmap.rows > buf->height) {
				dy--;
			}
		}
	} else {
		dx = (int)(buf->x - (ascender - descender) / 2 + base_line + F26D6_TO_DOUBLE(metrics->horiBearingY - metrics->height));
		dy = (int)(buf->y + F26D6_TO_INT32(metrics->horiBearingX));

		if (dx + glyph->bitmap.rows > buf->width) {
			dx = buf->width - glyph->bitmap.rows;
		}
	}

	if (gbox->char_class == CHAR_CLASS_COMBINING) {
		// Japanese Sound Marks (move to right-top corner)
		if (gbox->unicode == 0x3099 || gbox->unicode == 0x309a ||
			gbox->unicode == 0x309b || gbox->unicode == 0x309c) {
				switch (gbox->obj.dir) {
					case DIR_TYPE_TTB:
						dx = (int)(buf->x + width / 2 - glyph->bitmap.width);
						dy = (int)buf->y;
						break;
					case DIR_TYPE_LTR:
						dx = (int)(buf->x + width - glyph->bitmap.width);
						dy = (int)(buf->y - ascender);
						break;
					default:
						g_assert_not_reached();
			}
		}
	}

	if (dy < 0) dy = 0;
	if (dx < 0) dx = 0;

	copy_glyph_image (gbox->obj.color,
					  gbox->rotate,
					  antialias, 
					  buf->data, 
					  dx,
					  dy,
					  buf->width,
					  buf->height,
					  buf->planes,
					  glyph->bitmap.buffer,
					  glyph->bitmap.width,
					  glyph->bitmap.rows,
					  glyph->bitmap.pitch);
done:
	if (!use_cache) {
		FT_Done_Glyph((FT_Glyph)glyph);
	}
	return;
}/*}}}*/

GList* glyphbox_render_combining (GList *p, double base_advance, BufferContainer *buf)
{
	double tmp_x, tmp_y;
	tmp_x = buf->x;
	tmp_y = buf->y;

	p = g_list_next(p);
	while (p) {
		double ad, of;
		Rect r_rect, t_rect;
		double scale_x, scale_y;
		TypesetObj *obj = p->data;
		GlyphBox *gbox = p->data;
		if (obj->type != TYPESET_OBJ_TYPE_GLYPH) break;
		if (gbox->char_class != CHAR_CLASS_COMBINING) break;

		gbox->obj.func_get_size(gbox, &ad, &of, &r_rect, &t_rect);
		get_scale((TypesetObj*)gbox, &scale_x, &scale_y);

		switch(obj->dir) {
			//case DIR_TYPE_TTB: buf->y += base_advance - gbox->height * scale_y; break;
			case DIR_TYPE_LTR: buf->x += base_advance - gbox->width * scale_x; break;
			case DIR_TYPE_RTL: g_assert_not_reached() ; break;
		}
		gbox->obj.func_render(gbox, buf);
		p = g_list_next(p);
	}
	buf->x = tmp_x;
	buf->y = tmp_y;
	return p;
}

static void scale_size (GlyphBox *gbox, double *advance, double *offset, Rect *real_size, Rect *typo_size)
{/*{{{*/
	double scale_x, scale_y;
	get_scale ((TypesetObj*)gbox, &scale_x, &scale_y);

	if (gbox->obj.dir == DIR_TYPE_TTB) {
		real_size->top    *= scale_y;
		real_size->bottom *= scale_y;
		real_size->left   *= scale_x;
		real_size->right  *= scale_x;
		typo_size->top    *= scale_y;
		typo_size->bottom *= scale_y;
		typo_size->left   *= scale_x;
		typo_size->right  *= scale_x;
		*advance *= scale_y;
		*offset  *= scale_y;
	} else {
		real_size->top    *= scale_y;
		real_size->bottom *= scale_y;
		real_size->left   *= scale_x;
		real_size->right  *= scale_x;
		typo_size->top    *= scale_y;
		typo_size->bottom *= scale_y;
		typo_size->left   *= scale_x;
		typo_size->right  *= scale_x;
		*advance *= scale_x;
		*offset  *= scale_x;
	}
}/*}}}*/

static void embolden (FT_GlyphSlot glyphslot, double bold_x, double bold_y)
{/*{{{*/
	FT_F26Dot6 str_x, str_y;
	FT_Pos off_x, off_y;

	if (bold_x == 0 && bold_y == 0) {
		return;
	}
	str_x = DOUBLE_TO_F26D6(glyphslot->face->size->metrics.x_ppem * bold_x * 0.01);
	str_y = DOUBLE_TO_F26D6(glyphslot->face->size->metrics.y_ppem * bold_y * 0.01);
	off_x = -str_x >> 1;
	off_y = -str_y >> 1;

	if (glyphslot->format == FT_GLYPH_FORMAT_OUTLINE) {
		FT_BBox bbox1, bbox2, bbox3;

		FT_Outline_Get_BBox(&glyphslot->outline, &bbox1);
		/*
		Freetype2.4.2 it was used by OSX build does not have FT_Outline_EmboldenXY.
		FT_Outline_EmboldenXY(&glyphslot->outline, str_x, str_y);
		*/
		FT_Outline_Embolden(&glyphslot->outline, str_x);
		
		FT_Outline_Get_BBox(&glyphslot->outline, &bbox2);

		FT_Outline_Translate(&glyphslot->outline, off_x, off_y);
		FT_Outline_Get_BBox(&glyphslot->outline, &bbox3);

		glyphslot->metrics.width  = bbox2.xMax - bbox2.xMin;
		glyphslot->metrics.height = bbox2.yMax - bbox2.yMin;
		
		glyphslot->metrics.horiBearingX += off_x;
		glyphslot->metrics.horiBearingY -= off_y;
		
		glyphslot->metrics.vertBearingX += off_x;
		glyphslot->metrics.vertBearingY -= off_y;
	}
	if (glyphslot->format == FT_GLYPH_FORMAT_BITMAP) {
		FT_GlyphSlot_Own_Bitmap(glyphslot);
		FT_Bitmap_Embolden(glyphslot->library, &glyphslot->bitmap, str_x, str_y);
		
		glyphslot->bitmap_left += F26D6_TO_INT32(off_x);
		glyphslot->bitmap_top  -= F26D6_TO_INT32(off_y);

		glyphslot->metrics.width  = INT32_TO_F26D6(glyphslot->bitmap.width);
		glyphslot->metrics.height = INT32_TO_F26D6(glyphslot->bitmap.rows);
		
		glyphslot->metrics.horiBearingX += off_x;
		glyphslot->metrics.horiBearingY -= off_y;
		
		glyphslot->metrics.vertBearingX += off_x;
		glyphslot->metrics.vertBearingY -= off_y;
	}
	return;
}/*}}}*/

/* vim: set fdm=marker ts=4 sw=4:*/
