/*
 * GIMP Tategaki Plug-in -- This is a plug-in for The GIMP 2
 * Copyright (C) 2006-2017 OHTSUKA, Yoshio <ohtsuka.yoshio@gmail.com>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 */
#include "typeset_obj_layout.h"
#include "typeset_obj_glyphbox.h"
#include "typeset_obj_line.h"

static double dmax(int argc, ...);
static double dmin(int argc, ...);
static double dsum(int argc, ...);
static double get_default_line_pitch (Layout *layout);

Layout *layout_create (Typesetter *typ, TypesetObj *parent)
{/*{{{*/
	GList *p;
	Layout *layout = NULL;
	layout = g_malloc0(sizeof(Layout));
	OBJ_INIT(layout,
			 parent,
			 TYPESET_OBJ_TYPE_LAYOUT,
			 typ,
			 DIR_TYPE_RTL,
			 layout_destroy,
			 layout_reset_style, 
			 layout_set_style,
			 layout_set_dir,
			 layout_set_dot,
			 layout_set_bold,
			 layout_set_color,
			 layout_get_glyph_id,
			 layout_combine_glyph,
			 layout_cache_glyph_image,
			 layout_compute_gbox_size,
			 layout_compute_self_size,
			 layout_get_size,
			 layout_render);
	layout->lines = NULL;

	p = layout->lines;
	while (p) {
		((TypesetObj*)p->data)->parent = (TypesetObj*)layout;
		p = g_list_next(p);
	}
	return layout;
}/*}}}*/

void layout_destroy (Layout *layout)
{/*{{{*/
	GList *p;
	p = g_list_first (layout->lines);
	while (p) {
		((TypesetObj*)p->data)->func_destroy(p->data);
		p = g_list_next(p);
	}
	g_list_free (p);

	g_free(layout);
}/*}}}*/

void layout_reset_style(Layout *layout)
{/*{{{*/
	GList *p = g_list_first (layout->lines);
	while (p) {
		((TypesetObj*)p->data)->func_reset_style(p->data);
		p = g_list_next(p);
	}
}/*}}}*/

void layout_set_style(Layout *layout, int style)
{/*{{{*/
	GList *p = g_list_first (layout->lines);
	while (p) {
		((TypesetObj*)p->data)->func_set_style(p->data, style);
		p = g_list_next(p);
	}
	layout->obj.style = style;
}/*}}}*/

void layout_set_dir(Layout *layout, DirType dir)
{/*{{{*/
	GList *p;

	layout->obj.dir = dir;
	switch (dir) {
		case DIR_TYPE_RTL: dir = DIR_TYPE_TTB; break;
		case DIR_TYPE_TTB: dir = DIR_TYPE_LTR; break;
		case DIR_TYPE_LTR: dir = DIR_TYPE_TTB; break;
	}

	p = g_list_first (layout->lines);
	while (p) {
		((TypesetObj*)p->data)->func_set_dir(p->data, dir);
		p = g_list_next(p);
	}
}/*}}}*/

void layout_set_dot(Layout *layout, DotType dot)
{/*{{{*/
	GList *p = g_list_first (layout->lines);
	while (p) {
		((TypesetObj*)p->data)->func_set_dot(p->data, dot);
		p = g_list_next(p);
	}
	layout->obj.dot = DOT_TYPE_NONE;
}/*}}}*/

void layout_set_bold(Layout *layout, double bold_x, double bold_y)
{/*{{{*/
	GList *p = g_list_first (layout->lines);
	while (p) {
		((TypesetObj*)p->data)->func_set_bold(p->data, bold_x, bold_y);
		p = g_list_next(p);
	}
	layout->obj.bold_x += bold_x;
	layout->obj.bold_y += bold_y;
}/*}}}*/

void layout_set_color(Layout *layout, Color color)
{/*{{{*/
	GList *p = g_list_first (layout->lines);
	while (p) {
		((TypesetObj*)p->data)->func_set_color(p->data, color);
		p = g_list_next(p);
	}
	layout->obj.color = color;
}/*}}}*/

void layout_get_glyph_id(Layout *layout)
{/*{{{*/
	GList *p = g_list_first (layout->lines);
	while (p) {
		((TypesetObj*)p->data)->func_get_glyph_id(p->data);
		p = g_list_next(p);
	}
}/*}}}*/

void layout_combine_glyph(Layout *layout)
{/*{{{*/
	GList *p;
	layout->lines = combine_glyph(layout->lines);
	p = g_list_first (layout->lines);
	while (p) {
		((TypesetObj*)p->data)->func_combine_glyph(p->data);
		p = g_list_next(p);
	}
}/*}}}*/

void layout_cache_glyph_image (Layout *layout)
{/*{{{*/
	GList *p;
	p = g_list_first (layout->lines);
	while (p) {
		((TypesetObj*)p->data)->func_cache_glyph_image(p->data);
		p = g_list_next(p);
	}
}/*}}}*/

void layout_compute_gbox_size(Layout *layout, TypesetObj *prev, TypesetObj *next)
{/*{{{*/
	GList *p = g_list_first (layout->lines);

	while (p) {
		((TypesetObj*)p->data)->func_compute_gbox_size(p->data,
													   p->prev?p->prev->data:NULL,
													   p->next?p->next->data:NULL);
		p = g_list_next(p);
	}
}/*}}}*/

void layout_compute_self_size(Layout *layout)
{/*{{{*/
	GList *p;
	double ad, of;
	Rect r_size, t_size;
	double max = 0;

	p = g_list_first (layout->lines);
	while (p) {
		TypesetObj *obj = p->data;
		obj->func_compute_self_size(p->data);
		obj->func_get_size(obj, &ad, &of, &r_size, &t_size);

		switch (layout->obj.dir) {
			case DIR_TYPE_RTL:
				switch (obj->dir) {
					case DIR_TYPE_TTB: max = MAX(max, t_size.bottom); break;
					case DIR_TYPE_RTL: max = MAX(max, t_size.top - t_size.bottom); break;
					case DIR_TYPE_LTR: max = MAX(max, t_size.top - t_size.bottom); break;
				}
				break;
			case DIR_TYPE_TTB:
				switch (obj->dir) {
					case DIR_TYPE_TTB: max = MAX(max, t_size.top - t_size.bottom); break;
					case DIR_TYPE_RTL: max = MAX(max, t_size.right); break;
					case DIR_TYPE_LTR: max = MAX(max, t_size.right); break;
				}
				break;
			case DIR_TYPE_LTR:
				g_assert_not_reached();
				break;
		}
		p = g_list_next(p);
	}

	p = g_list_first (layout->lines);
	while (p) {
		TypesetObj *obj = p->data;
		if (obj->type == TYPESET_OBJ_TYPE_LINE) {
			line_update_height_and_offset((Line*)obj, max);
		}
		p = g_list_next(p);
	}
}/*}}}*/

void layout_get_size (Layout *layout, double *advance, double *offset, Rect *real_size, Rect *typo_size)
{/*{{{*/
	GList *p;
	double linepitch;
	double ad, of;
	Rect r_size, t_size, prev_r_size, prev_t_size;
	Rect zero_size = {0,0,0,0};

	*advance   = 0;
	*offset    = 0;
	*real_size = zero_size;
	*typo_size = zero_size;

	linepitch = get_default_line_pitch(layout);
	prev_r_size = zero_size;
	prev_t_size = zero_size;

	p = g_list_first(layout->lines);
	while (p) {
		((TypesetObj*)p->data)->func_get_size(p->data, &ad, &of, &r_size, &t_size);

		switch (layout->obj.dir) {
			case DIR_TYPE_RTL:
				typo_size->top    = MIN(typo_size->top,    t_size.top);
				typo_size->bottom = MAX(typo_size->bottom, t_size.bottom);
				typo_size->left   = 0;
				typo_size->right += t_size.right - t_size.left;
				if (p->prev && linepitch > t_size.right - prev_t_size.left) {
					typo_size->right += linepitch - (t_size.right - prev_t_size.left);
				}
				real_size->top    = MIN(real_size->top,    r_size.top);
				real_size->bottom = MAX(real_size->bottom, r_size.bottom);
				real_size->left   = r_size.left - t_size.left; //FIXME:O̍s˂ʂĂOtꍇɑΉ?
				real_size->right  = typo_size->right - t_size.right + r_size.right;
				break;

			case DIR_TYPE_TTB:
				typo_size->top     = 0;
				typo_size->bottom += t_size.bottom - t_size.top;
				if (p->prev && linepitch > prev_t_size.bottom - t_size.top) {
					typo_size->bottom += linepitch - (prev_t_size.bottom - t_size.top);
				}
				typo_size->left    = MIN(typo_size->left,  t_size.left);
				typo_size->right   = MAX(typo_size->right, t_size.right);
				real_size->top     = MIN(real_size->top, r_size.top - t_size.top);
				real_size->bottom  = typo_size->bottom - t_size.bottom + r_size.bottom;
				real_size->left   = MIN(real_size->left,  r_size.left);
				real_size->right  = MAX(real_size->right, r_size.right);
				break;

			case DIR_TYPE_LTR:
				g_assert_not_reached();
		}
		prev_r_size = r_size;
		prev_t_size = t_size;
		p = g_list_next(p);
	}

	/*
	// top }CiX bottom vX left }CiX right vX
	p = g_list_first (layout->lines);
	while (p) {
		double ad, of;
		Rect r_size, t_size;
		TypesetObj *obj = p->data;

		obj->func_get_size(p->data, &ad, &of, &r_size, &t_size);
		
		if (layout->obj.dir == DIR_TYPE_RTL) {
			switch (obj->dir) {
				case DIR_TYPE_TTB:
					typo_size->top    = MIN(typo_size->top,    t_size.top);
					typo_size->bottom = MAX(typo_size->bottom, t_size.bottom);
					typo_size->left   = 0;
					typo_size->right += MAX(linepitch, t_size.right - t_size.left);
					real_size->top    = MIN(real_size->top,    r_size.top);
					real_size->bottom = MAX(real_size->bottom, r_size.bottom);
					real_size->left   = r_size.left - t_size.left; //FIXME:O̍s˂ʂĂOtꍇɑΉ?
					real_size->right  = typo_size->right - t_size.right + r_size.right;
					break;
				case DIR_TYPE_LTR:
				case DIR_TYPE_RTL:
					g_assert_not_reached();
					break;
			}
		} else if (layout->obj.dir == DIR_TYPE_TTB) {
			switch (obj->dir) {
				case DIR_TYPE_LTR:
					typo_size->top     = 0;
					typo_size->bottom += MAX(linepitch, t_size.bottom - t_size.top);
					typo_size->left    = MIN(typo_size->left,  t_size.left);
					typo_size->right   = MAX(typo_size->right, t_size.right);

					real_size->top     = MIN(real_size->top, r_size.top - t_size.top);
					real_size->bottom  = typo_size->bottom - t_size.bottom + r_size.bottom;
					real_size->left   = MIN(real_size->left,  r_size.left);
					real_size->right  = MAX(real_size->right, r_size.right);


					break;
				case DIR_TYPE_TTB:
				case DIR_TYPE_RTL:
					g_assert_not_reached();
					break;
			}
		} else {
			//DIR_TYPE_LTR
			g_assert_not_reached();
		}
		p = g_list_next(p);
	}
	*/
}/*}}}*/

void layout_render (Layout *layout, BufferContainer *buf)
{/*{{{*/
	double ad, of;
	Rect real_size, typo_size;
	double default_line_pitch;
	GList *p;

	switch (layout->obj.dir) {
		case DIR_TYPE_RTL:
			buf->x = buf->width;
			buf->y = 0;
			break;
		case DIR_TYPE_LTR:
		case DIR_TYPE_TTB:
			buf->x = 0;
			buf->y = 0;
			break;
	}
	default_line_pitch = get_default_line_pitch (layout);

	p = g_list_first(layout->lines);
	if (p == NULL) return;
	((TypesetObj*)p->data)->func_get_size(p->data, &ad, &of, &real_size, &typo_size);
	switch (layout->obj.dir) {
		case DIR_TYPE_RTL:
			buf->x -= MAX(typo_size.right, real_size.right);
			buf->y -= MIN(typo_size.top, real_size.top);
			break;
		case DIR_TYPE_TTB:
			buf->x -= MIN(typo_size.left, real_size.left);
			buf->y -= MIN(typo_size.top, real_size.top);
			break;
		case DIR_TYPE_LTR:
			g_assert_not_reached();
			break;
	}

	while (p) {
		Rect prev_typo_size;

		((TypesetObj*)p->data)->func_render(p->data, buf);

		prev_typo_size = typo_size;
		p = g_list_next(p);

		if (p) {
			((TypesetObj*)p->data)->func_get_size(p->data, &ad, &of, &real_size, &typo_size);
			switch (layout->obj.dir) {
				case DIR_TYPE_RTL: buf->x -= MAX(default_line_pitch, typo_size.right - prev_typo_size.left); break;
				case DIR_TYPE_TTB: buf->y += MAX(default_line_pitch, prev_typo_size.bottom - typo_size.top); break;
				case DIR_TYPE_LTR: g_assert_not_reached();
			}
		}
	}
}/*}}}*/

int layout_set_lines_from_sequence (Layout *layout, GList *in)
{/*{{{*/
	GList *p = in;
	GList *elems = NULL;
	GList *lines = NULL;

	if (layout->lines) {
		p = layout->lines;
		while (p) {
			((TypesetObj*)p->data)->func_destroy(p->data);
			p = g_list_next(p);
		}
		layout->lines = NULL;
	}

	p = in;
	while (p) {
		guint32 unicode = 0;
		TypesetObj *obj = NULL;
		GlyphBox *gbox = NULL;

		obj = (TypesetObj*)p->data;
		obj->parent = (TypesetObj*)layout;

		if (obj->type == TYPESET_OBJ_TYPE_GLYPH) {
			gbox = (GlyphBox*)p->data;
			unicode = gbox->unicode;
		}

		p = g_list_next(p);

		if (unicode != 0x0a){
			elems = g_list_append(elems, obj);
		}
		if (unicode == 0x0a || p == NULL) {
			Line *line = line_create(layout->obj.typesetter, (TypesetObj*)layout, elems, DIR_TYPE_TTB);
			lines = g_list_append(lines, line);
			elems = NULL;
		}
	}
	layout->lines = lines;
	return 0;
}/*}}}*/

int layout_alloc_buffer (Layout *layout, BufferContainer *buf)
{/*{{{*/
	double width = 0;
	double height = 0;
	double ad, of;
	Rect r_size, t_size;

	buf->typesetter_option = &layout->obj.typesetter->option;
	buf->render_mode = buf->typesetter_option->render_mode;
	buf->x = 0;
	buf->y = 0;

	switch (layout->obj.typesetter->option.render_mode) {
		case RENDER_MODE_GA:   buf->planes = 2; break;
		case RENDER_MODE_RGBA: buf->planes = 4; break;
		default: g_assert_not_reached(); break;
	}

	layout_get_size(layout, &ad, &of, &r_size, &t_size);
	if (layout->obj.dir == DIR_TYPE_RTL) {
		width = MAX(t_size.right, r_size.right) - MIN(t_size.left, r_size.left);
		height = MAX(t_size.bottom, r_size.bottom) - MIN(t_size.top, r_size.top);
	} else {
		width = MAX(t_size.right, r_size.right) - MIN(t_size.left, r_size.left);
		height = MAX(t_size.bottom, r_size.bottom) - MIN(t_size.top, r_size.top);
	}
	buf->width = (int)ceil(width);
	buf->height = (int)ceil(height);
	
	//FIXME: Allocate 1px extra memory to absorb error.Probably embolding is the cause of the error.
	buf->data = g_malloc0(sizeof(gint8) * (buf->width +1) * (buf->height +1) * buf->planes);
	if (buf->data == NULL) goto error;

	return 0;
error:
	if (buf->data) g_free(buf->data);
	buf->data = NULL;
	buf->width = 0;
	buf->height = 0;
	buf->typesetter_option = NULL;
	buf->render_mode = 0;
	buf->x = 0;
	buf->y = 0;
	return 1;
}/*}}}*/

static double dmax(int argc, ...)
{/*{{{*/
	va_list args;
	double arg;
	double max_value;
	int i;
	if (argc < 1) return 0;

	va_start(args , argc);
	for (i = 0 ; i < argc; i++) {
		arg = va_arg(args , double);
		max_value = (i == 0) ? arg : MAX(max_value, arg);
	}
	va_end(args);
	return max_value;
}/*}}}*/

static double dmin(int argc, ...)
{/*{{{*/
	va_list args;
	double arg;
	double min_value;
	int i;
	if (argc < 1) return 0;

	va_start(args , argc);
	for (i = 0 ; i < argc; i++) {
		arg = va_arg(args , double);
		min_value = (i == 0) ? arg : MIN(min_value, arg);
	}
	va_end(args);
	return min_value;
}/*}}}*/

static double dsum(int argc, ...)
{/*{{{*/
	va_list args;
	double arg;
	double sum_value = 0;
	int i;
	if (argc < 1) return 0;

	va_start(args , argc);
	for (i = 0 ; i < argc; i++) {
		arg = va_arg(args , double);
		sum_value += arg;
	}
	va_end(args);
	return sum_value;
}/*}}}*/

static double get_default_line_pitch (Layout *layout)
{/*{{{*/
	FT_Face face = layout->obj.typesetter->face_table[layout->obj.style].face;
	if (layout->obj.dir == DIR_TYPE_TTB) {
		return (double)face->size->metrics.x_ppem * (double)layout->obj.typesetter->option.gyou_okuri / 100.0;
	} else {
		return (double)face->size->metrics.y_ppem * (double)layout->obj.typesetter->option.gyou_okuri / 100.0;
	}
}/*}}}*/

/* vim: set fdm=marker ts=4 sw=4:*/