/*
 * GIMP Tategaki Plug-in -- This is a plug-in for The GIMP 2
 * Copyright (C) 2006-2017 OHTSUKA, Yoshio <ohtsuka.yoshio@gmail.com>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 */
#include "typeset_obj_line.h"
#include "typeset_obj_ruby.h"
#include "typeset_obj_glyphbox.h"
#include "typeset_parser.h"

static double get_default_char_pitch (Line *line);

Line* line_create(Typesetter *typ, TypesetObj *parent, GList *elements, int dir)
{/*{{{*/
	GList *glist;
	Line *line = g_slice_alloc (sizeof(Line));
	OBJ_INIT(line,
			 parent,
			 TYPESET_OBJ_TYPE_LINE,
			 typ,
			 dir,
			 line_destroy,
			 line_reset_style,
			 line_set_style,
			 line_set_dir,
			 line_set_dot,
			 line_set_bold,
			 line_set_color,
			 line_get_glyph_id,
			 line_combine_glyph,
			 line_cache_glyph_image,
			 line_compute_gbox_size,
			 line_compute_self_size,
			 line_get_size,
			 line_render);
	line->elements = elements;
	line->char_pitch = get_default_char_pitch(line);

	glist = elements;
	while (glist) {
		TypesetObj *obj = glist->data;
		obj->parent = (TypesetObj*)line;
		obj->dir = line->obj.dir;

		glist = g_list_next (glist);
	}
	return line;
}/*}}}*/

void line_destroy(Line *line)
{/*{{{*/
	GList *p;
	
	p = g_list_first(line->elements);
	while (p) {
		((TypesetObj*)(p->data))->func_destroy(p->data);
		p = g_list_next(p);
	}
	g_list_free(p);

	g_slice_free1(sizeof(Line), line);
}/*}}}*/

void line_reset_style(Line *line)
{/*{{{*/
	GList *p = line->elements;
	while (p) {
		((TypesetObj*)p->data)->func_reset_style(p->data);
		p = g_list_next(p);
	}
}/*}}}*/

void line_set_style(Line *line, int style)
{/*{{{*/
	GList *p = line->elements;
	while (p) {
		((TypesetObj*)p->data)->func_set_style(p->data, style);
		p = g_list_next(p);
	}
	line->obj.style = style;
}/*}}}*/

void line_set_dir (Line *line, DirType dir)
{/*{{{*/
	GList *p = line->elements;
	while (p) {
		((TypesetObj*)p->data)->func_set_dir(p->data, dir);
		p = g_list_next(p);
	}
	line->obj.dir = dir;
}/*}}}*/

void line_set_dot (Line *line, DotType dot)
{/*{{{*/
	GList *p = line->elements;
	while (p) {
		((TypesetObj*)p->data)->func_set_dot(p->data, dot);
		p = g_list_next(p);
	}
	line->obj.dot = DOT_TYPE_NONE;
}/*}}}*/

void line_set_bold (Line *line, double bold_x, double bold_y)
{/*{{{*/
	GList *p = line->elements;
	while (p) {
		((TypesetObj*)p->data)->func_set_bold(p->data, bold_x, bold_y);
		p = g_list_next(p);
	}
	line->obj.bold_x += bold_x;
	line->obj.bold_y += bold_y;
}/*}}}*/

void line_set_color (Line *line, Color color)
{/*{{{*/
	GList *p = line->elements;
	while (p) {
		((TypesetObj*)p->data)->func_set_color(p->data, color);
		p = g_list_next(p);
	}
	line->obj.color = color;
}/*}}}*/

void line_get_glyph_id (Line *line)
{/*{{{*/
	GList *p = line->elements;
	while (p) {
		((TypesetObj*)p->data)->func_get_glyph_id(p->data);
		p = g_list_next(p);
	}
}/*}}}*/


void line_combine_glyph (Line *line)
{/*{{{*/
	GList *p;
	line->elements = combine_glyph(line->elements);
	p = g_list_first (line->elements);
	while (p) {
		((TypesetObj*)p->data)->func_combine_glyph(p->data);
		p = g_list_next(p);
	}
}/*}}}*/

void line_cache_glyph_image (Line *line)
{/*{{{*/
	GList *p = line->elements;
	while (p) {
		((TypesetObj*)(p->data))->func_cache_glyph_image(p->data);
		p = g_list_next(p);
	}
}/*}}}*/

void line_compute_gbox_size (Line *line, TypesetObj *prev, TypesetObj *next)
{/*{{{*/
	GList *p = line->elements;
	
	while (p) {
		((TypesetObj*)p->data)->func_compute_gbox_size(p->data,
													   p->prev?p->prev->data:NULL,
													   p->next?p->next->data:NULL);
		p = g_list_next(p);
	}
}/*}}}*/

void line_compute_self_size (Line *line) 
{/*{{{*/
	double ad, of;
	Rect rsize, tsize;
	GList *p = g_list_first(line->elements);

	line->ruby_width = 0;
	line->vert_elem_num = 0;

	while (p) {
		TypesetObj *obj = p->data;
		obj->func_compute_self_size(obj);

		if (obj->type == TYPESET_OBJ_TYPE_RUBY) {
			Ruby *ruby = (Ruby*)obj;

			line->ruby_width = MAX(line->ruby_width, ruby->ruby_width);

			if (ruby->have_space) {
				line->vert_elem_num++;
			} else {
				line->vert_elem_num += ruby->parent_num;
			}

			p = g_list_next(p);
			continue;
		}
		line->vert_elem_num++;
		p = g_list_next(p);
	}
	line_get_size(line, &ad, &of, &rsize, &tsize);
	if (line->obj.dir == DIR_TYPE_TTB) {
		line->typo_height = tsize.bottom - tsize.top;//̒lJustifyƕς邱Ƃ
		line->line_width = tsize.right - tsize.left;
	} else {
		line->typo_height = tsize.right - tsize.left;
		line->line_width = tsize.bottom - tsize.top;
	}
}/*}}}*/

void line_get_size (Line *line, double *advance, double *offset, Rect *real_size, Rect *typo_size)
{/*{{{*/
	double ad, of;
	Rect r_size, t_size;
	double pitch;
	GList *p;

	*advance   = 0;
	*offset    = 0;
	real_size->top    = 0;
	real_size->bottom = 0;
	real_size->left   = 0;
	real_size->right  = 0;
	typo_size->top    = 0;
	typo_size->bottom = 0;
	typo_size->left   = 0;
	typo_size->right  = 0;

	pitch = line->char_pitch;

	p = g_list_first (line->elements);
	while (p) {
		TypesetObj *obj = p->data;
		obj->func_get_size(p->data, &ad, &of, &r_size, &t_size);

		if (line->obj.dir == DIR_TYPE_TTB) {
			real_size->top     = MIN(real_size->top, *advance + r_size.top);
			real_size->bottom  = *advance + r_size.bottom;
			real_size->left    = MIN(real_size->left,   r_size.left);
			real_size->right   = MAX(real_size->right,  r_size.right);
			typo_size->top     = MIN(typo_size->top, *advance + t_size.top);
			typo_size->bottom  = *advance + t_size.bottom;
			typo_size->left    = MIN(typo_size->left,   t_size.left);
			typo_size->right   = MAX(typo_size->right,  t_size.right);
		}
		if (line->obj.dir == DIR_TYPE_LTR) {
			real_size->top    = MIN(real_size->top,    r_size.top);
			real_size->bottom = MAX(real_size->bottom, r_size.bottom);
			real_size->left   = MIN(real_size->left, *advance + r_size.left);
			real_size->right  = *advance + r_size.right;
			typo_size->top    = MIN(typo_size->top, t_size.top);
			typo_size->bottom = MAX(typo_size->top, t_size.bottom);
			typo_size->left   = MIN(typo_size->left, *advance + t_size.left);
			typo_size->right  = *advance + t_size.right;
		}

		// _̈Z
		if (obj->dot != DOT_TYPE_NONE) {
			double ad, of;
			Rect dot_rsize, dot_tsize;
			dot_get_size(obj->dot, &ad, &of, &dot_rsize, &dot_tsize);
			if (obj->dir == DIR_TYPE_TTB) {
				typo_size->right = MAX(typo_size->right, t_size.right + dot_tsize.right - dot_tsize.left);
			} else {
				typo_size->top = MIN(typo_size->top, t_size.top - dot_tsize.top - dot_tsize.bottom);
			}
		}

		p = g_list_next(p);

		// Hook for Combining Char
		while (p) {
			TypesetObj *obj = p->data;
			GlyphBox *gbox = p->data;
			double cm_ad, cm_of;
			Rect cm_rsize, cm_tsize;
			if (obj->type != TYPESET_OBJ_TYPE_GLYPH) break;
			if (gbox->char_class != CHAR_CLASS_COMBINING) break;

			obj->func_get_size(p->data, &cm_ad, &cm_of, &cm_rsize, &cm_tsize);

			if (line->obj.dir == DIR_TYPE_TTB) {
				real_size->top    = MIN(real_size->top, *advance + cm_rsize.top);
				real_size->bottom = MAX(real_size->bottom, *advance + cm_rsize.bottom);
				real_size->left   = MIN(real_size->left, cm_rsize.left);
				real_size->right  = MAX(real_size->right, cm_rsize.right);
				//typo_size̕ύX͂ȂB
			}
			if (line->obj.dir == DIR_TYPE_LTR) {
				real_size->top    = MAX(real_size->top, cm_rsize.top);
				real_size->bottom = MIN(real_size->bottom, cm_rsize.bottom);
				real_size->left   = MIN(real_size->left, cm_rsize.left);
				real_size->right  = MAX(real_size->right, *advance + cm_rsize.right);
				//typo_size̕ύX͂ȂB
			}
			p = g_list_next(p);
		}


		*advance += ad + pitch + of;
	}

	//remove last 1 char pitch.
	if (*advance > pitch) {
		*advance -= pitch;
	}
}/*}}}*/

void line_render (Line *line, BufferContainer *buf)
{/*{{{*/
	GList *p = line->elements;
	double line_ad, line_of;
	Rect line_rsize, line_tsize;
	double ad, of;
	Rect rsize, tsize;
	double top_offset;
	double orig_x, orig_y;
	double base_line;

	//save position
	orig_x = buf->x;
	orig_y = buf->y;

	top_offset = line->offset;

	line_get_size(line, &line_ad, &line_of, &line_rsize, &line_tsize);

	switch(line->obj.dir) {
		case DIR_TYPE_TTB:
			buf->y += top_offset;
			base_line = buf->x;
			break;
		case DIR_TYPE_LTR:
			buf->x += top_offset;
			base_line = buf->y;
			break;
		case DIR_TYPE_RTL:
			buf->x -= top_offset;
			base_line = buf->y;
			break;
		default:
			g_assert_not_reached();
	}

	while (p) {
		TypesetObj *obj = p->data;
		obj->func_get_size(p->data, &ad, &of, &rsize, &tsize);

		// Move to offset
		switch(line->obj.dir) {
			case DIR_TYPE_TTB: buf->y += of; buf->x = base_line; break;
			case DIR_TYPE_LTR: buf->x += of; buf->y = base_line; break;
			case DIR_TYPE_RTL: buf->x -= of; buf->y = base_line; break;
		}
		((TypesetObj*)(p->data))->func_render(p->data, buf);

		// Hook for Rendering Combining Char 
		p = glyphbox_render_combining (p, ad, buf);

		//Render dot
		if (obj->dot != DOT_TYPE_NONE) {
			double tmp_x = buf->x;
			double tmp_y = buf->y;
			double ad, of;
			Rect dot_rsize, dot_tsize;
			dot_get_size(obj->dot, &ad, &of, &dot_rsize, &dot_tsize);

			if (line->obj.typesetter->option.dot_align == TRUE) {
				if (obj->dir == DIR_TYPE_TTB) {
					buf->x += line_tsize.right - (dot_tsize.right - dot_tsize.left);
					buf->y += ((tsize.bottom - tsize.top) - (dot_tsize.bottom - dot_tsize.top)) / 2.0;
				} else {
					buf->x += ((tsize.left + tsize.right) - (dot_tsize.left + dot_tsize.right)) / 2.0;
					buf->y += line_tsize.top;
				}
			} else {
				if (obj->dir == DIR_TYPE_TTB) {
					buf->x += tsize.right;
					buf->y += ((tsize.bottom - tsize.top) - (dot_tsize.bottom - dot_tsize.top)) / 2.0;
				} else {
					buf->x += ((tsize.left + tsize.right) - (dot_tsize.left + dot_tsize.right)) / 2.0;
					buf->y += tsize.top + dot_tsize.top - dot_tsize.bottom;
				}
			}
			dot_render(obj->dot, obj->color, buf);
			buf->x = tmp_x;
			buf->y = tmp_y;
		}

		// Move to next position
		switch(line->obj.dir) {
			case DIR_TYPE_TTB: buf->y += ad + line->char_pitch; break;
			case DIR_TYPE_LTR: buf->x += ad + line->char_pitch; break;
			case DIR_TYPE_RTL: buf->x -= ad + line->char_pitch; break;
		}
	}
	//restore position
	buf->x = orig_x;
	buf->y = orig_y;
}/*}}}*/

void line_update_height_and_offset (Line *line, double max_typo_height)
{/*{{{*/
	switch (line->obj.typesetter->option.line_align) {
		case CENTER:
			line->offset = (max_typo_height - line->typo_height) / 2;
			line->typo_height = line->typo_height;
			break;
		case BOTTOM:
			line->offset = max_typo_height - line->typo_height;
			line->typo_height = line->typo_height;
			break;
		case BOTH:
			//required to compute again.
			line->char_pitch = get_default_char_pitch(line) + 
							   1.0 * (max_typo_height - line->typo_height) / (line->vert_elem_num-1);
			line->offset = 0;
			line->typo_height = max_typo_height;
			break;
		case TOP:
		default:
			line->offset = 0;
			line->typo_height = line->typo_height;
			break;
	}
	line_compute_self_size(line);//re-compute children
}/*}}}*/

//fobOpR[h̏o
void line_debug_print (Line *line, gpointer data)
{/*{{{*/
	GList *glist = line->elements;
	while (glist) {
		GlyphBox *gbox = glist->data;
		DBGMSG("code=0x%04x\n", gbox->code);
		glist = g_list_next(glist);
	}
	DBGMSG("---------------\n");
}/*}}}*/

//e̍Ƌl߂vZB
static double get_default_char_pitch (Line *line)
{
	FT_Face face = line->obj.typesetter->face_table[line->obj.style].face;
	return (double)face->size->metrics.y_ppem * ((double)line->obj.typesetter->option.moji_okuri - 100.0) / 100.0;
}

/* vim: set fdm=marker ts=4 sw=4:*/