/*
 * GIMP Tategaki Plug-in -- This is a plug-in for The GIMP 2
 * Copyright (C) 2006-2017 OHTSUKA, Yoshio <ohtsuka.yoshio@gmail.com>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 */
#include "typeset_parser.h"
#include "typeset_obj_line.h"
#include "typeset_obj_ruby.h"
#include "typeset_obj_glyphbox.h"

static double get_char_pitch (Ruby *ruby);
static void ruby_get_parent_size (Ruby *ruby, double *advance, double *offset, Rect *real_size, Rect *typo_size);
static void ruby_get_ruby_size (Ruby *ruby, double *advance, double *offset, Rect *real_size, Rect *typo_size);


Ruby* ruby_create(Typesetter *typ, TypesetObj *parent, GList *seq_parent, GList *seq_ruby, int ruby_type, gboolean have_space, gboolean enable_trans, double max_trans_rate)
{/*{{{*/
	Ruby *ruby = NULL;
	
	ruby = g_slice_alloc(sizeof(Ruby));
	OBJ_INIT(ruby,
			 parent,
			 TYPESET_OBJ_TYPE_RUBY,
			 typ, DIR_TYPE_TTB,
			 ruby_destroy,
			 ruby_reset_style,
			 ruby_set_style,
			 ruby_set_dir,
			 ruby_set_dot,
			 ruby_set_bold,
			 ruby_set_color,
			 ruby_get_glyph_id,
			 ruby_combine_glyph,
			 ruby_cache_glyph_image,
			 ruby_compute_gbox_size,
			 ruby_compute_self_size,
			 ruby_get_size,
			 ruby_render);
	ruby->height = 0;
	ruby->ruby_type = ruby_type;
	ruby->have_space = have_space;
	ruby->does_trans = enable_trans;
	ruby->max_trans_rate = max_trans_rate;
	ruby->trans_rate = 1;
	ruby->parent = seq_parent;
	ruby->ruby = seq_ruby;

	ruby->parent = NULL;
	ruby->ruby = NULL;
	ruby_set_parent(ruby, seq_parent);
	ruby_set_ruby(ruby, seq_ruby);

	return ruby;
}/*}}}*/

void ruby_destroy (Ruby *ruby) {
	GList *p;

	p = g_list_first(ruby->parent);
	while (p) {
		((TypesetObj*)p->data)->func_destroy(p->data);
		p = g_list_next(p);
	}
	g_list_free(p);

	p = g_list_first(ruby->ruby);
	while (p) {
		((TypesetObj*)p->data)->func_destroy(p->data);
		p = g_list_next(p);
	}
	g_list_free(p);

	g_slice_free1(sizeof(Ruby), ruby);
}

void ruby_reset_style (Ruby *ruby)
{/*{{{*/
	GList *p;
	int ruby_style;

	p = ruby->parent;
	while (p) {
		((TypesetObj*)p->data)->func_reset_style(p->data);
		p = g_list_next(p);
	}

	p = ruby->ruby;
	while (p) {
		((TypesetObj*)p->data)->func_reset_style(p->data);
		p = g_list_next(p);
	}

	if (ruby->obj.typesetter->face_table[FONT_STYLE_RUBY].face == NULL) {
		ruby_style = FONT_STYLE_RUBY2;
	} else{
		ruby_style = FONT_STYLE_RUBY;
	}
	p = ruby->ruby;
	while (p) {
		TypesetObj *obj = p->data;
		((TypesetObj*)obj)->func_set_style(obj, ruby_style);
		p = g_list_next(p);
	}
}/*}}}*/

void ruby_set_style (Ruby *ruby, int style)
{/*{{{*/
	GList *p;

	p = ruby->parent;
	while (p) {
		((TypesetObj*)p->data)->func_set_style(p->data, style);
		p = g_list_next(p);
	}
	/*
	p = ruby->ruby;
	while (p) {
		((TypesetObj*)p->data)->func_set_style(p->data, style);
		p = g_list_next(p);
	}
	*/
	ruby->obj.style = style;
}/*}}}*/

void ruby_set_dir (Ruby *ruby, DirType dir)
{/*{{{*/
	GList *p;

	p = ruby->parent;
	while (p) {
		((TypesetObj*)p->data)->func_set_dir(p->data, dir);
		p = g_list_next(p);
	}

	p = ruby->ruby;
	while (p) {
		((TypesetObj*)p->data)->func_set_dir(p->data, dir);
		p = g_list_next(p);
	}

	ruby->obj.dir = dir;
}/*}}}*/

void ruby_set_dot (Ruby *ruby, DotType dot)
{/*{{{*/
	GList *p;

	p = ruby->parent;
	while (p) {
		((TypesetObj*)p->data)->func_set_dot(p->data, dot);
		p = g_list_next(p);
	}

	p = ruby->ruby;
	while (p) {
		((TypesetObj*)p->data)->func_set_dot(p->data, DOT_TYPE_NONE);
		p = g_list_next(p);
	}

	ruby->obj.dot = DOT_TYPE_NONE;
}/*}}}*/

void ruby_set_bold (Ruby *ruby, double bold_x, double bold_y)
{/*{{{*/
	GList *p;

	p = ruby->parent;
	while (p) {
		((TypesetObj*)p->data)->func_set_bold(p->data, bold_x, bold_y);
		p = g_list_next(p);
	}
	/*
	p = ruby->ruby;
	while (p) {
		((TypesetObj*)p->data)->func_set_bold(p->data, bold_x, bold_y);
		p = g_list_next(p);
	}
	*/
	ruby->obj.bold_x += bold_x;
	ruby->obj.bold_y += bold_y;
}/*}}}*/

void ruby_set_color (Ruby *ruby, Color color)
{/*{{{*/
	GList *p;

	p = ruby->parent;
	while (p) {
		((TypesetObj*)p->data)->func_set_color(p->data, color);
		p = g_list_next(p);
	}

	p = ruby->ruby;
	while (p) {
		((TypesetObj*)p->data)->func_set_color(p->data, color);
		p = g_list_next(p);
	}

	ruby->obj.color = color;
}/*}}}*/

void ruby_get_glyph_id (Ruby *ruby)
{/*{{{*/
	GList *p;
	p = ruby->parent;
	while (p) {
		((TypesetObj*)p->data)->func_get_glyph_id(p->data);
		p = g_list_next(p);
	}
	p = ruby->ruby;
	while (p) {
		((TypesetObj*)p->data)->func_get_glyph_id(p->data);
		p = g_list_next(p);
	}
}/*}}}*/

void ruby_combine_glyph (Ruby *ruby)
{/*{{{*/
	GList *p;
	TypesetObj* obj;
	ruby->parent = combine_glyph(ruby->parent);
	ruby->ruby = combine_glyph(ruby->ruby);

	p = ruby->parent;
	while (p) {
		obj = (TypesetObj*)p->data;
		if (obj->type != TYPESET_OBJ_TYPE_GLYPH) {
			obj->func_combine_glyph(obj);
		}
		p = g_list_next(p);
	}
	p = ruby->ruby;
	while (p) {
		obj = (TypesetObj*)p->data;
		if (obj->type != TYPESET_OBJ_TYPE_GLYPH) {
			obj->func_combine_glyph(obj);
		}
		p = g_list_next(p);
	}
}/*}}}*/

void ruby_cache_glyph_image (Ruby *ruby)
{/*{{{*/
	GList *p;
	p = ruby->parent;
	while (p) {
		((TypesetObj*)p->data)->func_cache_glyph_image(p->data);
		p = g_list_next(p);
	}
	p = ruby->ruby;
	while (p) {
		((TypesetObj*)p->data)->func_cache_glyph_image(p->data);
		p = g_list_next(p);
	}
}/*}}}*/

void ruby_compute_gbox_size (Ruby *ruby, TypesetObj *prev, TypesetObj *next)
{/*{{{*/
	GList *p;

	p = ruby->parent;
	while (p) {
		((TypesetObj*)p->data)->func_compute_gbox_size(p->data,
													   p->prev?p->prev->data:NULL,
													   p->next?p->next->data:NULL);
		p = g_list_next(p);
	}

	p = ruby->ruby;
	while (p) {
		((TypesetObj*)p->data)->func_compute_gbox_size(p->data,
													   p->prev?p->prev->data:NULL,
													   p->next?p->next->data:NULL);
		p = g_list_next(p);
	}
}/*}}}*/

void ruby_compute_self_size (Ruby *ruby)
{/*{{{*/
	GList *p;
	double scale_x, scale_y;
	double pitch = get_char_pitch(ruby);

	get_scale((TypesetObj*)ruby, &scale_x, &scale_y);
	pitch *= scale_y;

	ruby->parent_height = 0;
	ruby->parent_width = 0;
	ruby->ruby_height = 0;
	ruby->ruby_width = 0;
	ruby->height = 0;
	ruby->trans_rate = 1;

	p = ruby->parent;
	while (p) {
		TypesetObj *obj = p->data;
		double ad, of;
		Rect r_size, t_size;

		obj->func_compute_self_size(obj);
		obj->func_get_size(obj, &ad, &of, &r_size, &t_size);
		
		if (ruby->obj.dir == DIR_TYPE_TTB) {
			ruby->parent_width = MAX(ruby->parent_width, t_size.right - t_size.left);
			ruby->parent_height += ad + of + pitch;
		} else {
			ruby->parent_width = MAX(ruby->parent_width, t_size.bottom - t_size.top);
			ruby->parent_height += ad + of + pitch;
		}
		p = g_list_next(p);
	}
	ruby->parent_height -= pitch;

	p = ruby->ruby;
	while (p) {
		TypesetObj *obj = p->data;
		double ad, of;
		Rect r_size, t_size;

		obj->func_compute_self_size(obj);
		obj->func_get_size(obj, &ad, &of, &r_size, &t_size);

		if (ruby->obj.dir == DIR_TYPE_TTB) {
			ruby->ruby_width = MAX(ruby->ruby_width, t_size.right - t_size.left);
			ruby->ruby_height += ad + of;
		} else {
			ruby->ruby_width = MAX(ruby->ruby_width, t_size.bottom - t_size.top);
			ruby->ruby_height += ad + of;
		}
		p = g_list_next(p);
	}

	if (ruby->does_trans) {
		double h = ruby->parent_height;
		double rmax = 0;

		if (ruby->parent_num > 1) {
			rmax = h;
		} else {
			FT_Face face = ruby->obj.typesetter->face_table[FONT_STYLE_RUBY].face ? 
				ruby->obj.typesetter->face_table[FONT_STYLE_RUBY].face : ruby->obj.typesetter->face_table[FONT_STYLE_RUBY2].face;
			rmax = face->size->metrics.y_ppem * 2;
			//if (ruby->ruby_num <= 2) {
			//	ruby->trans_rate = 1;
			//}
			rmax = h;
		}

		if (ruby->does_trans) {
			if (ruby->ruby_height <= rmax) {
				ruby->trans_rate = 1;
			} else {
				ruby->trans_rate = rmax / ruby->ruby_height;
				if (ruby->trans_rate < ruby->max_trans_rate) {
					ruby->trans_rate = ruby->max_trans_rate;
					ruby->ruby_height = ruby->ruby_height * ruby->trans_rate;
				} else {
					ruby->ruby_height = rmax;
				}
			}
		}
	}

	if (ruby->have_space) {
		ruby->height = MAX(ruby->parent_height, ruby->ruby_height);
	} else {
		ruby->height = ruby->parent_height;
	}
}/*}}}*/

void ruby_get_parent_size (Ruby *ruby, double *advance, double *offset, Rect *real_size, Rect *typo_size)
{/*{{{*/
	GList *p;
	double ad, of;
	Rect r_size, t_size;
	Rect zero_size = {0, 0, 0, 0};
	double pitch = get_char_pitch(ruby);

	*offset = 0;
	*advance = 0;
	*real_size = zero_size;
	*typo_size = zero_size;

	p = ruby->parent;
	while (p) {
		((TypesetObj*)p->data)->func_get_size(p->data, &ad, &of, &r_size, &t_size);
		if (ruby->obj.dir == DIR_TYPE_TTB) {
			real_size->top     = MIN(real_size->top, *advance + r_size.top);
			real_size->bottom  = *advance + r_size.bottom;
			real_size->left    = MIN(real_size->left, r_size.left);
			real_size->right   = MAX(real_size->right, r_size.right);
			typo_size->top     = MIN(typo_size->top, *advance + t_size.top);
			typo_size->bottom  = *advance + t_size.bottom;
			typo_size->left    = MIN(typo_size->left, t_size.left);
			typo_size->right   = MAX(typo_size->right, t_size.right);
		}
		if (ruby->obj.dir == DIR_TYPE_LTR) {
			real_size->top     = MIN(real_size->top, r_size.top);
			real_size->bottom  = MAX(real_size->bottom, r_size.bottom);
			real_size->left    = MIN(real_size->left, *advance + t_size.left);
			real_size->right   = *advance + r_size.right;
			typo_size->top     = MIN(typo_size->top, t_size.top);
			typo_size->bottom  = MAX(typo_size->bottom, t_size.bottom);
			typo_size->left    = MIN(typo_size->left, *advance * t_size.left);
			typo_size->right   = *advance + t_size.right;
		}
		*advance += ad + pitch + of;
		p = g_list_next(p);
	}
	if (*advance > 0) {
		*advance -= pitch;
	}
}/*}}}*/

//ruby->tranŝɕςƁÅ֐Ŏ擾TCYƈقȂ邱Ƃ̂Œ
static void ruby_get_ruby_size (Ruby *ruby, double *advance, double *offset, Rect *real_size, Rect *typo_size)
{/*{{{*/
	GList *p;
	double ad, of;
	Rect r_size, t_size;
	Rect zero_size = {0, 0, 0, 0};

	*offset = 0;
	*advance = 0;
	*real_size = zero_size;
	*typo_size = zero_size;

	p = ruby->ruby;
	while (p) {
		((TypesetObj*)p->data)->func_get_size(p->data, &ad, &of, &r_size, &t_size);
		if (ruby->obj.dir == DIR_TYPE_TTB) {
			real_size->top     = MIN(real_size->top, *advance + r_size.top);
			real_size->bottom  = *advance + r_size.bottom;
			real_size->left    = MIN(real_size->left,   r_size.left);
			real_size->right   = MAX(real_size->right,  r_size.right);
			typo_size->top     = MIN(typo_size->top, *advance + t_size.top);
			typo_size->bottom  = *advance + t_size.bottom;
			typo_size->left    = MIN(typo_size->left,   t_size.left);
			typo_size->right   = MAX(typo_size->right,  t_size.right);
		}
		if (ruby->obj.dir == DIR_TYPE_LTR) {
			real_size->top     = MIN(real_size->top, r_size.top);
			real_size->bottom  = MAX(real_size->bottom, r_size.bottom);
			real_size->left    = MIN(real_size->left, *advance + t_size.left);
			real_size->right   = *advance + r_size.right;
			typo_size->top     = MIN(typo_size->top, t_size.top);
			typo_size->bottom  = MAX(typo_size->bottom, t_size.bottom);
			typo_size->left    = MIN(typo_size->left, *advance * t_size.left);
			typo_size->right   = *advance + t_size.right;
		}
		*advance += ad + of;
		p = g_list_next(p);
	}
}/*}}}*/

void ruby_get_size (Ruby *ruby, double *advance, double *offset, Rect *real_size, Rect *typo_size)
{/*{{{*/
	double parent_ad, parent_of, ruby_ad, ruby_of;
	Rect pr_size, pt_size, rr_size, rt_size;
	const Rect zero_size = {0, 0, 0, 0};

	*advance   = 0;
	*offset    = 0;
	*real_size = zero_size;
	*typo_size = zero_size;

	parent_ad = parent_of = 0;
	ruby_ad = ruby_of = 0;

	ruby_get_parent_size(ruby, &parent_ad, &parent_of, &pr_size, &pt_size);

	ruby_get_ruby_size(ruby, &ruby_ad, &ruby_of, &rr_size, &rt_size);

	// _̈Z
	if (((TypesetObj*)ruby->parent->data)->dot != DOT_TYPE_NONE) {
		double ad, of;
		Rect dot_rsize, dot_tsize;
		TypesetObj *obj = ruby->parent->data;
		dot_get_size(obj->dot, &ad, &of, &dot_rsize, &dot_tsize);
		if (obj->dir == DIR_TYPE_TTB) {
			pt_size.right += dot_tsize.right - dot_tsize.left;
		} else {
			pt_size.top += dot_tsize.top - dot_tsize.bottom;
		}
	}

	//if parent length >= ruby length
	if (ruby_ad <= parent_ad) {
		*advance = parent_ad;
		*offset = parent_of;

		switch (ruby->obj.dir) {
			case DIR_TYPE_TTB:
				real_size->top    = MIN(pr_size.top, rr_size.top);
				real_size->bottom = MAX(pr_size.bottom, rr_size.bottom);
				real_size->left   = MIN(pr_size.left,  pt_size.right - rt_size.left + rr_size.left);
				real_size->right  = MAX(pr_size.right, pt_size.right - rt_size.left + rr_size.right);
				typo_size->top    = 0;
				typo_size->bottom = pt_size.bottom;
				typo_size->left   = pt_size.left;
				typo_size->right  = pt_size.right - rt_size.left + rt_size.right;
				break;
			case DIR_TYPE_LTR:
				real_size->top    = MAX(pr_size.top,    pt_size.top - rt_size.bottom + rr_size.top);
				real_size->bottom = MIN(pr_size.bottom, pt_size.top - rt_size.bottom + rr_size.bottom);
				real_size->left   = MIN(pr_size.left,  rr_size.left);
				real_size->right  = MAX(pr_size.right, rr_size.right);
				typo_size->top    = pt_size.top - rt_size.bottom + rt_size.top;
				typo_size->bottom = pt_size.bottom;
				typo_size->left   = pt_size.left;
				typo_size->right  = pt_size.right;
				break;
			default:
				g_assert_not_reached();
		}
		return;
	}

	//parent length < ruby length
	if (ruby->have_space) {
		*advance = ruby_ad;
		*offset = 0;
		switch (ruby->obj.dir) {
			case DIR_TYPE_TTB:
				real_size->top    = MIN(pr_size.top, rr_size.top);
				real_size->bottom = MAX(pr_size.bottom, rr_size.bottom);
				real_size->left   = MIN(pr_size.left,  pt_size.right - rt_size.left + rr_size.left);
				real_size->right  = MAX(pr_size.right, pt_size.right - rt_size.left + rr_size.right);
				typo_size->top    = 0;
				typo_size->bottom = rt_size.bottom;
				typo_size->left   = pt_size.left;
				typo_size->right  = pt_size.right - rt_size.left + rt_size.right;
				break;
			case DIR_TYPE_LTR:
				real_size->top    = MAX(pr_size.top,    pt_size.top - rt_size.bottom + rr_size.top);
				real_size->bottom = MIN(pr_size.bottom, pt_size.top - rt_size.bottom + rr_size.bottom);
				real_size->left   = MIN(pr_size.left,  rr_size.left);
				real_size->right  = MAX(pr_size.right, rr_size.right);
				typo_size->top    = pt_size.top - rt_size.bottom + rt_size.top;
				typo_size->bottom = pt_size.bottom;
				typo_size->left   = rt_size.left;
				typo_size->right  = rt_size.right;
				break;
			default:
				g_assert_not_reached();
		}
	} else {
		*advance = parent_ad;
		*offset = 0;
		switch (ruby->obj.dir) {
			case DIR_TYPE_TTB:
				real_size->top    = MIN(pr_size.top, rr_size.top);
				real_size->bottom = MAX(pr_size.bottom, rr_size.bottom);
				real_size->left   = MIN(pr_size.left,  pt_size.right - rt_size.left + rr_size.left);
				real_size->right  = MAX(pr_size.right, pt_size.right - rt_size.left + rr_size.right);
				typo_size->top    = 0;
				typo_size->bottom = pt_size.bottom;
				typo_size->left   = pt_size.left;
				typo_size->right  = pt_size.right - rt_size.left + rt_size.right;
				break;
			case DIR_TYPE_LTR:
				real_size->top    = MAX(pr_size.top,    pt_size.top - rt_size.bottom + rr_size.top);
				real_size->bottom = MIN(pr_size.bottom, pt_size.top - rt_size.bottom + rr_size.bottom);
				real_size->left   = MIN(pr_size.left,  rr_size.left);
				real_size->right  = MAX(pr_size.right, rr_size.right);
				typo_size->top    = pt_size.top - rt_size.bottom + rt_size.top;
				typo_size->bottom = pt_size.bottom;
				typo_size->left   = pt_size.left;
				typo_size->right  = 0;
				break;
			default:
				g_assert_not_reached();
		}
	}
	return;
}/*}}}*/

void ruby_render (Ruby *ruby, BufferContainer *buf)
{/*{{{*/
	GList *p;
	double orig_x, orig_y;
	double pitch = get_char_pitch(ruby);
	double p_top, p_pitch;
	double r_top, r_pitch;
	double scale_x, scale_y;

	//save position
	orig_x = buf->x;
	orig_y = buf->y;

	get_scale((TypesetObj*)ruby, &scale_x, &scale_y);

	//
	//parent
	//
	p_top = 0;
	p_pitch = pitch;
	if (ruby->have_space && ruby->parent_height < ruby->ruby_height) {
		switch (ruby->ruby_type) {
			case RUBY_TYPE_TOP:
				p_top   = 0;
				p_pitch = pitch;
				break;
			case RUBY_TYPE_CENTER:
				p_top   = (ruby->height - ruby->parent_height)/2;
				p_pitch = pitch;
				break;
			case RUBY_TYPE_BOTTOM:
				p_top   = ruby->height - ruby->parent_height;
				p_pitch = pitch;
				break;
			case RUBY_TYPE_JUSTALL:
				if (ruby->parent_num > 1) {
					p_top   = 0;
					p_pitch = (ruby->height - ruby->parent_height + (ruby->parent_num -1)*p_pitch) / (ruby->parent_num - 1);
				} else {
					p_top   = (ruby->height - ruby->parent_height) / 2;
					p_pitch = 0;
				}
				break;
			case RUBY_TYPE_JUST121:
			default:
				if (ruby->parent_num > 1) {
					double space = ruby->height - ruby->parent_height + (ruby->parent_num -1)*p_pitch;
					p_top   = space/4;
					p_pitch = (space/2) / (ruby->parent_num - 1);
				} else {
					p_top   = (ruby->height - ruby->parent_height + (ruby->parent_num -1)*p_pitch) / 2;
					p_pitch = 0;
				}
				break;
		}
	}

	switch (ruby->obj.dir) {
		case DIR_TYPE_TTB:
			p_top   *= scale_y;
			p_pitch *= scale_y; 
			break;
		case DIR_TYPE_LTR:
			p_top   *= scale_x;
			p_pitch *= scale_x;
			break;
		case DIR_TYPE_RTL:
			g_assert_not_reached();
	}


	if (ruby->obj.dir == DIR_TYPE_TTB) {
		buf->y += p_top;
	} else {
		buf->x += p_top;
	}

	p = ruby->parent;
	while (p) {
		double ad, of;
		Rect r_size, t_size;
		TypesetObj *obj = p->data;
		GlyphBox *gbox = p->data;

		if (obj->type == TYPESET_OBJ_TYPE_GLYPH &&
			gbox->char_class == CHAR_CLASS_COMBINING) {
			p = g_list_next(p);
			continue;
		}

		((TypesetObj*)(p->data))->func_get_size(p->data, &ad, &of, &r_size, &t_size);
		// Move to offset
		switch(ruby->obj.dir) {
			case DIR_TYPE_TTB: buf->y += of; break;
			case DIR_TYPE_LTR: buf->x += of; break;
			case DIR_TYPE_RTL: buf->x -= of; break;
		}
		((TypesetObj*)(p->data))->func_render(p->data, buf);

		// Hook for Rendering Combining Char 
		p = glyphbox_render_combining (p, ad, buf);

		//Render dot
		if (obj->dot != DOT_TYPE_NONE) {
			double tmp_x = buf->x;
			double tmp_y = buf->y;
			double ad, of;
			Rect dot_rsize, dot_tsize;
			dot_get_size(obj->dot, &ad, &of, &dot_rsize, &dot_tsize);

			if (obj->dir == DIR_TYPE_TTB) {
				buf->x += t_size.right + ruby->ruby_width;
				buf->y += ((t_size.bottom - t_size.top) - (dot_tsize.bottom - dot_tsize.top)) / 2.0;
			} else {
				buf->x += ((t_size.left + t_size.right) - (dot_tsize.left + dot_tsize.right)) / 2.0;
				buf->y += t_size.top + dot_tsize.top - dot_tsize.bottom - ruby->ruby_width;
			}
			dot_render(obj->dot, obj->color, buf);
			buf->x = tmp_x;
			buf->y = tmp_y;
		}

		// Move to next position
		switch(ruby->obj.dir) {
			case DIR_TYPE_TTB: buf->y += (ad + p_pitch) ; break;
			case DIR_TYPE_LTR: buf->x += (ad + p_pitch) ; break;
			case DIR_TYPE_RTL: buf->x -= (ad + p_pitch) ; break;
		}
		continue;
	}
	//
	//ruby
	//
	r_top = r_pitch = 0;

	//restore position
	buf->x = orig_x;
	buf->y = orig_y;

	if (ruby->parent_height >= ruby->ruby_height) {
		switch (ruby->ruby_type) {
			case RUBY_TYPE_TOP:
				r_top   = 0;
				r_pitch = 0;
				break;
			case RUBY_TYPE_CENTER:
				r_top   = (ruby->parent_height - ruby->ruby_height)/2;
				r_pitch = 0;
				break;
			case RUBY_TYPE_BOTTOM:
				r_top = ruby->parent_height - ruby->ruby_height;
				r_pitch = 0;
				break;
			case RUBY_TYPE_JUSTALL:
				if (ruby->ruby_num == 1) {
					r_top   = (ruby->parent_height - ruby->ruby_height)/2;
					r_pitch = 0;
				} else {
					r_top   = 0;
					r_pitch = (ruby->parent_height - ruby->ruby_height) / (ruby->ruby_num - 1);
				}
				break;
			case RUBY_TYPE_JUST121:
			default:
				if (ruby->ruby_num == 1) {
					r_top = (ruby->parent_height - ruby->ruby_height)/2;
				} else {
					double space = ruby->parent_height - ruby->ruby_height;
					r_top   = space/4;
					r_pitch = (space/2) / (ruby->ruby_num - 1);
				}
				break;
		}
	} else {
		switch (ruby->ruby_type) {
			case RUBY_TYPE_TOP:
				r_top   = 0;
				r_pitch = 0;
				break;
			case RUBY_TYPE_BOTTOM:
				if (ruby->have_space) {
					r_top   = 0;
					r_pitch = 0;
				} else {
					r_top   = ruby->parent_height - ruby->ruby_height;
					r_pitch = 0;
				}
				break;
			case RUBY_TYPE_CENTER:
			case RUBY_TYPE_JUSTALL:
			case RUBY_TYPE_JUST121:
			default:
				if (ruby->have_space) {
					r_top   = 0;
					r_pitch = 0;
				} else {
					r_top = (ruby->parent_height - ruby->ruby_height) /2;
					r_pitch = 0;
				}
				break;
		}
	}

	switch (ruby->obj.dir) {
		case DIR_TYPE_TTB:
			r_top   *= scale_y;
			r_pitch *= scale_y; 
			break;
		case DIR_TYPE_LTR:
			r_top   *= scale_x;
			r_pitch *= scale_x;
			break;
		case DIR_TYPE_RTL:
			g_assert_not_reached();
	}
	
	{
		double dummy1, dummy2;
		Rect dummy_size, parent_typo_size, ruby_typo_size;
		ruby_get_parent_size(ruby, &dummy1, &dummy2, &dummy_size, &parent_typo_size);
		ruby_get_ruby_size(ruby, &dummy1, &dummy2, &dummy_size, &ruby_typo_size);
		if (ruby->obj.dir == DIR_TYPE_TTB) {
			buf->x += (parent_typo_size.right - ruby_typo_size.left);
			buf->y += r_top;
		} else {
			buf->x += r_top;
			buf->y += (parent_typo_size.top - ruby_typo_size.bottom);
		}
	}

	if (ruby->does_trans == FALSE) {
		p = ruby->ruby;
		while (p) {
			double ad, of;
			Rect real_size, typo_size;
			((TypesetObj*)(p->data))->func_get_size(p->data, &ad, &of, &real_size, &typo_size);
			// Move to offset
			switch(ruby->obj.dir) {
				case DIR_TYPE_TTB: buf->y += of; break;
				case DIR_TYPE_LTR: buf->x += of; break;
				case DIR_TYPE_RTL: buf->x -= of; break;
			}
			((TypesetObj*)(p->data))->func_render(p->data, buf);

			// Hook for Rendering Combining Char 
			p = glyphbox_render_combining (p, ad, buf);

			// Move to next position
			switch(ruby->obj.dir) {
				case DIR_TYPE_TTB: buf->y += (ad + r_pitch); break;
				case DIR_TYPE_LTR: buf->x += (ad + r_pitch); break;
				case DIR_TYPE_RTL: buf->x -= (ad + r_pitch); break;
			}
		}
	} else {
		p = ruby->ruby;
		while (p) {
			double ad, of;
			Rect real_size, typo_size;

			((TypesetObj*)p->data)->func_get_size(p->data, &ad, &of, &real_size, &typo_size);
			// Move to offset
			switch(ruby->obj.dir) {
				case DIR_TYPE_TTB: buf->y += of; break;
				case DIR_TYPE_LTR: buf->x += of; break;
				case DIR_TYPE_RTL: buf->x -= of; break;
			}
			((TypesetObj*)p->data)->func_render(p->data, buf);

			// Hook for Rendering Combining Char 
			p = glyphbox_render_combining (p, ad, buf);

			// Move to next position
			switch(ruby->obj.dir) {
				case DIR_TYPE_TTB: buf->y += (ad + r_pitch); break;
				case DIR_TYPE_LTR: buf->x += (ad + r_pitch); break;
				case DIR_TYPE_RTL: buf->x -= (ad + r_pitch); break;
			}
		}
	}

	//restore position
	buf->x = orig_x;
	buf->y = orig_y;
}/*}}}*/

void ruby_set_parent(Ruby *ruby, GList *seq)
{/*{{{*/
	GList *p;

	//list_destroy_seq(ruby->parent); seqruby->parentƓ̉\
	ruby->parent = seq;
	ruby->parent_num = 0;

	p = g_list_first(ruby->parent);
	while (p) {
		TypesetObj *obj = p->data;
		GlyphBox *gbox = (GlyphBox*)p->data;
		obj->parent = (TypesetObj*)ruby;

		if (obj->type == TYPESET_OBJ_TYPE_GLYPH &&
			gbox->char_class == CHAR_CLASS_COMBINING) {
			p = g_list_next(p);
			continue;
		}
		ruby->parent_num++;
		p = g_list_next(p);
	}
}/*}}}*/

void ruby_set_ruby(Ruby *ruby, GList *seq)
{/*{{{*/
	GList *p;
	int ruby_style = FONT_STYLE_RUBY;

	//list_destroy_seq(ruby->ruby);seqruby->rubyƓ̉\
	ruby->ruby = seq;
	ruby->ruby_num = 0;

	if (ruby->obj.typesetter->face_table[FONT_STYLE_RUBY].face == NULL) {
		ruby_style = FONT_STYLE_RUBY2;
	} else {
		ruby_style = FONT_STYLE_RUBY;
	}

	p = g_list_first(ruby->ruby);
	while (p) {
		TypesetObj *obj = p->data;
		GlyphBox *gbox = (GlyphBox*)p->data;

		obj->parent = (TypesetObj*)ruby;
		obj->func_set_style(obj, ruby_style);

		if (obj->type == TYPESET_OBJ_TYPE_GLYPH &&
			gbox->char_class == CHAR_CLASS_COMBINING) {
			p = g_list_next(p);
			continue;
		}
		ruby->ruby_num++;
		p = g_list_next(p);
	}
}/*}}}*/

static double get_char_pitch (Ruby *ruby)
{
	TypesetObj *parent = NULL;
	if (ruby->obj.parent) {
		parent = ruby->obj.parent;
	}
	while (parent) {
		switch (parent->type) {
			case TYPESET_OBJ_TYPE_LINE: return ((Line*)parent)->char_pitch;
			case TYPESET_OBJ_TYPE_YOKO: return 0;
			default: break;
		}
		parent = parent->parent;
	}
	return 0;
}

/* vim: set fdm=marker ts=4 sw=4:*/
