/*
 * GIMP Tategaki Plug-in -- This is a plug-in for The GIMP 2
 * Copyright (C) 2006-2017 OHTSUKA, Yoshio <ohtsuka.yoshio@gmail.com>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 */
#include <stdlib.h>
#include "font_list.h"
#include "typeset_parser.h"
#include "typeset_obj.h"
#include "typeset_obj_glyphbox.h"
#include "typeset_obj_ruby.h"
#include "typeset_obj_yoko.h"
#include "typeset_obj_line.h"

static gchar* list_to_string(GList *seq);
static gboolean list_to_double(GList *seq, double *result);
static gboolean list_to_int(GList *seq, int *result);

gboolean list_is_match_str (GList *list, char *str)
{/*{{{*/
	GList *glist = list;
	int len = (int)strlen(str);
	int i = 0;
	
	if (len != g_list_length(glist)) {
		return FALSE;
	}

	for (i=0; i < len ; i++) {
		GlyphBox *gbox = glist->data;
		if (gbox->obj.type != TYPESET_OBJ_TYPE_GLYPH) {
			return FALSE;
		}
		if (str[i] != gbox->unicode) {
			return FALSE;
		}
		glist = g_list_next(glist);
	}
	if (i != len) return FALSE;
	return TRUE;
}/*}}}*/

GList* get_gbox_sequence (Typesetter *typ, gchar *str, FontFlag flag)
{/*{{{*/
	gchar *p;
	GList *seq = NULL;

	p = str;
	while (*p) {
		GlyphBox *gbox = glyphbox_create(NULL);
		guint32 unicode;
		unicode = g_utf8_get_char(p);

		//replace U+309b to U+3099, replace U+309c to U+309a.
		//because combined glyph is better. but if font have no 
		//combined glyph then the unicode and glyph is back to 
		//U+309b/U+309c in glyphbox_combine_glyph() function.
		switch (unicode) {
			case 0x309b: unicode = 0x3099; break;
			case 0x309c: unicode = 0x309a; break;
		}

		gbox->obj.typesetter = typ;
		gbox->obj.dir = DIR_TYPE_TTB;
		gbox->unicode = unicode;
		gbox->obj.style = FONT_STYLE_BASE;
		gbox->obj.fontflag = flag;
		gbox->obj.color = typ->option.color;

		p = g_utf8_next_char(p);

		if (*p) {
			unicode = g_utf8_get_char(p);
			if (unicode >= 0xE0100 && unicode <= 0xE01EF) {
				gbox->variant_selector = unicode;
				p = g_utf8_next_char(p);
			} else {
				gbox->variant_selector = 0;
			}
		}
		seq = g_list_append(seq, gbox);
	}
	if (g_list_length(seq) == 0) {
		g_list_free(seq);
		return NULL;
	}
	return g_list_first(seq);
}/*}}}*/

/// seq̈ʒuR}hċAIɒTĕԂB
/// Eseq̃Xg͕ύXȂB
/// Eout͐V쐬ꂽTypesetObj**z**BR}hᖳƂ͈ꕶ̃XgԂB
///   out͌ĂяoŃXg̃Ȃ΂ȂȂB
/// EnextseqƓXǵÅJnʒuBnext͎seqƂȂB
static void list_parse_command (Typesetter *typ, GList *seq, GList **out, GList **next)
{
	GList *p      = NULL;
	GList *tokens = NULL;
	GList *token  = NULL;
	TypesetObj *obj;

	if (seq == NULL) {
		*out = NULL;
		*next = NULL;
		return;
	}

	p = seq;
	obj = (TypesetObj*)p->data;

	if (obj->type == TYPESET_OBJ_TYPE_GLYPH) {
		GlyphBox *gbox = (GlyphBox*)obj;

		//escape cmd
		if (gbox->unicode == '\\') {
			p = g_list_next(p);
			if (p) {
				*out = g_list_append(NULL, p->data);
				*next = g_list_next(p);
			} else {
				*out  = NULL;
				*next = NULL;
			}
			return;
		}
		
		//check if command
		if (gbox->unicode == '[') {
			gboolean quoted = FALSE;
			p = g_list_next(p);

			//split to token
			tokens = NULL;
			token = NULL;
			while (p) {
				GList *cmd_out, *cmd_next;
				
				//escape(ex. escape space)
				if (((TypesetObj*)p->data)->type == TYPESET_OBJ_TYPE_GLYPH) {
					GlyphBox *gbox = (GlyphBox*)p->data;
					if (gbox->unicode == '\\') {
						p = g_list_next(p);
						if (p == NULL) { goto PURGE_AS_NOT_COMMAND; }

						token = g_list_append(token, p->data);
						p = g_list_next(p);
						if (p == NULL) { goto PURGE_AS_NOT_COMMAND; }
						continue;
					}
				}
				//quote
				if (((TypesetObj*)p->data)->type == TYPESET_OBJ_TYPE_GLYPH) {
					GlyphBox *gbox = (GlyphBox*)p->data;
					if (gbox->unicode == '"') {
						p = g_list_next(p);
						if (p == NULL) { goto PURGE_AS_NOT_COMMAND; }
						quoted ^= TRUE;
						continue;
					}
				}

				list_parse_command (typ, p, &cmd_out, &cmd_next);
				if (cmd_out == NULL) goto PURGE_AS_NOT_COMMAND;
				
				if (quoted == FALSE && ((TypesetObj*)cmd_out->data)->type == TYPESET_OBJ_TYPE_GLYPH) {
					GlyphBox *token_gbox = (GlyphBox*)cmd_out->data;
					//purge!
					if (token_gbox->unicode == ']') {
						tokens = g_list_append(tokens, g_list_first(token));
						token = NULL;
						p = cmd_next;
						break;
					}
					
					//skip space
					if (token_gbox->unicode == ' ') {
						if (token != NULL) {
							tokens = g_list_append(tokens, g_list_first(token));
							token = NULL;
						}
						p = cmd_next;
						continue;
					}
				}
				token = g_list_concat(token, cmd_out);
				p = cmd_next;
			}
			
			//check cmd token
			{
				int cmd_type = TYPESET_CMD_TYPE_NONE;
				int tokens_count;
				GList *t0, *t1, *t2, *t3;
				gpointer aobj;

				tokens_count = g_list_length(tokens);
				if (tokens_count == 0) {
					goto PURGE_AS_NOT_COMMAND;
				}

				t0 = g_list_first(tokens)->data;//command name
				t1 = tokens_count >= 2 ? g_list_nth(tokens, 1)->data : NULL;
				t2 = tokens_count >= 3 ? g_list_nth(tokens, 2)->data : NULL;
				t3 = tokens_count >= 4 ? g_list_nth(tokens, 3)->data : NULL;

				if (list_is_match_str(t0, "ruby"))  cmd_type = TYPESET_CMD_TYPE_RUBY_RJJE;
				if (list_is_match_str(t0, "rt"))    cmd_type = TYPESET_CMD_TYPE_RUBY_RT;
				if (list_is_match_str(t0, "rc"))    cmd_type = TYPESET_CMD_TYPE_RUBY_RC;
				if (list_is_match_str(t0, "rb"))    cmd_type = TYPESET_CMD_TYPE_RUBY_RB;
				if (list_is_match_str(t0, "rja"))   cmd_type = TYPESET_CMD_TYPE_RUBY_RJA;
				if (list_is_match_str(t0, "rjj"))   cmd_type = TYPESET_CMD_TYPE_RUBY_RJJ;
				if (list_is_match_str(t0, "rte"))   cmd_type = TYPESET_CMD_TYPE_RUBY_RTE;
				if (list_is_match_str(t0, "rce"))   cmd_type = TYPESET_CMD_TYPE_RUBY_RCE;
				if (list_is_match_str(t0, "rbe"))   cmd_type = TYPESET_CMD_TYPE_RUBY_RBE;
				if (list_is_match_str(t0, "rjae"))  cmd_type = TYPESET_CMD_TYPE_RUBY_RJAE;
				if (list_is_match_str(t0, "rjje"))  cmd_type = TYPESET_CMD_TYPE_RUBY_RJJE;
				if (list_is_match_str(t0, "yoko"))  cmd_type = TYPESET_CMD_TYPE_YOKO;
				if (list_is_match_str(t0, "dot"))   cmd_type = TYPESET_CMD_TYPE_DOT;
				if (list_is_match_str(t0, "goma"))  cmd_type = TYPESET_CMD_TYPE_GOMA;
				if (list_is_match_str(t0, "scale")) cmd_type = TYPESET_CMD_TYPE_SCALE;
				if (list_is_match_str(t0, "bold"))  cmd_type = TYPESET_CMD_TYPE_BOLD;
				if (list_is_match_str(t0, "font"))  cmd_type = TYPESET_CMD_TYPE_FONT;
				if (list_is_match_str(t0, "color")) cmd_type = TYPESET_CMD_TYPE_COLOR;

				if (cmd_type == TYPESET_CMD_TYPE_NONE) {
					goto PURGE_AS_NOT_COMMAND;
				}
				
				aobj = NULL;
				switch (cmd_type) {
#define RUBY_ARG_CHECK if (tokens_count != 3 || t1 == NULL || t2 == NULL) {goto PURGE_AS_NOT_COMMAND;}
					case TYPESET_CMD_TYPE_RUBY_RT:  RUBY_ARG_CHECK; aobj = ruby_create(typ, NULL, t1, t2, RUBY_TYPE_TOP,    FALSE, typ->option.ruby_trans, typ->option.ruby_max_trans_rate); break;
					case TYPESET_CMD_TYPE_RUBY_RC:  RUBY_ARG_CHECK; aobj = ruby_create(typ, NULL, t1, t2, RUBY_TYPE_CENTER, FALSE, typ->option.ruby_trans, typ->option.ruby_max_trans_rate); break;
					case TYPESET_CMD_TYPE_RUBY_RB:  RUBY_ARG_CHECK; aobj = ruby_create(typ, NULL, t1, t2, RUBY_TYPE_BOTTOM, FALSE, typ->option.ruby_trans, typ->option.ruby_max_trans_rate); break;
					case TYPESET_CMD_TYPE_RUBY_RJA: RUBY_ARG_CHECK; aobj = ruby_create(typ, NULL, t1, t2, RUBY_TYPE_JUSTALL,FALSE, typ->option.ruby_trans, typ->option.ruby_max_trans_rate); break;
					case TYPESET_CMD_TYPE_RUBY_RJJ: RUBY_ARG_CHECK; aobj = ruby_create(typ, NULL, t1, t2, RUBY_TYPE_JUST121,FALSE, typ->option.ruby_trans, typ->option.ruby_max_trans_rate); break;
					case TYPESET_CMD_TYPE_RUBY_RTE: RUBY_ARG_CHECK; aobj = ruby_create(typ, NULL, t1, t2, RUBY_TYPE_TOP,    TRUE,  typ->option.ruby_trans, typ->option.ruby_max_trans_rate); break;
					case TYPESET_CMD_TYPE_RUBY_RCE: RUBY_ARG_CHECK; aobj = ruby_create(typ, NULL, t1, t2, RUBY_TYPE_CENTER, TRUE,  typ->option.ruby_trans, typ->option.ruby_max_trans_rate); break;
					case TYPESET_CMD_TYPE_RUBY_RBE: RUBY_ARG_CHECK; aobj = ruby_create(typ, NULL, t1, t2, RUBY_TYPE_BOTTOM, TRUE,  typ->option.ruby_trans, typ->option.ruby_max_trans_rate); break;
					case TYPESET_CMD_TYPE_RUBY_RJJE:RUBY_ARG_CHECK; aobj = ruby_create(typ, NULL, t1, t2, RUBY_TYPE_JUST121,TRUE,  typ->option.ruby_trans, typ->option.ruby_max_trans_rate); break;
#undef RUBY_ARG_CHECK
					
					case TYPESET_CMD_TYPE_YOKO: {
						double trans = 0;
						AlignType align = CENTER;
						if (tokens_count < 2 || tokens_count > 4) {
							goto PURGE_AS_NOT_COMMAND;
						}
						if (t1 == NULL) {
							goto PURGE_AS_NOT_COMMAND;
						}
						if (t2 != NULL) {
							double tmp;
							if (list_to_double(t2, &tmp) && tmp > 0) {
								trans = tmp;
							}
						}
						if (t3 != NULL) {
							if (((TypesetObj*)t3->data)->type == TYPESET_OBJ_TYPE_GLYPH) {
								switch (((GlyphBox*)t3->data)->unicode) {
									case 'T': //TOP
									case 't': align = TOP; break;
									case 'C': //CENTER
									case 'c': align = CENTER; break;
									case 'B': //BOTTOM
									case 'b': align = BOTTOM; break;
								}
							}
						}
						aobj = yoko_create(typ, NULL, t1, trans, align);
						break;
					}
					case TYPESET_CMD_TYPE_DOT:
					case TYPESET_CMD_TYPE_GOMA: {
						DotType dot_type = DOT_TYPE_MARU;
						GList *p2 = t1;
						if (tokens_count != 2) {
							goto PURGE_AS_NOT_COMMAND;
						}
						if (t1 == NULL) {
							goto PURGE_AS_NOT_COMMAND;
						}
						switch (cmd_type) {
							case TYPESET_CMD_TYPE_DOT:  dot_type = DOT_TYPE_MARU; break;
							case TYPESET_CMD_TYPE_GOMA: dot_type = DOT_TYPE_GOMA; break;
						}
						while (p2) {
							((TypesetObj*)p2->data)->func_set_dot(p2->data, dot_type);
							p2 = g_list_next(p2);
						}
						*out  = t1;
						*next = p;
						return;
					}

					case TYPESET_CMD_TYPE_SCALE: {
						GList *p2 = t1;
						double tmp;
						double scale_x = 1;
						double scale_y = 1;
						if (tokens_count < 2 || tokens_count > 4) {
							goto PURGE_AS_NOT_COMMAND;
						}
						if (t2 != NULL && list_to_double(t2, &tmp) && tmp > 0) {
							scale_x = tmp;
						}
						if (t3 != NULL && list_to_double(t3, &tmp) && tmp > 0) {
							scale_y = tmp;
						} else {
							scale_y = scale_x;
						}
						while (p2) {
							((TypesetObj*)p2->data)->scale_x *= scale_x;
							((TypesetObj*)p2->data)->scale_y *= scale_y;
							p2 = g_list_next(p2);
						}
						*out  = t1;
						*next = p;
						return;
					}

					case TYPESET_CMD_TYPE_BOLD: {
						GList *p2 = t1;
						double tmp;
						double bold_x = 4;
						double bold_y = 4;
						if (tokens_count < 2 || tokens_count > 4) {
							goto PURGE_AS_NOT_COMMAND;
						}
						if (t1 == NULL) {
							goto PURGE_AS_NOT_COMMAND;
						}
						if (t2 != NULL && list_to_double(t2, &tmp)) {
							bold_x = tmp;
						}
						if (t3 != NULL && list_to_double(t3, &tmp)) {
							bold_y = tmp;
						} else {
							bold_y = bold_x;
						}
						while (p2) {
							((TypesetObj*)p2->data)->func_set_bold(p2->data, bold_x, bold_y);
							p2 = g_list_next(p2);
						}
						*out  = t1;
						*next = p;
						return;
					}

					case TYPESET_CMD_TYPE_FONT: {
						GList *p2 = t1;
						gchar* font_name = NULL;
						int i, style_id;
						GList *font;
						FontStyle *style;
						FontInfo *tmp, *fontinfo;
						if (tokens_count != 3) {
							goto PURGE_AS_NOT_COMMAND;
						}

						if (t1 == NULL) {
							goto PURGE_AS_NOT_COMMAND;
						}
						if (t2 != NULL) {
							font_name = list_to_string(t2);
						} else {
							goto DONE_TYPESET_CMD_TYPE_FONT;
						}

						fontinfo = NULL;
						font = typ->font_list;
						while (font) {
							tmp = font->data;
							if (g_strcmp0(font_name, tmp->name) == 0 ||
								g_strcmp0(font_name, tmp->name_en) == 0 ||
								g_strcmp0(font_name, tmp->name_old) == 0 ||
								g_strcmp0(font_name, tmp->name_ps) == 0) {
								fontinfo = tmp;
								break;
							}
							font = g_list_next(font);
						}
						if (fontinfo == NULL) {
							goto DONE_TYPESET_CMD_TYPE_FONT;
						}

						style_id = -1;
						for (i = FONT_STYLE_USER; i < FONT_STYLE_MAX; i++) {
							if (typ->face_table[i].style) {
								if (g_strcmp0(typ->face_table[i].style->file_path, fontinfo->path) == 0 &&
									typ->face_table[i].style->face_index == fontinfo->face_index) {
									style_id = i;
								}
							}
						}

						if (style_id == -1) {
							style_id = typesetter_get_empty_user_style_slot (typ);
							if (style_id != -1) {
								style = typesetter_fontstyle_create(fontinfo->name,
																	fontinfo->path,
																	fontinfo->face_index,
																	typ->face_table[FONT_STYLE_BASE].style->dpi_x,
																	typ->face_table[FONT_STYLE_BASE].style->dpi_y,
																	typ->face_table[FONT_STYLE_BASE].style->size_x,
																	typ->face_table[FONT_STYLE_BASE].style->size_y,
																	typ->face_table[FONT_STYLE_BASE].style->unit);
								typesetter_set_style(typ, style_id, style);
							}
						}

						if (style_id != -1) {
							while (p2) {
								((TypesetObj*)p2->data)->func_set_style(p2->data, style_id);
								p2 = g_list_next(p2);
							}
						}
DONE_TYPESET_CMD_TYPE_FONT:
						if (font_name) g_free(font_name);
						*out  = t1;
						*next = p;
						return;
					}

					case TYPESET_CMD_TYPE_COLOR: {
						GList *p2 = t1;
						gchar *color_name = NULL;
						Color color;

						if (tokens_count != 3) {
							goto PURGE_AS_NOT_COMMAND;
						}
						if (t1 == NULL) {
							goto PURGE_AS_NOT_COMMAND;
						}
						if (t2 != NULL) {
							color_name = list_to_string(t2);
						}
						if (color_name && color_from_str(color_name, &color)) {
							switch (typ->option.render_mode) {
								case RENDER_MODE_RGBA:
									/* do nothing */
									break;
								case RENDER_MODE_GA:
									color.r = color_to_intensity_uchar(color);
									color.g = color.r;
									color.b = color.r;
									color.a = 0xFF;
									break;
								default:
									g_assert_not_reached();
									break;
							}
							while (p2) {
								((TypesetObj*)p2->data)->func_set_color(p2->data, color);
								p2 = g_list_next(p2);
							}
						}
						if (color_name) g_free(color_name);
						*out  = t1;
						*next = p;
						return;
					}

					default:
						goto PURGE_AS_NOT_COMMAND;
				}
				*out  = g_list_append(NULL, aobj);
				*next = p;
				return;
			}
		}
	}

PURGE_AS_NOT_COMMAND:
	*out = g_list_append(NULL, seq->data);
	*next = g_list_next(seq);
	if (tokens) {
		while (tokens) {
			if (tokens->data) g_list_free(tokens->data);
			tokens = g_list_next(tokens);
		}
		g_list_free(tokens);
	}
	return;
}

///seqR}hp[XAVzĕԂB
///seq͔j󂳂̂ŁAԒl̂قgƁB
GList* list_parse (Typesetter *typ, GList *seq)
{/*{{{*/
	GList *new_seq = NULL;
	GList *p = seq;

	while (p) {
		GList *p2, *out, *next;
		list_parse_command(typ, p, &out, &next);
		p2 = out;
		while (p2) {
			new_seq = g_list_append(new_seq, p2->data);
			p2 = g_list_next(p2);
		}
		if (out) {
			g_list_free(out);
		}
		p = next;
	}
	g_list_free(seq);
	return g_list_first(new_seq);
}/*}}}*/

void list_reset_style (Typesetter *typ, GList *seq)
{/*{{{*/
	GList *p = seq;

	while (p) {
		TypesetObj* obj = p->data;
		obj->func_reset_style(obj);
		p = g_list_next(p);
	}
}/*}}}*/

///seq玩IɏcR}h쐬ēւĂB
///seq͊{IɕύXȂA擪̕ւꍇ
///seq̃|C^ς邱Ƃ̂ŒӁB
GList* list_auto_yoko (Typesetter *typ, GList *seq)
{/*{{{*/
	GList *p = seq;

	if (typ->option.auto_yoko == FALSE) {
		return seq;
	}

	while (p) {
		TypesetObj* obj = p->data;
		if (obj->type == TYPESET_OBJ_TYPE_GLYPH) {
			GlyphBox *gbox = p->data;
			if (gbox->char_class == CHAR_CLASS_NUMBER ||
				gbox->char_class == CHAR_CLASS_EXCLAM) {
				int char_class = gbox->char_class;
				GList *num_list = NULL;
				DotType dot;
				GList *p2;
				int pos = g_list_position(seq, p);

				dot = gbox->obj.dot;
				if (gbox->char_class == CHAR_CLASS_EXCLAM && typ->option.dot_skip_yaku) {
					dot = DOT_TYPE_NONE;
				}

				p2 = p;
				while (p2) {
					GList *next;
					if (((TypesetObj*)p2->data)->type != TYPESET_OBJ_TYPE_GLYPH ||
						((GlyphBox*)p2->data)->char_class != char_class) {
							break;
					}
					next = g_list_next(p2);
					seq = g_list_remove_link(seq, p2);
					num_list = g_list_concat(num_list, p2);
					p2 = next;
				}

				num_list = g_list_first(num_list);

				if (g_list_length(num_list) <= typ->option.auto_yoko_digit) {
					Yoko *yoko = yoko_create(typ, gbox->obj.parent, num_list, 0, CENTER);
					yoko->obj.func_set_dot(yoko, dot);
					seq = g_list_insert(seq, yoko, pos);
					p = g_list_nth(seq, pos);
				} else {
					while (g_list_length(num_list)) {
						Yoko *yoko;
						GList *p3 = g_list_nth(num_list, 0);
						num_list = g_list_remove_link(num_list, p3);
						yoko = yoko_create(typ, gbox->obj.parent, p3, 0, CENTER);
						yoko->obj.func_set_dot(yoko, dot);
						seq = g_list_insert(seq, yoko, pos++);
					}
					p = g_list_nth(seq, pos);
				}
				continue;
			}
		}
		if (obj->type == TYPESET_OBJ_TYPE_RUBY) {
			Ruby *ruby = p->data;
			ruby_set_parent(ruby, list_auto_yoko(ruby->obj.typesetter, ruby->parent));
			ruby_set_ruby(ruby, list_auto_yoko(ruby->obj.typesetter, ruby->ruby));
		}
		p = g_list_next(p);
	}

	return seq;
}/*}}}*/

void list_destroy_seq (GList *seq) {
	GList *p = seq;
	while (p) {
		TypesetObj* obj = p->data;
		obj->func_destroy(p->data);
		p = g_list_next(p);
	}
	g_list_free(seq);
}

//sequtf-8BԂlg_freeŉKvB
static gchar* list_to_string(GList *seq)
{
	GList *p = g_list_first(seq);
	GString *gstr = g_string_sized_new(16);
	while (p) {
		GlyphBox *gbox = p->data;
		gstr = g_string_append_unichar(gstr, gbox->unicode);
		p = g_list_next(p);
	}
	return g_string_free(gstr, FALSE);
}

//seqdouble^BԂl͐ꍇTrue
static gboolean list_to_double(GList *seq, double *result)
{
	gchar *utf8 = list_to_string(seq);
	if (utf8) {
		*result = atof(utf8);
		g_free(utf8);
		return TRUE;
	}
	return FALSE;
}

//seqint^BԂl͐ꍇTrue
static gboolean list_to_int(GList *seq, int *result)
{
	gchar *utf8 = list_to_string(seq);
	if (utf8) {
		*result = atoi(utf8);
		g_free(utf8);
		return TRUE;
	}
	return FALSE;
}

/* vim: set fdm=marker ts=4 sw=4:*/