/*
 * GIMP Tategaki Plug-in -- This is a plug-in for The GIMP 2
 * Copyright (C) 2006-2017 OHTSUKA, Yoshio <ohtsuka.yoshio@gmail.com>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 */
#include <glib.h>
#include <math.h>
#include "tategaki.h"
#include "typesetter.h"
#include "font_list.h"
#include "encoding.h"
#include "typeset_obj.h"
#include "typeset_obj_layout.h"
#include "typeset_obj_line.h"
#include "typeset_obj_glyphbox.h"
#include "typeset_obj_ruby.h"
#include "typeset_parser.h"

static void debug_print_font(FT_Face face);
static FontFlag create_font_flag (TypesetterOption* option);

#define IS_ERROR(error, msg) \
	if ((error)) { \
		ERRMSG( msg " %d\n", (error)); return 1;\
	}


FT_Error face_set_size (FT_Face face, FontStyle *style)
{/*{{{*/
	switch (style->unit) {
		case POINT:
			return  FT_Set_Char_Size(face, 
					style->size_x,
					style->size_y,
					style->dpi_x,
					style->dpi_y);
			break;
		case PIXEL:
		default:
			return  FT_Set_Pixel_Sizes(face,
					F26D6_TO_INT32(style->size_x),
					F26D6_TO_INT32(style->size_y));
			break;
	}
	return 0;
}/*}}}*/

double face_get_base_line (FT_Face face)
{/*{{{*/
	double as, ds, lg;
	get_typo_metrics(face, &as, &ds, &lg);
	return -ds;
}/*}}}*/

Typesetter* typesetter_create()
{/*{{{*/
	Typesetter *typ;
	FT_Error error = 0;
	typ = g_malloc0(sizeof(Typesetter));

	error = FT_Init_FreeType(&(typ->library));
	if (error) {
		ERRMSG("failed to initialize freetype. error code=%d\n", error);
		return 0;
	}
	typ->face_table = g_malloc0(sizeof(FontFaceTable) * FONT_STYLE_MAX);

	return typ;
}/*}}}*/

int typesetter_destroy(Typesetter *typ)
{/*{{{*/
	if (typ->library) {
		typesetter_reset_style (typ);
		if (typ->face_table) {
			g_free(typ->face_table);
			typ->face_table = NULL;
		}
		if (typ->library) {
			FT_Done_FreeType(typ->library);
			typ->library = NULL;
		}
	}
	g_free(typ);
	return 0;
}/*}}}*/

int typesetter_reset(Typesetter *typ)
{/*{{{*/
	typesetter_reset_style (typ);
	return 0;
}/*}}}*/

void typesetter_reset_style (Typesetter *typ)
{/*{{{*/
	int i;
	for (i=0; i<FONT_STYLE_MAX; i++) {
		if (typ->face_table[i].face) {
			FT_Done_Face(typ->face_table[i].face);
			vert_cache_table_destroy (typ->face_table[i].vcache_table);
			glyph_cache_table_destroy (typ->face_table[i].gcache_table);
			gsub_table_destroy(typ->face_table[i].gsub_table);
		}
		if (typ->face_table[i].style) {
			typesetter_fontstyle_destroy(typ->face_table[i].style);
		}
	}
	memset(typ->face_table, 0x00, sizeof(FontFaceTable) * FONT_STYLE_MAX);
}/*}}}*/

int typesetter_set_style (Typesetter *typ, guint32 style_type, FontStyle *style)
{/*{{{*/
	FT_Face face;
	FT_Error error;
	gchar *font_path = NULL;

	if (style_type > FONT_STYLE_MAX) {
		return 1;
	}
	if (style == NULL || style->name == NULL) {
		return 1;
	}

#ifdef _WIN32
	font_path = g_win32_locale_filename_from_utf8(style->file_path);
#else
	font_path = style->file_path;
#endif

	error = FT_New_Face(typ->library, font_path, style->face_index, &face);

#ifdef _WIN32
	if (font_path) {
		g_free(font_path);
		font_path = NULL;
	}
#endif

	IS_ERROR(error, "failed to get font face");
	typ->face_table[style_type].face = face;
	typ->face_table[style_type].vcache_table = vert_cache_table_create(face);
	typ->face_table[style_type].gcache_table = glyph_cache_table_create();
	typ->face_table[style_type].gsub_table = gsub_table_create(face);
	typ->face_table[style_type].style = style;
	
	error = face_set_size(face, style);
	IS_ERROR(error, "failed to set char size");

	//set charmap if it is null
	check_face_charmap(face);
	
	if (FT_IS_SFNT(face)) {
		typ->face_table[style_type].os2 = FT_Get_Sfnt_Table(face, ft_sfnt_os2);
		if (typ->face_table[style_type].os2->version == 0xFFFF) {
			typ->face_table[style_type].os2 = NULL;
		}
	} else {
		typ->face_table[style_type].os2 = NULL;
	}

	//compute base-line
	typ->face_table[style_type].base_line = face_get_base_line(face);

#ifdef DEBUG
	debug_print_font(face);
#endif
	if (style_type == FONT_STYLE_BASE) {
		FontStyle *ruby2_style = typesetter_fontstyle_copy(style);
		ruby2_style->size_x /= 2;
		ruby2_style->size_y /= 2;
		typesetter_set_style (typ, FONT_STYLE_RUBY2, ruby2_style);
	}
	return 0;
}/*}}}*/

int typesetter_get_empty_user_style_slot (Typesetter *typ)
{/*{{{*/
	int i;
	for (i = FONT_STYLE_USER; i < FONT_STYLE_MAX; i++) {
		if (typ->face_table[i].face == NULL) {
			return i;
		}
	}
	return -1;
}/*}}}*/

//tHg̃[hƃ_ÕIvV
static FontFlag create_font_flag (TypesetterOption* option)
{/*{{{*/
	FontFlag flag;

	flag.load   = FT_LOAD_DEFAULT;
	flag.render = 0;

	if (option->antialias == TRUE) {
		flag.load   |= FT_LOAD_NO_BITMAP;
		flag.render |= FT_RENDER_MODE_NORMAL;
	} else {
		flag.load   |= FT_LOAD_MONOCHROME;
		flag.render |= FT_RENDER_MODE_MONO;
	}
	flag.load |= FT_LOAD_IGNORE_GLOBAL_ADVANCE_WIDTH;
	return flag;
}/*}}}*/

static void debug_print_font(FT_Face face)
{/*{{{*/
	WNGMSG("family name:%s\n", face->family_name);
	WNGMSG("Face Num:%d, Face Index:%d\n", (int)face->num_faces, (int)face->face_index );
	WNGMSG("-----------------------------------\n");
	WNGMSG("FONT FACE FLAGS\n");
	WNGMSG("HAS_HORIZONTAL  =%d\n",	(0!=FT_HAS_HORIZONTAL(face)));
	WNGMSG("HAS_VERTICAL    =%d\n",	0!=FT_HAS_VERTICAL(face));
	WNGMSG("HAS_KERNING     =%d\n",	0!=FT_HAS_KERNING(face));
	WNGMSG("IS_SCALABLE     =%d\n",	0!=FT_IS_SCALABLE(face));
	WNGMSG("IS_SFNT         =%d\n",	0!=FT_IS_SFNT(face));
	WNGMSG("IS_FIXED_WIDTH  =%d\n",	0!=FT_IS_FIXED_WIDTH(face));
	WNGMSG("HAS_FIXED_SIZES =%d\n",	0!=FT_HAS_FIXED_SIZES(face));
	WNGMSG("HAS_GLYPH_NAMES =%d\n",	0!=FT_HAS_GLYPH_NAMES(face));
	WNGMSG("HAS_MULTIPLE_MASTERS=%d\n", 0!=FT_HAS_MULTIPLE_MASTERS(face));
	WNGMSG("-----------------------------------\n");
	WNGMSG("Charmap Num:%d\n", face->num_charmaps);
{
	int i;
	for (i=0; i<face->num_charmaps; i++) {
		char *enc, *plat, *eid;
		WNGMSG(face->charmap == face->charmaps[i] ? ">> ":"   ");

		WNGMSG("%d : ", i);
		switch (face->charmaps[i]->encoding) {
			case 0: enc="NONE "; break;
			case ft_encoding_symbol:         enc="MS_SYMBOL";break;
			case ft_encoding_unicode:        enc="UNICODE";break;
			case ft_encoding_sjis:           enc="SJIS";break;
			case ft_encoding_gb2312:         enc="GB2312";break;
			case ft_encoding_big5:           enc="BIG5";break;
			case ft_encoding_wansung:        enc="WANSUNG";break;
			case ft_encoding_johab:          enc="JOHAB";break;
			case ft_encoding_adobe_standard: enc="ADOBE_STANDARD";break;
			case ft_encoding_adobe_expert:   enc="ADOBE_EXPERT";break;
			case ft_encoding_adobe_custom:   enc="ADOBE_CUSTOM";break;
			case ft_encoding_latin_1:        enc="ADOBE_LATIN_1";break;
			case ft_encoding_latin_2:        enc="FT_ENCODING_GLD_LATIN_2"; break;
			case ft_encoding_apple_roman:    enc="APPLE_ROMAN";break;
			default: enc="UNKNOWNENC!";break;
		}
		switch (face->charmaps[i]->platform_id) {
			case TT_PLATFORM_APPLE_UNICODE: plat="APPLE_UNICODE";break;
			case TT_PLATFORM_MACINTOSH    : plat="MACINTOSH";break;
			case TT_PLATFORM_ISO          : plat="ISO";break;
			case TT_PLATFORM_MICROSOFT    : plat="MICROSOFT";break;
			case TT_PLATFORM_CUSTOM       : plat="CUSTOM";break;
			case TT_PLATFORM_ADOBE        : plat="ADOBE";break;
			default:plat="UNKNOWNPLATFORM!";break;
		}
		switch (face->charmaps[i]->encoding_id) {
			case TT_APPLE_ID_DEFAULT    :eid="DEFAULT";break;
			case TT_APPLE_ID_UNICODE_1_1:eid="UNICODE_1_1";break;
			case TT_APPLE_ID_ISO_10646  :eid="ISO_10646";break;
			case TT_APPLE_ID_UNICODE_2_0:eid="UNICODE_2_0";break;
			case TT_APPLE_ID_UNICODE_32 :eid="UNICODE_32";break;
			default: eid="UNKNOWNENCID!"; break;
		}
		WNGMSG("%s, plat=%s, enc=%s", enc, plat, eid);
		WNGMSG("\n");
	}
}
}/*}}}*/

int typesetter_render(Typesetter *typ, gchar *in, guint8 **outbuf, gint *outwidth, gint *outheight)
{/*{{{*/
	gchar *normalized = NULL;
	Layout *layout = NULL;
	BufferContainer buf;
	FontFlag flag;
	GList *seq = NULL;

	if (in == NULL) goto error;

	normalized = utf8_nfc_except_cjk_kanji(in);

	flag = create_font_flag(&typ->option);

	seq = get_gbox_sequence(typ, normalized, flag);
	if (seq == NULL) goto error;

	list_reset_style(typ, seq);
	seq = list_parse (typ, seq);
	seq = list_auto_yoko(typ, seq);

	dot_init(typ);
	layout = layout_create(typ, NULL);
	if (layout == NULL) goto error;

	layout_set_lines_from_sequence(layout, seq);
	
	layout->obj.func_set_dir(layout, typ->option.text_dir);

	layout->obj.func_get_glyph_id(layout);
	layout->obj.func_combine_glyph(layout);
	layout->obj.func_cache_glyph_image(layout);
	layout->obj.func_compute_gbox_size(layout, NULL, NULL);
	layout->obj.func_compute_self_size(layout);
	
	if (layout_alloc_buffer (layout, &buf)) goto error;
	layout->obj.func_render(layout, &buf);

	layout_destroy(layout);
	dot_destroy();

	*outwidth = buf.width;
	*outheight = buf.height;
	*outbuf = buf.data;
	if (normalized) g_free(normalized);
	if (seq) g_list_free(seq);
	return 0;
error:
	*outwidth = 0;
	*outheight = 0;
	*outbuf = NULL;
	if (normalized) g_free(normalized);
	if (seq) g_list_free(seq);
	return 1;
}/*}}}*/

int typesetter_get_pixel_size (Typesetter *typ, guint32 style_type)
{/*{{{*/
	FT_Face face = typ->face_table[style_type].face;
	if (!face) return 0;
	return face->size->metrics.x_ppem;
}/*}}}*/

FontStyle* typesetter_fontstyle_create(gchar *name,
									   gchar *file_path,
									   guint face_index,
									   guint dpi_x,
									   guint dpi_y,
									   FT_F26Dot6 size_x, 
									   FT_F26Dot6 size_y,
									   FontSizeUnit unit)
{/*{{{*/
	FontStyle *style = g_malloc0(sizeof(FontStyle));
	style->name = g_strdup(name);
	style->file_path = g_strdup(file_path);
	style->face_index = face_index;
	style->size_x = size_x;
	style->size_y = size_y;
	style->dpi_x = dpi_x;
	style->dpi_y = dpi_y;
	style->unit = unit;
	return style;
}/*}}}*/

FontStyle* typesetter_fontstyle_copy(FontStyle *style)
{/*{{{*/
	FontStyle *ret = typesetter_fontstyle_create(style->name,
												 style->file_path,
												 style->face_index,
												 style->dpi_x,
												 style->dpi_y,
												 style->size_x,
												 style->size_y,
												 style->unit);
	return ret;
}/*}}}*/

void typesetter_fontstyle_destroy(FontStyle *style)
{/*{{{*/
	if (style->name) g_free(style->name);
	if (style->file_path) g_free(style->file_path);
	g_free(style);
}/*}}}*/

/* vim: set fdm=marker ts=4 sw=4:*/
